#define NO_MSGS
#include <stdlib.h>
#include "Msgs.h"
#include "physics.h"
#include "tree.h"
#include "traverse.h"
#include "vop.h"
#include "do_grav.h"
#include "integrate.h"
#include "placebods.h"

/* And these control the actual interaction: */
static const float *BodyPos;		/* position of body - read only */
static float MassAccum;
static float *BodyPhip;
static float *BodyAcc;
static int BodyNterms;
static float eps2;

/* These should be controlled by main, but I can't be bothered to
   incorporate the relevant command line switches */
static int useGravOffset = 0;
static float GravOffset[NDIM];

static int 
BHGravMAC(Key_t key, const tcell *tc){
  cell *cp = Contents(tc);
  Vxd(float dr);
  Ivec_t iv;
  int ncut= 0;
  Vxd(float srcpos);

  IncrCounter(&BHMacCnt);
  if( useGravOffset ){
    VxVV(srcpos, = cp->pos, + GravOffset);
  }else{
    VxV(srcpos, = cp->pos);
  }

  VxVVx(dr, = BodyPos, - srcpos);
  dr2 = Dotx(dr, dr);
  if( dr2 < cp->rcrit * cp->rcrit ){
    Msgf(("dr2=%g, rcrit2=%g REJECT\n", dr2, cp->rcrit*cp->rcrit));
    return MAC_REJECT;
  }

  Msgf(("dr2=%g, rcrit2=%g ACCEPT\n", dr2, cp->rcrit*cp->rcrit));
  /* errsum? nterms? */
  /* This is gross and lazy.  But we really don't plan to use this code
     anyway */
  iv.u.used.mass = cp->mass;
  VVx(iv.u.used.pos, = srcpos);
  do_grav(&iv, &iv+1, BodyPos, &MassAccum, BodyAcc, BodyPhip, &eps2, &ncut);
  BodyNterms++;
  return MAC_ACCEPT;
}

static void 
BHGravDoBodies(const sib_group *sg){
  _tbody *_tb = Tbodies(sg);
  tbody *tbods = (tbody *)_tb;
  Ivec_t *iv = alloca(sg->ntbody * sizeof(Ivec_t));
  int i;
  int ncut = 0;

  for(i=0; i<sg->ntbody; i++){
    iv[i].u.used.mass = tbods[i].mass;
    if( useGravOffset ){
      VVV(iv[i].u.used.pos, = tbods[i].pos, + GravOffset);
    }else{
      VV(iv[i].u.used.pos, = tbods[i].pos);
    }
  }
  do_grav(iv, iv+sg->ntbody, BodyPos, &MassAccum, BodyAcc, BodyPhip, &eps2, &ncut);
  BodyNterms += sg->ntbody;
}

void
BHTraverse(float eps, float GNewt, gravdiags_t *diags){
  int b, p;
  bucket_t *tmp;

  eps2 = eps*eps;
  PlaceBodsBegin(NDIM, sizeof(body), decompranges, nbuckets, nextbodies, bucketdividers);
  DiagsZero(diags);
  /* We could use BucketToMem to replace most of the outer loops here,
     and to better encapsulate the details fo the bucket
     implementation.  On the other hand, this is a little more frugal
     with temporary space, and may be a little easier on the
     cache/tlb... */
  for(b=0; b<nbuckets; b++){
    bucket_t *bk;
    bk = &bodies[b];
    for(p=0; p<bk->npgs; p++){
      PgId pg = bk->pages[p];
      int nbod = bk->bytes_per_page[p]/sizeof(body);
      PgHndl hndl = PgRef(pg); 
      body *base = PgAddr(hndl);
      int i;

      /* And within each page, loop over bodies */
      for(i=0; i<nbod; i++){
	body *bp; 
	bp = &base[i];

	/* Set up appropriate static vars so the callbacks know where to
	   write the results */
	BodyPos = bp->pos;
	BodyAcc = bp->acc;
	BodyPhip = &bp->phi;
	BodyNterms = 0;
    
	/* Initialize accumulators to zero */
	MassAccum = 0.F;
	VS(BodyAcc, = 0.F);
	*BodyPhip = 0.F;
	
	TraverseSg(srcroot, BHGravMAC, BHGravDoBodies);
	bp->nterms = BodyNterms;
	VS(bp->acc, *= GNewt);
	bp->phi *= GNewt;
	Msgf(("body %d Nterms=%d, acc=(%g %g %g)\n", bp->ident, BodyNterms,
	      bp->acc[0], bp->acc[1], bp->acc[2]));
	/* Push the body to its new position and velocity. */
	DiagsAdd(diags, bp, 0.5*dt);
	IntegrateBody(bp);
	bp->key = KeyFromFloats(bp->pos, &UniverseBbox, nbodybits, key_order);
	if( KeyEQZ(bp->key) ){
	  Error("Body outside of UniverseBBox.  Need to handle this case!\n");
	}
	/* Save it into the appropriate bucket.  If buffers fill up,
	   this will generate communication. */
	PlaceBody(bp, bp->key);
	PagePoll();
      }
      PgUnref(hndl, PG_DIRTY);
    }
    BucketDestroy(&bodies[b]);
  }
  PagePollTillDone();
  PlaceBodsEnd();
  DiagsFinish(diags);

  /* Swap nextbodies and bodies */
  tmp = nextbodies;
  nextbodies = bodies;
  bodies = tmp;
}
