#define NO_MSGS
#include "Msgs.h"
#include "physics.h"
#include "tree.h"
#include "traverse.h"
#include "vop.h"
#include "do_grav.h"
#include "integrate.h"
#include "placebods.h"
#include "timers.h"

Counter_t BHMacCnt;

/* And these control the actual interaction: */
static const float *BodyPos;		/* position of body - read only */
static float MassAccum;
static int BodyNterms;
static float eps2;
static Ivec_t *iv, baseivec[16384];

static int 
BHvGravMAC(Key_t key, const tcell *tc){
  cell *cp = Contents(tc);
  Vxd(float dr);
  Vxd(float srcpos);

  IncrCounter(&BHMacCnt);
  VxV(srcpos, = cp->pos);
  VxVVx(dr, = BodyPos, - srcpos);
  dr2 = Dotx(dr, dr);
  if( dr2 < cp->rcrit * cp->rcrit ){
    Msgf(("dr2=%g, rcrit2=%g REJECT\n", dr2, cp->rcrit*cp->rcrit));
    return MAC_REJECT;
  }

  Msgf(("dr2=%g, rcrit2=%g ACCEPT\n", dr2, cp->rcrit*cp->rcrit));
  /* errsum? nterms? */
  /* This is gross and lazy.  But we really don't plan to use this code
     anyway */
  iv->u.used.mass = cp->mass;
  VVx(iv->u.used.pos, = srcpos);
  iv++;
  BodyNterms++;
  return MAC_ACCEPT;
}

static void 
BHvGravDoBodies(const sib_group *sg){
  _tbody *_tb = Tbodies(sg);
  tbody *tbods = (tbody *)_tb;
  int i;

  for(i=0; i<sg->ntbody; i++){
    iv->u.used.mass = tbods[i].mass;
    VV(iv->u.used.pos, = tbods[i].pos);
    iv++;
  }
  BodyNterms += sg->ntbody;
}

void
BHvTraverse(float eps, float GNewt, gravdiags_t *diags){
  int b, p;
  bucket_t *tmp;

  eps2 = eps*eps;
  PlaceBodsBegin(NDIM, sizeof(body), decompranges, nbuckets, nextbodies, bucketdividers);
  DiagsZero(diags);
  /* We could use BucketToMem to replace most of the outer loops here,
     and to better encapsulate the details fo the bucket
     implementation.  On the other hand, this is a little more frugal
     with temporary space, and may be a little easier on the
     cache/tlb... */
  for(b=0; b<nbuckets; b++){
    bucket_t *bk;
    bk = &bodies[b];
    for(p=0; p<bk->npgs; p++){
      PgId pg = bk->pages[p];
      int nbod = bk->bytes_per_page[p]/sizeof(body);
      PgHndl hndl = PgRef(pg); 
      body *base = PgAddr(hndl);
      int i;

      /* And within each page, loop over bodies */
      for(i=0; i<nbod; i++){
	body *bp; 
	int ncut;
	bp = &base[i];

	/* Set up appropriate static vars so the callbacks know where to
	   write the results */
	BodyNterms = 0;
	iv = baseivec;
    
	/* Initialize accumulators to zero */
	MassAccum = 0.F;
	BodyPos = bp->pos;
	VS(bp->acc, = 0.F);
	bp->phi = 0.F;
	
	TraverseSg(srcroot, BHvGravMAC, BHvGravDoBodies);
	ncut = 0;
	do_grav(baseivec, iv, bp->pos, &MassAccum, bp->acc, &bp->phi, &eps2, &ncut);
	bp->nterms = BodyNterms;
	VS(bp->acc, *= GNewt);
	bp->phi *= GNewt;
	Msgf(("body %d Nterms=%d, acc=(%g %g %g)\n", bp->ident, BodyNterms,
	      bp->acc[0], bp->acc[1], bp->acc[2]));
	/* Push the body to its new position and velocity. */
	DiagsAdd(diags, bp, 0.5*dt);
	IntegrateBody(bp);
	bp->key = KeyFromFloats(bp->pos, &UniverseBbox, nbodybits, key_order);
	if( KeyEQZ(bp->key) ){
	  Error("New position (%g %g %g) out of UniverseBbox\n",
		bp->pos[0], bp->pos[1], bp->pos[2]);
	}
	/* Save it into the appropriate bucket.  If buffers fill up,
	   this will generate communication. */
	PlaceBody(bp, bp->key);
      }
      PgUnref(hndl, PG_DIRTY);
    }
    BucketDestroy(&bodies[b]);
  }
  PagePollTillDone();
  PlaceBodsEnd();
  DiagsFinish(diags);

  /* Swap nextbodies and bodies */
  tmp = nextbodies;
  nextbodies = bodies;
  bodies = tmp;
}
