/*
 *  Buckets for managing multi-page objects (but small enough to fit in memory!
 */

#ifndef _BuckeTSdotH_
#define _BuckeTSdotH_
#include <stddef.h>
#include "page.h"

/* A Bucket is a (possibly) out-of-core logical block of storage that
   you can add to but not subtract from.  You BucketCreate() the
   bucket and then BucketAddData to it one piece at a time.  It's up
   to you to remember where the boundaries are between data objects.
   As far as the bucket is concerned, they all just get concatenated
   together.

   If this were C++ there would be an iterator over the pages in a bucket.
   But this is C, so you have to roll your own.  See the code fragment
   at the end for hints.  The BucketToMem source might also be helpful.

   BucketToMem offers an alternative scheme for iteration.  It ASSUMES
   that the bucket is small enough to fit in memory.  Perhaps you
   called Malloc(BucketSize(bk)) to get enough space.  (Presumably you
   have lots of buckets or you wouldn't be using out-of-core in the
   first place).

   BucketLastItemOnPage unrefs the bucket's 'current' page.  You MUST
   DO THIS when you are done adding to a bucket so that the paging
   system has the freedom to swap the last page in and out.  You can
   still add to a bucket after calling BucketLastItemOnPage, with the
   only restriction being that you are obliged to call
   BucketLastItemOnPage when you are done.  Another reason to call
   BLIOP is that for some reason you know that subsequent objects
   should be on a new page.
   
   This interface works on one bucket at a time.  If you have many
   buckets, and/or many processors, you need to keep track of them
   yourself.  But see placebods.h for help.
   */

typedef struct{
  int npgs;
  int npgs_allocated;
  PgId *pages;
  int *bytes_per_page;
  /* internal state for adding data */
  PgHndl hndl;
  void *ptr;
  size_t space;
}bucket_t ;

void BucketCreate(bucket_t *);
void BucketDestroy(bucket_t *);
void BucketToMem(bucket_t *, void *location);
void BucketAddData(bucket_t *, void *data, size_t sz);
void BucketLastItemOnPage(bucket_t *);
size_t BucketSize(bucket_t *);

#if 0 && defined(COMMENT) 
/* Example of iterating over an array of buckets (outer loop), and
   for each bucket, over its contents (inner two loops). */
typedef <something> body;

void
doBodies( void (*callback)(const body *) ){
  int b, p;
  
  for(b=0; b<nbuckets; b++){
    bucket_t *bk;
    bk = &bodies[b];
    for(p=0; p<bk->npgs; p++){
      PgId pg = bk->pages[p];
      int nbod = bk->bytes_per_page[p]/sizeof(body);
      PgHndl hndl = PgRef(pg); 
      body *base = PgAddr(hndl);
      int i;

      /* And within each page, loop over bodies */
      for(i=0; i<nbod; i++){
	callback(&base[i]);
      }
      PgUnref(hndl, PG_CLEAN);
    }
  }
}  
#endif /* COMMENT */

#endif /* _BuckeTSdotH_ */

