/*
 * compute new-style decompositions by looking at the tree.
 * This is much easier in the new OOC regime because we have
 * an shared-address-space model where any processor is free to
 * look at the pages of another.  Of course efficiency dictates
 * that this should be used sparingly.
 */

#include <stddef.h>
#include <math.h>
#include <values.h>
#include <stdlib.h>
#include "Assert.h"
#include "Malloc.h"
#include "mpmy.h"
#include "Msgs.h"

#include "tree.h"
#include "traverse.h"
#include "decomp.h"

static double
getTotalwgt(oocptr rootsg, SplitCellValue_t cCost, SplitBodsValues_t bVisit){
  sib_group *sg;
  PgHndl sghndl;
  int i;
  double sum = 0.F;

  sg = OOCFollow(rootsg, &sghndl);

  Msg("abmdone", ("getTotalWgt(%s)\n", PrintSg(sg)));
  if( cCost ){
    for(i=0; i<sg->ncell; i++){
      sum += cCost(sg->key, &sg->tcells[i]);
      Msg("abmdone", ("icell=%d, sum=%g\n", i, sum));
    }
  }
  
  if(  bVisit ){
    SplitTbodyInfo *tbinfo;
    tbinfo = alloca(sg->ntbody * sizeof(*tbinfo));
    (*bVisit)(sg, tbinfo);
    for(i=0; i<sg->ntbody; i++){
      sum += tbinfo[i].value;
      Msg("abmdone", ("ibody=%d, sum=%g\n", i, sum));
    }
  }
  
  PgUnref(sghndl, PG_CLEAN);
  return sum;
}

/*
 * myspeed is a relative measure of speed.  The only thing that matters
 * is that it is consistent across all processors.  BogoMIPS, gk9 MN/s
 * or some measurement from the last timestep are fine.
 * On return, answer[i] has the hi and low keys for each processor i's decomp
 * on the next timestep.
 */
void
decomp(double myspeed, krange_t *answer, int ndim, oocptr rootsg,
       SplitCellValue_t cCost, SplitBodsValues_t bVisit){
  double *allspeeds;
  double *speedsum;
  double totalspeed;
  double totalWgt;
  double targetCost;
  int i;
  Key_t *targetkeys;
  Key_t bestKey;
  Key_t kstart, bkey0, bkey7;

  allspeeds = Malloc(MPMY_Nproc() * sizeof(*allspeeds));
  speedsum = Malloc(MPMY_Nproc() * sizeof(*speedsum));
  targetkeys = Malloc( (MPMY_Nproc()+1) * sizeof(Key_t));
  MPMY_AllGather(&myspeed, 1, MPMY_DOUBLE, allspeeds);
  /* kstart = 0x100000..., a body key */
  bkey0 = KeyLshift(KeyInt(0x1), ndim* ((KEYBITS-1)/ndim));
  bkey7 = KeyOr( bkey0, KeyAddInt(bkey0, -1));
  kstart = bkey0;
  
  /* speedsum holds the fraction of the target sum of the wgts
     through the top of the indexed processor. */
  speedsum[0] = allspeeds[0];
  for(i=1; i<MPMY_Nproc(); i++){
    speedsum[i] = speedsum[i-1] + allspeeds[i];
  }
  totalspeed = speedsum[MPMY_Nproc() - 1];
  totalWgt = getTotalwgt(rootsg, cCost, bVisit);

  /* Processor p will look for target[p].  Note that there p-1
     targets, corresponding to the split point between processor i
     and i+1 for i in 0,...,p-2 */
  
  if( totalWgt > 0. ){
    if( MPMY_Procnum() != MPMY_Nproc()-1 ){
      targetCost = totalWgt*(speedsum[MPMY_Procnum()]/totalspeed);
      Msgf(("ksrch for targetCost=%g in tree with total weight: %g, myspeedsum=%g, totalspeed=%g\n",
	    targetCost, totalWgt, speedsum[MPMY_Procnum()], totalspeed));
      bestKey = kSrch(targetCost, rootsg, ndim, cCost, bVisit);
      Msgf(("best key found: %s\n", PrintKey(bestKey)));
    }else{
      bestKey = bkey7;
      Msgf(("decomp:  bestKey = %s\n", PrintKey(bestKey)));
    }
    Msg("abmdone", ("Calling PagePollDone in " __FUNCTION__ "\n"));
    PagePollTillDone();
  }else{
    /* There is nothing in the tree.  Just interpolate keys! */
    Msg("abmdone", ("totalWgt==0., no call to PagePollDone!\n"));
    bestKey = KeyInterpolate(bkey0, bkey7, speedsum[MPMY_Procnum()]/totalspeed);
  }

  Msgf(("procnum=%d bestkey=%s\n", MPMY_Procnum(), PrintKey(bestKey)));
  MPMY_AllGather(&bestKey, sizeof(Key_t), MPMY_CHAR, targetkeys+1);
  /* the first lo key should be 0x100000 */
  targetkeys[0] = bkey0;

  for(i=0; i<MPMY_Nproc(); i++){
    answer[i].lo = targetkeys[i];
    answer[i].hi = KeyAddInt(targetkeys[i+1], -1);
    assert( KeyGE(answer[i].hi, answer[i].lo) );
  }
  /* The last hi value should be 1777... */
  answer[MPMY_Nproc()-1].hi = bkey7;

  Free(targetkeys);
  Free(allspeeds);
  Free(speedsum);
}
