#ifndef _DecOMPdotH_
#define _DecOMPdotH_

#include "tree.h"
#include "key.h"
#include "page.h"

/* A parallel decomposition in the new world order requires two steps.

   First, we decompose into processors using decomp().  This step will
   require some interprocessor communication.  The result is a list of
   key pairs indicating ranges.  In this step we are trying to balance
   the workload divided by the processor speed.  We don't necessarily
   want equal work on each processor.  Use decomp() for this step.
   The 'cost' callbacks should reflect workload.
   It invokes global communication routines and everybody is guaranteed
   to get exactly the same results.

   Then we (optionally) split up the range of values assigned to processors
   into 'buckets'.  The buckets should be small enough to fit comfortably
   in memory.  The number may be different on each node.  Some nodes
   may not even do this step.  Use splitKeys() for this step.  The
   'cost' callbacks should reflect memory usage.  There is no
   global communication, but there IS a tree-traversal which might
   invoke communication if remote parts of the tree are examined. */

/* decomp and splitKeys fill in arrays of krange_t on return.  */
typedef struct {
  Key_t lo;
  Key_t hi;
} krange_t;


/* The two call back prototype typedefs are used by decomp() and
   splitKeys().  SplitCellValue_t is easy: it takes a tcell and
   returns a value.   */
typedef double (*SplitCellValue_t)(Key_t, tcell * );

/* SplitBodsValue_t is trickier.  Since the tree does not record
   knowledge about where one tbody ends and the next one starts, we
   can't ask one-at-a-time.  Thus, we ask about all of them at once.
   Similarly, the tree does not record information about tbody keys, so
   it's up to the caller to provide some way to recover a key from a
   tbody.  Note that this will be called relatively few times
   (O(lgN)), so there's little harm if reconstructing a key is slow.
   It's better to be slow, but save memory, than to waste the
   memory recording a key.  The array of SplitTbodyInfo should be
   filled in by the SplitBodsValues_t callback.
*/
typedef struct {
  double value;
  Key_t key;
} SplitTbodyInfo;

typedef void (*SplitBodsValues_t)(sib_group * , SplitTbodyInfo *);

/* kSrch searches for the value in key space at which the accumulated
   'values' of the bodies and cells add up to a target value.  It's
   exactly what you want when you search for a key-value at which to
   decompose the key space.  It can be used to decompose based on either
   memory or time by suitably changing the callbacks. */
Key_t kSrch(double target, oocptr, int ndim, SplitCellValue_t, SplitBodsValues_t);

/* The principal user callable functions */
void decomp(double myspeed, krange_t procranges[], int ndim, oocptr rootsg,
	    SplitCellValue_t cellSz, SplitBodsValues_t bodySzs);
void splitKeys(int nbuckets, krange_t results[], int ndim, 
	       oocptr rootsg, Key_t lo, Key_t hi, 
	       SplitCellValue_t cellSz, SplitBodsValues_t bodySzs);

#endif /* _DecOMPdotH_ */
