#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <signal.h>
#include "protos.h"
#include "keycvt.h"
#include "gc.h"
#include "mpmy.h"
#include "mpmy_abnormal.h"
#include "randoms.h"
#include "Msgs.h"
#include "key.h"
#include "keycvt.h"
#include "malloc.h"
#include "fastflpt.h"
#include "page.h"
#include "Malloc.h"
#include "timers.h"
#include "singlio.h"
#include "error.h"
#include "Assert.h"

#include "physics.h"
#include "cmdline.h"
#include "vop.h"
#include "buckets.h"
#include "placebods.h"
#include "fiddledirs.h"
#include "sampledecomp.h"

#include "tree.h"
#include "traverse.h"
#include "decomp.h"

static void ReportStats(void);
static Timer_t CreateTm, InitDecompTm;

tbbox UniverseBbox;
double Mtot;
int nbodybits;

/* bodies is an array of buckets holding the bodies we are working on
   now, i.e., the current timestep */
bucket_t *bodies;
/* nextbodies is an array of buckets holding the bodies being assembled
   for the next timestep */
bucket_t *nextbodies;
/* decompranges records the range of body keys for each processor */
krange_t *decompranges;			/* filled by decomp */
/* bucketdividers records the rang of body keys for each bucket */
krange_t *bucketdividers;		/* filled by splitKeys */

/* The root of the source tree */
oocptr srcroot;

static void
CreateBodies(krange_t *decompranges){
  ran_state rs;
  int i, ii;
  body bod;
  int nobj, first_id;
  int nsamp, isamp;
  Key_t *samples;
  int bodies_per_samp;
  bucket_t initial_bucket;
  int p;

  StartTimer(&CreateTm);
  UniverseBbox.ndim = NDIM;
  /* Make the Univers twice as big as our particles */
  VS(UniverseBbox.rmin, = -1.2);
  VS(UniverseBbox.sz, = 2.4);

  Mtot = 0.;
  ran_init(10*(1+MPMY_Procnum()), &rs);

  BucketCreate(&initial_bucket);
  NobjInitial(gnobj, MPMY_Nproc(), MPMY_Procnum(), &nobj, &first_id);
  printf("Creating %d bodies starting at %d on %d\n", 
	 nobj, first_id, MPMY_Procnum());
  /* Let's do about 50 samples per bucket.  That should be plenty to
       get a good decomp.  How much space does that use (on node 0)?
       400bytes*Nproc*nbuckets =
       400bytes*Nproc*(sizeof(bodies)/sizeof(mem)) It's unlikely that
       sizeof(bodies)/sizeof(mem) > 100.  So it's less than
       40kB*Nproc.  Which should fit comfortably almost anywhere. */
  if( nobj > 0 ){
    nsamp = 50*nbuckets;
    bodies_per_samp = (nobj-1)/nsamp + 1;
  }else{
    nsamp = 0;
    bodies_per_samp = 1;
  }
  samples = Malloc(sizeof(Key_t)*nsamp);
  ii = bodies_per_samp;		/* get the first one */
  isamp = 0;
  for(i=0; i<nobj; i++){
    float r;
    bod.mass = 2.F*uniform_rand(&rs)/gnobj;
    bod.nterms = 1;
    Mtot += bod.mass;
    sphere_rand(&rs, NDIM, bod.pos);
    bod.key = KeyFromFloats(bod.pos, &UniverseBbox, nbodybits, key_order);

    r = sqrtf_fast(Dot(bod.pos, bod.pos));
    sphere_rand(&rs, NDIM, bod.vel);
      
    /* if v \propto r, we get approximately 0 energy.  But what's the
     constant?  Gaussian might be more interesting... */
    VS(bod.vel, *= r);
    bod.ident = first_id + i;
    if( bod.ident%100000 == 0 )
      printf("body %d added to initial bucket\n", bod.ident);
    bod.key = KeyFromFloats(bod.pos, &UniverseBbox, nbodybits, key_order);
    BucketAddData(&initial_bucket, &bod, sizeof(body));
    if( ii == bodies_per_samp ){
      ii = 0;
      samples[isamp++] = bod.key;
    }
    ii++;
  }
  StopTimer(&CreateTm);

  StartTimer(&InitDecompTm);
  BucketLastItemOnPage(&initial_bucket);
  printf("Bodies created in out-of-core storage\n");
  assert(isamp <= nsamp);
  nsamp = isamp;

  SampleDecomp(1.0, nbuckets, nsamp, NDIM, samples, 
	       decompranges, bucketdividers);
  PlaceBodsBegin(NDIM, sizeof(body), decompranges, nbuckets, bodies, bucketdividers);
  /* Loop through initial_bucket and call PlaceBody() for each*/
  ii = 0;
  for(p=0; p<initial_bucket.npgs; p++){
    PgId pg = initial_bucket.pages[p];
    int nbod = initial_bucket.bytes_per_page[p]/sizeof(body);
    PgHndl hndl = PgRef(pg); 
    body *bp = PgAddr(hndl);
    int i;
    
    /* And within each page, loop over bodies */
    for(i=0; i<nbod; i++){
      PlaceBody(&bp[i], bp[i].key);
      if( ii++%100000 == 0 )
	printf("Placed body %d\n", ii);
    }
    PgUnref(hndl, PG_CLEAN);
  }
  PlaceBodsEnd();
  BucketDestroy(&initial_bucket);

  MPMY_Combine(&Mtot, &Mtot, 1, MPMY_DOUBLE, MPMY_SUM);
  StopTimer(&InitDecompTm);
}

static void
ReadBodies(const char *filename, krange_t *decompranges){
  Error("ReadBodies not implemented\n");
}

/* With the 'bucket' strategy, looping over all bodies is 
   a little uglier than for(i=0; i<nobj; i++).  So we encapsulate
   the simplest case here with a callback.  Use with care - it
   swaps the entire disk-resident contents of the 'btab'.  */
void
doBodies( void (*callback)(const body *) ){
  int b, p;
  
  for(b=0; b<nbuckets; b++){
    bucket_t *bk;
    bk = &bodies[b];
    for(p=0; p<bk->npgs; p++){
      PgId pg = bk->pages[p];
      int nbod = bk->bytes_per_page[p]/sizeof(body);
      PgHndl hndl = PgRef(pg); 
      body *base = PgAddr(hndl);
      int i;

      /* And within each page, loop over bodies */
      for(i=0; i<nbod; i++){
	callback(&base[i]);
      }
      PgUnref(hndl, PG_CLEAN);
    }
  }
}  

static void
PrintBodyToStdout(const body *b){
  printf("%s\n", PrintBodyContents(b));
}

void alrmhndlr(int sig){
  Error("Caught Alarm.  We're outahere!\n");
}

int
main(int argc, char **argv){
  tree_t thetree;
  Timer_t BuildTm, NlgNTm, DLTm, StepTm;
  int step;
  int i;
  bucket_t *buckets1, *buckets2;
  gravdiags_t diags;

  /* Initialize everything.  Maximum debugging */
  MPMY_Init(&argc, &argv);
  MPMY_OnAbnormal(MPMY_SystemAbort);
  singlPrintf("#Command line:\n#");
  for(i=0; i<argc; i++){
    singlPrintf("%s ", argv[i]);
  }
  singlPrintf("\n");

  ParseArgs(argc, argv);
  FiddleDirs(newdir_fmt, MPMY_Procnum());
  /* Chdir yet again if we dump core.  This works even if we have multiple
     processes in the same directory.  On the other hand, it's often
     easier to find 'core' than core.<??>/core.  Fiddledirs has already
     placed us in 'newdir', which ought to be on a local disk, so 
     shouldn't thrash nfs if we crash. */
  MPMY_OnAbnormal(MPMY_Abchdir);
  sprintf(MPMY_Abchdir_arg, "core.%d", MPMY_Procnum());

  if( timeout > 0 ){
    signal(SIGALRM, alrmhndlr);
  }
    
  if( msg_arg )
    Msg_turnon(msg_arg);
  else
    Msg_set_enable(0);

  if( debug ){
    /* Do extended sanity checks on every sg_tmp, as it is created */
    tree17keyFromTbody = &keyFromTbody;
    /* Do extended sanity checks on malloc heap whenever we touch it */
    malloc_debug(2);
  }

  printf("Running on %s, procnum=%d, nprocs=%d\n", 
	 MPMY_Physnode(), MPMY_Procnum(), MPMY_Nproc());

  /* Set up the page tables. */
  PgInit(lgpgsize, wsetpages, swapfile);

  /* Timers */
  EnableWCTimer(&CreateTm, "Create");
  EnableWCTimer(&InitDecompTm, "InitDecomp");
  EnableWCTimer(&SortTm, "Sort");
  EnableWCTimer(&BuildTm, "Build");
  EnableWCTimer(&StepTm, "Step");
  EnableWCTimer(&RemoteTm, "Remote");
  EnableWCTimer(&PgPollTm, "Poll");
  EnableWCTimer(&PgPollLoopTm, "PollLoop");
  EnableWCTimer(&PgPollTillDoneTm, "PollTillDone");
  EnableCounter(&PgPollLoopCnt, "PollLoop");
  EnableCounter(&PgInCnt, "Pagein");
  EnableCounter(&PgOutCnt, "Pageout");
  EnableCounter(&PgRemoteInCnt, "Remotein");
  EnableCounter(&PgRemoteSendCnt, "Remoteout");
  if( do_dl ){
    EnableWCTimer(&DLTm, "DL");
    EnableWCTimer(&DLPollWaste, "DLPoll");
    EnableCounter(&DLMacCnt, "DLMac");
    EnableCounter(&DLUncertainMacCnt, "DLUncertain");
    EnableCounter(&DLUncertainSplitSinkCnt, "DLUncSplSink");
    EnableCounter(&DLPeelCnt, "DLPeel");
    EnableCounter(&DLDepthCnt, "DLDepth");
  }
  if( do_nlgn ){
    EnableWCTimer(&NlgNTm, "NlgN");
    EnableCounter(&BHMacCnt, "BHMac");
  }
    
  fflush(stdout);


  printf("Creating bodies in %d buckets\n", nbuckets);
  decompranges = Malloc((MPMY_Nproc())*sizeof(krange_t));
  bucketdividers = Malloc((nbuckets)*sizeof(krange_t));
  buckets1 = Malloc(nbuckets*sizeof(bucket_t));
  buckets2 = Malloc(nbuckets*sizeof(bucket_t));

  bodies = buckets1;
  nextbodies = buckets2;
  nbodybits = (KEYBITS-1)/NDIM;

  /* Create or read the body data */
  if( datafile ) {
    ReadBodies(datafile, decompranges);
  }else{
    CreateBodies(decompranges);
  }

  singlPrintf("Done creating bodies\n");
  /* PgStats(); */
  if( _Msg_test("memleak") )
    malloc_print();
  fflush(stdout);

  /* Iterate */
  for(step=0; step<nsteps; step++){
    StartTimer(&StepTm);

    /* The timeout is per-step. */
    if( timeout>0 )
      alarm(timeout);

    if( _Msg_test("memleak") )
      malloc_print();

    /* BUILD (and SORT) */
    StartTimer(&BuildTm);
    printf("Starting tree build with occupancy=%d, NDIM=%d\n", occupancy, NDIM);
    TreeBuildStart(&thetree, NDIM, occupancy, &GravBfuncs);
    SortAndAddToTree(&thetree);

    if( do_print_tree ){
      doBodies(PrintBodyToStdout);
      fflush(stdout);
    }

    srcroot = TreeBuildFinish(&thetree);
    StopTimer(&BuildTm);

    singlPrintf("Buildtree, finished in %.2f WCsec.\n", ReadTimer(&BuildTm));
    /* PgStats(); */
    fflush(stdout);
  
    /* PRINT */
    if( do_print_tree ){
      BHPrintTree(1, printf);
      singlPrintf("After printing\n");
      PgStats();
      if( _Msg_test("memleak") )
	malloc_print();
      fflush(stdout);
    }

    if( do_check_tree ){
      chktree(srcroot);
    }

    /* Decomp for next timestep.  We set decompranges and bucketdividers
     now to define the decomposition on the next timestep.  We should
     also update the Unviverse bbox to guarantee that nothing moves outside
     the bbox (or we can implement other bc).
     */
    decomp(1.0, decompranges, NDIM, srcroot, &cellCost, &bodyCostVisit);
    splitKeys(nbuckets, bucketdividers, NDIM, srcroot,
	      decompranges[MPMY_Procnum()].lo, 
	      decompranges[MPMY_Procnum()].hi, 
	      &cellSz, &bodySzs);
    /* What about a new Universe?? */

    /* NLGN WALK */
    if( do_nlgn ){
      StartTimer(&NlgNTm);
      
      if( do_vectorize ){
	BHvTraverse(0.01 /* epsilon */, 1.0 /* GNewt */, &diags);
      }else{
	BHTraverse(0.01 /* epsilon */, 1.0 /* GNewt */, &diags);
      }

      StopTimer(&NlgNTm);
      singlPrintf("After NlgN\n");
      PgStats();
      printf("NlgN computed total time: %.2f\n",
	     ReadTimer(&NlgNTm));
    }else if( do_dl ){
      StartTimer(&DLTm);
      DLTraverse(0.01 /* epsilon */, 1.0 /* GNewt */, &diags);
      StopTimer(&DLTm);
      PgStats();
      printf("DL computed, total time: %.2f\n", ReadTimer(&DLTm));
    }
    if( do_dl || do_nlgn ){
      DiagsPrint(&diags);
      fflush(stdout);
    }

    /* CLEAN UP */
    TreeFree(&thetree);
    if( PgNrefcnt > 0 ){
      Warning("After TreeFree: %d refcnt pages.\n", PgNrefcnt);
    }

    if( verbose || _Msg_test("memleak") )
      malloc_print();

    ReportStats();
    StopTimer(&StepTm);
    OutputTimers(singlPrintf);
    OutputCounters(singlPrintf);
    ClearEnabledTimers();
    printf("steptm after clear: %f\n", ReadTimer(&StepTm));
    ClearEnabledCounters();
    fflush(stdout);
  }

  PgTerm();
  for(i=0; i<nbuckets; i++){
    BucketDestroy(&bodies[i]);
  }
  Free(buckets2);
  Free(buckets1);
  Free(bucketdividers);
  Free(decompranges);

  if( verbose || _Msg_test("memleak") ){
    Msg_do("Just before exiting, memory looks like:\n");
    malloc_print();
  }
  MPMY_Finalize();
  exit(0);
}

/* This looks an awful lot like the internals of SumCounter.  It's
   probably redundant */
int
ireduction(int v, int *min, int *max, double *sum){
  MPMY_Comm_request req;

  MPMY_ICombine_Init(&req);
  if( min ){
    *min = v;
    MPMY_ICombine(min, min, 1, MPMY_INT, MPMY_MIN, req);
  }
  if( max ){
    *max = v;
    MPMY_ICombine(max, max, 1, MPMY_INT, MPMY_MAX, req);
  }
  if( sum ){
    *sum = (double)v;
    MPMY_ICombine(sum, sum, 1, MPMY_DOUBLE, MPMY_SUM, req);
  }
  MPMY_ICombine_Wait(req);
  return 0;
}

static void
ReportStats(void){
  int min, max;
  double sum;

  singlPrintf("%12s%12s%12s%12s\n", "", "min", "max", "sum"); 
  ireduction(malloc_heapsz(), &min, &max, &sum);
  singlPrintf("%12s%10.2fMB%10.2fMB%10.2fMB\n",
	      "Malloc heapsz", min/1.e6, max/1.e6, sum/1.e6);
  ireduction(malloc_used(), &min, &max, &sum);
  singlPrintf("%12s%10.2fMB%10.2fMB%10.2fMB\n",
	      "Malloc inuse", min/1.e6, max/1.e6, sum/1.e6);
}
