#include <stdlib.h>
#include <string.h>

#include "hwclock.h"
#include "keycvt.h"
#include "Malloc.h"
#include "error.h"
#include "Msgs.h"
#include "mpmy.h"
#include "abm.h"
#include "chn.h"

#include "decomp.h"
#include "buckets.h"
#include "placebods.h"

static int nbuckets;
static bucket_t *tobuckets;
static krange_t *bucketranges;
static krange_t *procranges;
static int *correspondents;
static int ncorr;
static ABM abm;
static Chn outchn;
static size_t unitsz;

#define ABM_PROCESS_BODY 0
#define ABM_MARKER 1
#define ABM_REPLY_MARKER 2
static ABMhndlr_t abm_process_body;
static ABMhndlr_t abm_process_marker;
static ABMhndlr_t abm_process_reply_marker;

ABMhndlr_t *hndlarray[] = {
  &abm_process_body,
  &abm_process_marker,
  &abm_process_reply_marker
};

/* A callback suitable for use in bsearch.  Unfortunately, this
 has to know where the key is hidden in the body pointer.  If this
 were passed in as an arg to placeBodsBegin then placebods would
 not need physics.h */

static int
cmp(const void *kp, const void *_krange){
  Key_t k = *(Key_t *)kp;
  const krange_t *krange = (const krange_t *)_krange;

  if( KeyLT(k, krange->lo) )
    return -1;
  else if( KeyGT(k, krange->hi) )
    return 1;
  else
    return 0;
}

void
PlaceBodsBegin(int ndim, size_t unitsz_, krange_t *procranges_, int nbuckets_, bucket_t *tobuckets_, krange_t *bucketranges_){
  int i;

  unitsz = unitsz_;
  nbuckets = nbuckets_;
  tobuckets = tobuckets_;
  procranges = procranges_;
  bucketranges = bucketranges_;
  correspondents = Malloc(MPMY_Nproc() * sizeof(int));

  for(i=0; i<nbuckets; i++){
    BucketCreate(&tobuckets[i]);
  }

  Msgf(("Placebodsbegin: %d buckets, %d procs\n", nbuckets, MPMY_Nproc()));
  Msgf(("procranges: \n"));
  for(i=0; i<MPMY_Nproc(); i++){
    Msgf(("%d %s ", i, PrintKey(procranges[i].lo)));
    Msgf(("%s\n", PrintKey(procranges[i].hi)));
  }
  Msgf(("buckets:\n"));
  for(i=0; i<nbuckets; i++){
    Msgf(("%d %s ", i, PrintKey(bucketranges[i].lo)));
    Msgf(("%s\n", PrintKey(bucketranges[i].hi)));
  }

  ChnInit(&outchn, sizeof(Key_t)+unitsz, 1000, Realloc_f);
  ABMSetup(&abm, 12000, 3938, sizeof(hndlarray)/sizeof(*hndlarray), hndlarray);
}

void
PlaceBodsEnd(void){
  int i;

  Msgf(("PlaceBodsEnd: bucketranges[%d] = (%s, ", nbuckets-1, PrintKey(bucketranges[nbuckets-1].lo)));
  Msgf(("%s)\n", PrintKey(bucketranges[nbuckets-1].hi)));
  ABMFlush(&abm);
  Msgf(("afterABMflush: bucketranges[%d] = (%s, ", nbuckets-1, PrintKey(bucketranges[nbuckets-1].lo)));
  Msgf(("%s)\n", PrintKey(bucketranges[nbuckets-1].hi)));

  /* The ABM library says you can't call ABMIamDone until you are sure that
     all your messages have been RECEIVED AND PROCESSED.  The only way to
     know that is to send a MARKER to every one of your correspondents and
     wait for a reply. */
  ncorr = 0;
  for(i=0; i<MPMY_Nproc(); i++){
    if( correspondents[i] ){
      ABMPost(&abm, i, 0, ABM_MARKER, NULL, NULL);
      ncorr++;
    }
  }
  while( ncorr ){
    ABMPollWait(&abm);
  }
  Free(correspondents);

  ABMIamDone(&abm);
  while( !ABMAllDone(&abm) ){
    Msgf(("!Alldone: bucketranges[%d] = (%s, ", nbuckets-1, PrintKey(bucketranges[nbuckets-1].lo)));
    Msgf(("%s)\n", PrintKey(bucketranges[nbuckets-1].hi)));
    ABMPollWait(&abm);
  }
  ABMShutdown(&abm);
  procranges = NULL;
  bucketranges = NULL;
  ChnTerminate(&outchn);
  for(i=0; i<nbuckets; i++){
    BucketLastItemOnPage(&tobuckets[i]);
  }
}

static void 
memcpy_and_chnfree(void *to, void *arg, int sz){
  memcpy(to, arg, sz);
  ChnFree(&outchn, arg);
}

static void
abm_process_body(int src, int len, void *ptr){
  krange_t *kr;
  int b;

  Msgf(("abm_process_body(src=%d, len=%d, ptr=%p)\n", src, len, ptr));
  Msgf(("abm_process_body: bucketranges[%d] = (%s, ", nbuckets-1, PrintKey(bucketranges[nbuckets-1].lo)));
  Msgf(("%s)\n", PrintKey(bucketranges[nbuckets-1].hi)));
  assert(unitsz+sizeof(Key_t) == len);
  /* The ptr contains {key, bodydata}.  So we can pass it directly
     to bsearch (which expects a key pointer) */
  kr = bsearch(ptr, bucketranges, nbuckets, sizeof(krange_t), cmp);
  b = kr - bucketranges;
  Msgf(("abm_process_body: b = %d\n", b));
  /* Be careful to add sizeof(Key_t) to get a ptr to the data. */
  BucketAddData(&tobuckets[b], (char*)ptr+sizeof(Key_t), unitsz);
}

static void
abm_process_marker(int src, int len, void *ptr){
  assert(len == 0);
  Msgf(("processing marker from %d, posting reply\n", src));
  ABMPost(&abm, src, 0, ABM_REPLY_MARKER, NULL, NULL);
}

static void
abm_process_reply_marker(int src, int len, void *ptr){
  assert(len == 0);
  assert( correspondents[src]  && ncorr > 0 );
  correspondents[src] = 0;
  ncorr--;
  Msgf(("processing reply_marker from %d, ncorr now %d\n", src, ncorr));
}

void
PlaceBody(void *_bp, Key_t key){
  int b;
  krange_t *kr;
  static hwclock_t tlastpoll;
  double tdiff;
  hwclock_t tnow;

  /* Figure out where the body belongs, 
     if on some other processor queue for delivery (abm?)
     if on this processor then determine which bucket and do BucketAddData

     When we're all done, we need to flush the last of the outqueues,
     and then another ABMdone exercise.  We can then swap 'buckets' and
     'tobuckets' to be ready for the next timestep.  Do we need another function?

     What about the initial BucketCreate and final BucketDestroy?  Where should that
     happen?
     */

  if( cmp(&key, &procranges[MPMY_Procnum()]) == 0 ){
    kr = bsearch(&key, bucketranges, nbuckets, sizeof(krange_t), cmp);
    b = kr - bucketranges;
    Msgf(("PlaceBody(k=%s), to local bucket %d\n", PrintKey(key), b));
    BucketAddData(&tobuckets[b], _bp, unitsz);
  }else{
    void *p;
    int proc;
    kr = bsearch(&key, procranges, MPMY_Nproc(), sizeof(krange_t), cmp);
    proc = kr - procranges;
    /* ABMPost isn't exactly right here.  It has the undesirable property that
       the message isn't actually copied into an out-queue until it is ready to
       be delivered.  This is not what we want, so we have to make an additional
       copy.  Furthermore, have a choice between:
       - requiring a callback argument that returns a key
       - sending an extra copy of the key along with the body data.
       yuck!
       */
    p = ChnAlloc(&outchn);
    *(Key_t*)p = key;
    memcpy((char *)p+sizeof(Key_t), _bp, unitsz);
    correspondents[proc] = 1;
    ABMPost(&abm, proc, unitsz+sizeof(Key_t), ABM_PROCESS_BODY, memcpy_and_chnfree, p);
    Msgf(("PlaceBody(k=%s), to remote processor %d\n", PrintKey(key), proc));
  }
  hwclock(&tnow);
  tdiff = (tnow - tlastpoll) * hwsec_per_clock() ;
  if( tdiff > 0.01 ){
    Msgf(("Time for a poll!, it's been %g sec\n", tdiff));
    ABMPoll(&abm);
    tlastpoll = tnow;
  }else{
    Msgf(("Don't poll, tdiff=%g\n", tdiff));
  }
}

