#include <stdlib.h>
#include <math.h>

#include "protos.h"
#include "mpmy.h"
#include "Malloc.h"
#include "Msgs.h"
#include "key.h"

#include "decomp.h"

#include "sampledecomp.h"

static int
kcompare(const void *_k1, const void *_k2){
  Key_t k1 = *(const Key_t *)_k1;
  Key_t k2 = *(const Key_t *)_k2;
  if( KeyGT(k1, k2) )
    return 1;
  if( KeyLT(k1, k2) )
    return -1;
  return 0;
}

/* On input, x is between 0 and nsamp.  Assume the 'samples' are the
   values of a function at abcissas 0.5, 1.5, 2.5, ..., nsamp-0.5.

   Figure out which two samples the x actually falls between.  Do
   a linear interpolation between them.  If its at one of the ends,
   assume that the f(0)=minkey and f(nsamp)=maxkey and interpolate
   appropriately */
static Key_t
KeyLinterp(double x, int nsamp, Key_t *samples, Key_t minkey, Key_t maxkey){
  int ilo, ihi;
  double xlo, xhi;
  Key_t klo, khi;

  ilo = floor(x - 0.5);
  if( ilo < 0 ){
    klo = minkey;
    xlo = 0.;
  }else{
    klo = samples[ilo];
    xlo = ilo + 0.5;
  }
  Msgf(("xlo=%g, ilo=%d, klo=%s\n", xlo, ilo, PrintKey(klo)));

  ihi = ilo + 1;
  if( ihi >= nsamp ){
    khi = maxkey;
    xhi = 1.0;
  }else{
    khi = samples[ihi];
    xhi = ihi + 0.5;
  }
  Msgf(("xhi=%g, ihi=%d, khi=%s\n", xhi, ihi, PrintKey(khi)));

  return KeyInterpolate(klo, khi, (x-xlo)/(xhi - xlo));
}

void
SampleDecomp(double myspeed, int mynbuckets, int mynsamp, int ndim,
	     Key_t *samples, krange_t *decompranges, krange_t *bucketranges){
  int amRoot;
  int nsamp, maxnbuck;
  int *allnsamp=NULL, *allnbuck=NULL;
  double *allspeeds=NULL;
  krange_t *bucks = NULL;
  double totalspeed, speedsum, speedfrac, lastfrac;
  Key_t bkey0, bkey7;
  Key_t *allsamples=NULL;
  Key_t k;
  int i, b;
  MPMY_Comm_request req;
  MPMY_Status stat;

  amRoot = MPMY_Procnum() == 0;

  if( amRoot ){
    allspeeds = Malloc(MPMY_Nproc() * sizeof(Key_t));
    allnsamp = Malloc(MPMY_Nproc() * sizeof(Key_t));
    allnbuck = Malloc(MPMY_Nproc() * sizeof(int));
  }
  MPMY_Gather(&myspeed, 1, MPMY_DOUBLE, allspeeds, 0);
  MPMY_Gather(&mynbuckets, 1, MPMY_INT, allnbuck, 0);
  MPMY_Combine(&mynsamp, &nsamp, 1, MPMY_INT, MPMY_SUM);
  MPMY_NGather(samples, mynsamp*sizeof(Key_t), MPMY_CHAR, (void **)&allsamples, 0);

  MPMY_Irecv(bucketranges, mynbuckets*sizeof(krange_t), 0, 5432, &req);

  if( amRoot ){
    /* Do all the work on node 0.  */
    qsort(allsamples, nsamp, sizeof(Key_t), kcompare);

    for(i=0; i<nsamp; i++){
      Msgf(("allsamples[%d] = %s\n", i, PrintKey(allsamples[i])));
    }
    totalspeed = 0.;
    maxnbuck = -1;
    for(i=0; i<MPMY_Nproc(); i++){
      totalspeed += allspeeds[i];
      maxnbuck = (allnbuck[i]> maxnbuck)? allnbuck[i] : maxnbuck;
    }
    bucks = Malloc(maxnbuck*sizeof(krange_t));

    speedsum = 0.;
    bkey0 = KeyLshift(KeyInt(0x1), ndim* ((KEYBITS-1)/ndim));  
    bkey7 = KeyOr( bkey0, KeyAddInt(bkey0, -1));
    decompranges[0].lo = bkey0;
    decompranges[MPMY_Nproc()-1].hi = bkey7;
    speedfrac = lastfrac = 0.;
    for(i=1; i<MPMY_Nproc(); i++){
      /* find the divider between proc i-1 and proc i */
      speedsum += allspeeds[i-1];
      lastfrac = speedfrac;
      speedfrac = nsamp * (speedsum / totalspeed);
      k = KeyLinterp(speedfrac, nsamp, allsamples, bkey0, bkey7);
      Msgf(("Decomp interpolate(%g) = %s\n", speedfrac, PrintKey(k)));
      decompranges[i].lo = k;
      decompranges[i-1].hi = KeyAddInt(k, -1);
      /* Now do the buckets for proc i-1 */
      bucks[0].lo = decompranges[i-1].lo;
      for(b=1; b<allnbuck[i-1]; b++){
	double sf;
	sf = ((double)b/(double)allnbuck[i-1]) * (speedfrac - lastfrac) + lastfrac;
	k = KeyLinterp(sf, nsamp, allsamples, bkey0, bkey7);
	Msgf(("Bkt interpolate(%g) = %s\n", sf, PrintKey(k)));
	bucks[b].lo = k;
	bucks[b-1].hi = KeyAddInt(k, -1);
	Msgf(("proc=%d, buck[%d] = {%s, ", i-1, b-1, PrintKey(bucks[b-1].lo)));
	Msgf(("%s}\n", PrintKey(bucks[b-1].hi)));
      }
      bucks[b-1].hi = decompranges[i-1].hi;
      Msgf(("proc=%d, buck[%d] = {%s, ", i-1, b-1, PrintKey(bucks[b-1].lo)));
      Msgf(("%s}\n", PrintKey(bucks[b-1].hi)));
      MPMY_send(bucks, sizeof(krange_t)*allnbuck[i-1], i-1, 5432);
    }
    lastfrac = speedfrac;
    speedfrac = nsamp;
    bucks[0].lo = decompranges[i-1].lo;
    for(b=1; b<allnbuck[i-1]; b++){
      double sf;
      sf = ((double)b/(double)allnbuck[i-1]) * (speedfrac - lastfrac) + lastfrac;
      k = KeyLinterp(sf, nsamp, allsamples, bkey0, bkey7);
      Msgf(("interpolate(%g) = %s\n", sf, PrintKey(k)));
      bucks[b].lo = k;
      bucks[b-1].hi = KeyAddInt(k, -1);
      Msgf(("proc=%d, buck[%d] = {%s, ", i-1, b-1, PrintKey(bucks[b-1].lo)));
      Msgf(("%s}\n", PrintKey(bucks[b-1].hi)));
    }
    bucks[b-1].hi = decompranges[i-1].hi;
    Msgf(("proc=%d, buck[%d] = {%s, ", i-1, b-1, PrintKey(bucks[b-1].lo)));
    Msgf(("%s}\n", PrintKey(bucks[b-1].hi)));
    MPMY_send(bucks, sizeof(krange_t)*allnbuck[i-1], i-1, 5432);
  }

  /* Now sort out all the broadcasts and scatters */
  MPMY_Wait(req, &stat);
  MPMY_Bcast(decompranges, sizeof(krange_t)*MPMY_Nproc(), MPMY_CHAR, 0);

  if( amRoot ){
    Free(allsamples);
    Free(allspeeds);
    Free(allnsamp);
    Free(allnbuck);
  }
}
