/* We split the sub-tree between high key and a low key into 
   'buckets' such that each is (significantly) smaller than memory.
   These buckets will then be handed to an in-core sort for tree-building */

#include <stdlib.h>

#include "key.h"
#include "tree.h"
#include "traverse.h"
#include "decomp.h"
#include "Msgs.h"

static Key_t lokey, hikey;
static SplitBodsValues_t bodySzs;
static SplitCellValue_t  cellSz;
static int ndim;

static double MeasureSz(oocptr oocsg);

/* tstk represents a 'range' of keys, from tstk000 to tstk777.
   Return true if any part of that range overlaps with [lo, hi]. */
static int
overlapRange(Key_t lo, Key_t hi, Key_t tstk, int ndim){
  int tstlev = TreeLevel(tstk, ndim);
  int maxlev = (KEYBITS-1)/ndim;
  int lshift = ndim * (maxlev - tstlev);
  Key_t tsthi, tstlo;
    
  tstlo = KeyLshift(tstk, lshift);
  tsthi = KeyLshiftFill(tstk, lshift);

  return ! (KeyLT(tsthi, lo) || KeyGT(tstlo, hi));
}

/* Return true if tstk is entirely inside the [lo,hi] range */
static int
insideRange(Key_t lo, Key_t hi, Key_t tstk, int ndim){
  int tstlev = TreeLevel(tstk, ndim);
  int maxlev = (KEYBITS-1)/ndim;
  int lshift = ndim * (maxlev - tstlev);
  Key_t tsthi, tstlo;
    
  tstlo = KeyLshift(tstk, lshift);
  tsthi = KeyLshiftFill(tstk, lshift);

  return KeyGE(tstlo, lo) && KeyLE(tsthi, hi);
}

static double
restrictedCellSz(Key_t key, tcell *tc){
  Key_t ckey = KeyOrInt( KeyLshift(key, ndim), tc->bits );

  if( !overlapRange(lokey, hikey, ckey, ndim) ){
    Msgf((" %s is outside range, return 0!\n", PrintKey(ckey)));
    return 0.;
  }else if( insideRange(lokey, hikey, ckey, ndim) ){
    Msgf((" %s is inside range, return %g\n", PrintKey(ckey), cellSz(ckey, tc)));
    return cellSz(ckey, tc);
  }else{
    Msgf((" %s has edge effects, recursively try daughter\n", PrintKey(ckey)));
    return MeasureSz(tc->daughter);
  }
}

static void
restrictedBodyVisit(sib_group *sg, SplitTbodyInfo *tbinfo){
  int i;
  
  (*bodySzs)(sg, tbinfo);
  /* Now set any to zero that are outside the range */
  for(i=0; i<sg->ntbody; i++){
    if( KeyGT(tbinfo[i].key, hikey) || KeyLT(tbinfo[i].key, lokey) ){
      tbinfo[i].value = 0.;
    }
  }
}

static
double MeasureSz(oocptr oocsg){
  double sum = 0.;
  sib_group *sg;
  SplitTbodyInfo *tbinfo;
  PgHndl sghndl;
  int i;

  sg = OOCFollow(oocsg, &sghndl);

  for(i=0; i<sg->ncell; i++){
    sum += restrictedCellSz(sg->key, &sg->tcells[i]);
  }

  tbinfo = alloca(sg->ntbody * sizeof(*tbinfo));
  restrictedBodyVisit(sg, tbinfo);
  for(i=0; i<sg->ntbody; i++){
    sum += tbinfo[i].value;
  }

  PgUnref(sghndl, PG_CLEAN);
  return sum;
}

void
splitKeys(int nbuckets, krange_t *results, int ndim_, oocptr rootsg,
	  Key_t lokey_, Key_t hikey_,
	  SplitCellValue_t cellSz_, SplitBodsValues_t bodySzs_){
  double sztotal;
  Key_t lastkey;
  int i;

  ndim = ndim_;
  lokey = lokey_;
  hikey = hikey_;
  cellSz = cellSz_;
  bodySzs = bodySzs_;
  sztotal = MeasureSz(rootsg);
  Msgf(("Bucketize(%s, ", PrintKey(lokey)));
  Msgf(("%s)\n", PrintKey(hikey)));
  Msgf(("Total size=%g\n", sztotal));

  results[0].lo = lokey_;
  lastkey = lokey_;
  for(i=0; i<nbuckets-1; i++){
    if( sztotal > 0 ){
      Msgf(("Calling ksrch(%g)\n", (i+1)*sztotal/nbuckets));
      results[i].hi = 
	kSrch((i+1)*sztotal/nbuckets, rootsg, ndim,
	      restrictedCellSz, restrictedBodyVisit);
      Msgf(("ksrch returned %s\n", PrintKey(results[i].hi)));
    }else{
      results[i].hi = KeyInterpolate(lokey_, hikey_, (double)(i+1)/nbuckets);
    }
    results[i+1].lo = KeyAddInt(results[i].hi, 1);
  }
  results[nbuckets-1].hi = hikey_;
  PagePollTillDone();
}

