/* Copyright 1998, John K. Salmon All Rights Reserved.  */

#ifndef _Tree17DOTh
#define _Tree17DOTh

#include <stddef.h>
#include <limits.h>
#include "key.h"
#include "page.h"

#define MAXNDIM 3
#define MAXM 32

/* C's typing mechanism isn't good enough to let us use these names
   properly, so we cheat and use the pre-processor.  Sort of a
   poor-man's template mechanism.  If you are careful, you can #define
   these to types of your own before THE FIRST #include "tree.h" in
   any source module.   Then you can write your callbacks using 
   your own type names and not have to filter out unhelpful type
   mismatch warnings. */

#ifndef _cellDataIntermediate
#define _cellDataIntermediate void
#endif

#ifndef _cellDataFinal
#define _cellDataFinal void
#endif

#ifndef _tbody
#define _tbody void
#endif

/* Definitions, etc. not for casual users */
#include "tree-internal.h"

#ifdef __cplusplus
extern "C"{
#endif /* __cplusplus */

/* TCELL:  The basic data structure for an internal node in the tree.
   It always has a daughter (otherwise it would never have been created).
   The 'offset' is added to the initial address to get the actual "physics"
   contents.  This violates all known rules of good coding style,
   but it keeps things together in memory -- "it's the bandwidth STUPID".
   Use the Contents() macro to find the contents of a tcell. */
typedef struct tcell_s{
  unsigned int bits : MAXNDIM;
  unsigned int offset : (sizeof(unsigned int)*CHAR_BIT - MAXNDIM);
  oocptr daughter;
} tcell;
#define Contents(tcf) ((_cellDataFinal *)((char *)(tcf) + (tcf)->offset))

/* SIB_GROUP: a group of siblings.  The plan is that we ALWAYS store
   siblings in contiguous memory locations.  The 'daughter' entry in a
   tcell points to one of these.  A sib_group consists of up to 1<<dim
   tcells AND up to m*(1<<ndim) tbodies.  I.e., all the sub-cells plus
   all the 'un-attached' tbodies that lie within a cell.  Because of
   the bizarre packing strategy, use the macro Tbodies() to find the
   tbodies in a cell.  */
typedef struct sib_group_s{
  Key_t key;
  unsigned char ncell;
  unsigned char ntbody;
  short tbody_offset;
  tcell tcells[DUMMY_ARRAY_LENGTH];
  /* followed by:
     ntbody tbodies
     ncell cells;
     
     But the sizes of these are not known explicitly, or even necessarily
     fixed so it would only lead to confusion to include them in the struct...
     */
} sib_group;
#define Tbodies(sgf)  ((_tbody *)((char *)(sgf) + sgf->tbody_offset))

/* MakeIntermediateCell computes the aggregation for a bunch of
   _tbodies and other _cellDataIntermediates into a new
   _cellDataIntermediate.  (I.e., a cellDataIntermediate might
   consist of centers-of-mass info, higher order moments, etc.)

   All _cellDataIntermediates are the same size (specified in the
   BuildFuncs_t structure).  Since there are never very many of them
   in storage at any one time (O(lgN)), it's not a problem for the
   user to just declare a union if there are different possible types.

   The value returned is the EVENTUAL SIZE OF THE RESULTING _cellDataFinal.
   The tree library records the returned value and reserves that much
   space for the 'to' argument of makeFinalCell (see below).  In most
   cases, this will just be sizeof(some_application_structure), but
   advanced applications will may require different amounts of space
   for different cells (e.g., 'p-adaptivity').

   As discussed elsewhere, the point of this two-step process is to
   allow for shrinkage between the intermediate stage (e.g., higher
   order moments) and the permanent cell data (e.g., rcrit).  */
typedef size_t 
(*makeIntermediateCell_t)(Key_t key, int ntbody, const _tbody **tbodyptrs,
			  int ncell, const _cellDataIntermediate **aggregates, 
			  _cellDataIntermediate *_to);

/* makeFinalCell convert a _cellDataIntermediate into a _cell, i.e.,
     the permanent representation of aggregate objects in the tree.
     While it's reasonable for _cellDataIntermediates to be slightly
     wasteful of space, _cells could easily dominate the space
     requirement of a program.  They are packed tightly into pages of
     memory, and they be as dense as possible.  The user is encouraged
     to eliminate any unnecessary 'baggage' during the repackaging
     process.  When makeIntermediateCell was called the implementation
     told treelib how much space would be required for the final cell.
     That value is now passed to makeFinalCell.  It is an error to
     write any more data.  The function signature of makeFinalCell is
     such that memcpy is a valid makeFinalCell. */
typedef void * 
(*makeFinalCell_t)(_cellDataFinal *to, const _cellDataIntermediate *from, size_t sz);

/* BuildFuncs is a container for the callbacks typedef'ed above.  You
 have to create one and initialize it before calling TreeBuildStart. */
typedef struct BuildFuncs_s{
  makeIntermediateCell_t makeIntermediateCell;
  makeFinalCell_t makeFinalCell;
  size_t intermediatesz;
} BuildFuncs_t;

void TreeBuildStart(tree_t *tp, int ndim, int m, BuildFuncs_t *Bfuncs);
void TreeBuildAdd(tree_t *tp, Key_t key, _tbody *tbody, size_t tbodysz);
oocptr TreeBuildFinish(tree_t *tp);
void TreeFree(tree_t *tp);
oocptr TreeRoot(tree_t *tp);
void TreeChkSibGroup(const sib_group *sg, const Key_t *bkeys, int ndim, int m);

/* some utilities for printing the contents of various internals */
char *PrintSg(const sib_group *sg);
char *PrintTc(const tcell *tc);

/* This one is for debugging only.  If DEBUG is set in tree17.c,
   then you can set this to something that will return the Key of a
   tbody *.  Lots more internal checking is then possible. */
extern Key_t (*tree17keyFromTbody)(const _tbody *);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* _TreeDOTh */

