#include <stdarg.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
/* Disregard warning in stdio.h.  It's because unistd.h #defines SEEK_CUR */
/* etc., and it's not the same as in stdio...  "Intel inside (tm)" */
#include <stdio.h>
#include <errno.h>
#include "protos.h"
#include "error.h"
#include "mpmy_io.h"
#include "Msgs.h"
#include "mpmy.h"
#include "Malloc.h"
#include "iozero.h"

#ifndef EINVAL
#define EINVAL 0
#endif

/* This will probably not work for multiple processes writing to NFS */
/* There needs to be a mpmy_nfsio.c which does what #define SERIALIZE_WRITES */
/* did in the original SDFwrite */

static int
MPMY_Global_Write(int fd, const void *buf, unsigned int nbytes);
static int
MPMY_Global_Read(int fd, void *buf, unsigned int nbytes);

#define NFILES 64
static struct _File{
    int fd;
    int iomode;
    char ungot;
    int ungot_active;
} _files[NFILES];

#define DEFAULT_PERMS 0644

MPMYFile *
MPMY_Fopen(const char *path, int mpmy_flags)
{
    int fd;
    int flags = 0;
    int iomode = MPMY_UNIX;

    Msgf(("Fopen %s\n", path));
    if (mpmy_flags & MPMY_RDONLY) flags |= O_RDONLY;
    if (mpmy_flags & MPMY_WRONLY) flags |= O_WRONLY;
    if (mpmy_flags & MPMY_RDWR)   flags |= O_RDWR;
    if (mpmy_flags & MPMY_APPEND) flags |= O_APPEND;
    if (mpmy_flags & MPMY_TRUNC)  flags |= O_TRUNC;
    if (mpmy_flags & MPMY_CREAT)  flags |= O_CREAT;

    /* We should really check that only one is set... */
    if (mpmy_flags & MPMY_MULTI) iomode = MPMY_MULTI;
    if (mpmy_flags & MPMY_SINGL) iomode = MPMY_SINGL;
    if (mpmy_flags & MPMY_UNIX) iomode = MPMY_UNIX;
    if (mpmy_flags & MPMY_IOZERO) iomode = MPMY_IOZERO;
    if (mpmy_flags & MPMY_INDEPENDENT) iomode = MPMY_INDEPENDENT;

    /* One must make sure that only one process creates or truncates a file */
    if ((flags & (O_CREAT|O_TRUNC)) && (iomode != MPMY_INDEPENDENT) ) {
	if (MPMY_Procnum() == 0) {
	    fd = open(path, flags, DEFAULT_PERMS);
	    MPMY_Sync();
	} else {
	    flags &= ~(O_CREAT|O_TRUNC);
	    MPMY_Sync();
	    fd = open(path, flags, DEFAULT_PERMS);
	}
    } else {
	fd = open(path, flags, DEFAULT_PERMS);
    }

    if(fd >= 0){
	_files[fd].iomode = iomode;
	_files[fd].ungot_active = 0;
	_files[fd].fd = fd;
	Msgf(("Fopen returns %d, iomode=%d, flags=0x%x\n", fd, iomode, flags));
	return &(_files[fd]);
    }else{
	Msgf(("Fopen fails, errno=%d\n", errno));
	return NULL;
    }
}

int
MPMY_Fclose(MPMYFile *Fp)
{
    struct _File *fp = (struct _File *)Fp;
    int ret;

    if( fp == NULL ){
	errno = EINVAL;
	return -1;
    }

    ret = close(fp->fd);
    fp->fd = -1;
    Msgf(("Close returns %d\n", ret));
    return ret;
}

int 
MPMY_Mkdir(const char *path, int mode){
    return mkdir0(path, mode);
}

int
MPMY_Fread(void *ptr, int size, int nitems, MPMYFile *Fp)
{
    struct _File *fp = (struct _File *)Fp;
    int ret = 0;

    if( fp == NULL ){
	errno = EINVAL;
	return -1;
    }

    switch (fp->iomode) {
    case MPMY_IOZERO:
	ret = read0(fp->fd, ptr, size*nitems);
	break;
    case MPMY_INDEPENDENT:
    case MPMY_UNIX:
	ret = read(fp->fd, ptr, size*nitems);
	break;
    case MPMY_MULTI:
	ret = MPMY_Global_Read(fp->fd, ptr, size*nitems);
	break;
    case MPMY_SINGL:
	if (MPMY_Procnum() == 0) {
	    ret = read(fp->fd, ptr, size*nitems);
	}
	MPMY_Bcast(&ret, 1, MPMY_INT, 0);
	if(ret > 0 )
	    MPMY_Bcast(ptr, ret, MPMY_CHAR, 0);
	break;
    default:
	ret =  -1;
	break;
    }

    Msgf(("Fread returns %d.\n", ret));
    if (ret % size) 
	SeriousWarning("MPMY_Fread has a problem, ret=%d, size=%d\n",
	      ret, size);
    return ret/size;
}

int
MPMY_Fwrite(void *ptr, int size, int nitems, MPMYFile *Fp)
{
    struct _File *fp = (struct _File *)Fp;
    int ret;

    if( fp == NULL ){
	errno = EINVAL;
	return -1;
    }
    
    switch (fp->iomode) {
    case MPMY_IOZERO:
	ret = write0(fp->fd, ptr, size*nitems);
	break;
    case MPMY_INDEPENDENT:
    case MPMY_UNIX:
	ret = write(fp->fd, ptr, size*nitems);
	break;
    case MPMY_MULTI:
	ret = MPMY_Global_Write(fp->fd, ptr, size*nitems);
	break;
    case MPMY_SINGL:
	if (MPMY_Procnum() == 0) {
	    ret = write(fp->fd, ptr, size*nitems);
	} else ret = size*nitems; /* no check for identical data */
	break;
    default:
	ret = -1;
	Shout("Bad iomode in Fread\n");
    }

    Msgf(("Fwrite returns %d.\n", ret));
    if (ret % size) 
	Shout("MPMY_Fwrite has a problem: ret=%d, size=%d\n", ret, size);
    return ret/size;
}

int
MPMY_Fseek(MPMYFile *Fp, long offset, int whence)
{
    struct _File *fp = (struct _File *)Fp;
    int ret;
    int real_whence = 0;
    int iomode;

    if( fp == NULL ){
	errno = EINVAL;
	return -1;
    }

    iomode = fp->iomode;
    if (iomode == MPMY_MULTI)
	Error("Illegal iomode for Fseek\n");
    if (iomode == MPMY_SINGL && MPMY_Procnum()) {
	return 0;
    }
    fp->ungot_active = 0; /* read the manual for ungetc */
    if (whence == MPMY_SEEK_SET) real_whence = SEEK_SET;
    if (whence == MPMY_SEEK_CUR) real_whence = SEEK_CUR;
    if (whence == MPMY_SEEK_END) real_whence = SEEK_END;

    if(iomode == MPMY_IOZERO)
	ret = lseek0(fp->fd, offset, real_whence);
    else
	ret = lseek(fp->fd, offset, real_whence);
    if (ret != -1) ret = 0;
    Msgf(("Fseek returns %d\n", ret));
    return ret;
}

int
MPMY_Ftell(MPMYFile *Fp)
{
    struct _File *fp = (struct _File *)Fp;
    int ret;
    int iomode;

    if( fp == NULL ){
	errno = EINVAL;
	return -1;
    }

    iomode = fp->iomode;
    if (iomode == MPMY_MULTI)
	Error("Illegal iomode for Ftell\n");

    if (iomode == MPMY_IOZERO)
	return tell0(fp->fd);

    if (MPMY_Procnum() == 0) {
	ret = lseek(fp->fd, 0, SEEK_CUR);
    }
    MPMY_Bcast(&ret, 1, MPMY_INT, 0);

    Msgf(("Ftell returns %d\n", ret));
    return ret;
}

int
MPMY_Flen(MPMYFile *Fp)
{
    struct _File *fp = (struct _File *)Fp;
    struct stat buf;
    int ret;

    if( fp == NULL ){
	errno = EINVAL;
	return -1;
    }

    if( fp->iomode == MPMY_IOZERO )
	return flen0(fp->fd);

    if (MPMY_Procnum() == 0) {
	/* We get a warning on delta because mpmy_nx.c includes protos.h */
	/* but not sys/stat.h, hence it doesn't prototype fstat.  Sigh. */
	fstat(fp->fd, &buf);
	ret = buf.st_size;
    }
    MPMY_Bcast(&ret, 1, MPMY_INT, 0);
    Msgf(("Flen returns %d\n", ret));
    return ret;
}

int
MPMY_Getc(MPMYFile *Fp)
{
    struct _File *fp = (struct _File *)Fp;
    int ret;
    char buf;

    if( fp == NULL ){
	errno = EINVAL;
	return -1;
    }

    if (fp->ungot_active) {
	fp->ungot_active = 0;
	return fp->ungot;
    } else {
	ret = MPMY_Fread(&buf, 1, 1, fp);
	if (ret == 0 ) return 0;
	if (ret == -1) return -1;
	return buf;
    }
}

int
MPMY_Ungetc(char c, MPMYFile *Fp)
{
    struct _File *fp = (struct _File *)Fp;
    if( fp == NULL ){
	errno = EINVAL;
	return -1;
    }

    if (c == -1) return -1;
    fp->ungot_active = 1;
    fp->ungot = c;
    return c;			/* is this right? */
}

int 
MPMY_Fseekrd(MPMYFile *Fp, long offset, int whence, void *buf, int reclen,
	     int nrecs)
{
    struct _File *fp = (struct _File *)Fp;
    int doseek;

    if( fp == NULL ){
	errno = EINVAL;
	return -1;
    }

    if( fp->iomode == MPMY_IOZERO ){
	return fseekrd0(fp->fd, offset, whence, buf, reclen, nrecs);
    }

    if( whence == MPMY_SEEK_CUR ){
	doseek = (offset != 0);
    }else if( whence == MPMY_SEEK_SET ){
	/* don't worry about errors.  If ftell returns -1, */
	/* doseek will be turned on, and the fseek below will */
	/* (probably) fail */
	doseek = (MPMY_Ftell(fp) != offset);
    }else{
	doseek = 1;
    }

    if( doseek ){
	if( MPMY_Fseek(fp, offset, whence) ){
	    if( whence == MPMY_SEEK_CUR && offset > 0 ){
		/* Make a final heroic effort to seek by reading forward! */
		char junk[BUFSIZ];
		int nleft = offset;
		while( nleft ){
		    int ntry = ( nleft > sizeof(junk) ) ? sizeof(junk) : nleft;
		    if( MPMY_Fread(junk, ntry, 1, fp) != 1 ){
			Error("fseekrd: incremental fread(%#lx, %d, 1, %#lx) failed, errno=%d\n",
			      (unsigned long)junk, ntry, 
			      (unsigned long)fp, errno);
			return -1;
		    }
		    nleft -= ntry;
		}
	    }else{
		Error("fseekrd: fseek(%#lx, %ld, %d) failed, errno=%d\n",
		      (unsigned long)fp, offset, whence, errno);
		return -1;
	    }
	}
    }
    if( MPMY_Fread(buf, reclen, nrecs, fp) != nrecs ){
	Error("fseekrd: fread(%#lx, %d, %d, %#lx) failed, errno=%d\n", 
	      (unsigned long)buf, reclen, nrecs, (unsigned long)fp, errno);
	return -1;
    }
    return nrecs;
}

static int
MPMY_Global_Write(int fd, const void *buf, unsigned int nbytes)
{
    int ret;
    unsigned long seeklen;
    int *sizes;
    int i;
    int nproc = MPMY_Nproc();
    int procnum = MPMY_Procnum();

    Msgf(("Global_Write %d...", nbytes));
    
    seeklen = tell0(fd);
    /* This is really 'scan', which could be implemented slightly faster... */
    sizes = Malloc(sizeof(int)*nproc);
    MPMY_AllGather(&nbytes, 1, MPMY_INT, sizes);
    for (i = 0; i < procnum; i++)
	seeklen += sizes[i];

    Msgf(("seeklen=%lu\n", seeklen));
    if (lseek(fd, seeklen, SEEK_SET) != seeklen)
	Error("lseek(seeklen=%lu) failed, errno=%d\n", seeklen, errno);

    ret = write(fd, buf, nbytes);

    /* Now figure out where the end is... */
    for( ; i<nproc; i++){
	seeklen += sizes[i];
    }
    if( lseek0(fd, seeklen, SEEK_SET) != seeklen ){
	SinglWarning("lseek0(seeklen=%lu) failed in Global_Write\n", seeklen);
	ret = -1;
    }
    /* shouldn't we do a combine on our return values too?? */

    Free(sizes);
    Msgf(("Global_Write returns %d.\n", ret));
    return ret;
}

static int
MPMY_Global_Read(int fd, void *buf, unsigned int nbytes)
{
    int ret;
    unsigned long seeklen;
    int *sizes;
    int i;
    int nproc = MPMY_Nproc();
    int procnum = MPMY_Procnum();

    Msgf(("Global_Read %d...", nbytes));

    seeklen = tell0(fd);
    /* This is really 'scan', which could be implemented slightly faster... */
    sizes = Malloc(sizeof(int)*nproc);
    MPMY_AllGather(&nbytes, 1, MPMY_INT, sizes);
    for (i = 0; i < procnum; i++)
	seeklen += sizes[i];

    Msgf(("seeklen=%lu\n", seeklen));
    if (lseek(fd, seeklen, SEEK_SET) != seeklen)
	Warning("lseek(seeklen=%lu) failed, errno=%d\n", seeklen, errno);

    ret = read(fd, buf, nbytes);

    /* Now figure out where the end is... */
    for( ; i<nproc; i++){
	seeklen += sizes[i];
    }
    if( lseek0(fd, seeklen, SEEK_SET) != seeklen ){
	SinglWarning("lseek0(seeklen=%lu) failed in Global_Read\n", seeklen);
	ret = -1;
    }
    /* shouldn't we do a combine on our return values too?? */

    Free(sizes);
    Msgf(("Global_Read returns %d.\n", ret));
    return ret;
}

#include "iozero.c"
#include "io_generic.c"
