#define page5DotC

#ifdef __OPTIMIZE__
#define NO_MSGS
#endif

/* The msgs in Remote can be expensive.  They are also the last best hope
   for tracking down some bugs */
/* #define NO_MSGS */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#include "protos.h"
#include "Assert.h"
#include "Malloc.h"
#include "Msgs.h"
#include "singlio.h"
#include "page.h"
#include "page-internal.h"
#include "verify.h"
#include "timers.h"
#include "pagehash.h"
#include "mpmy.h"

#ifndef INLINE
#ifdef __GNUC__
#define INLINE __inline__
#else
#define INLINE
#endif
#endif

/* Is this a sun-ism, a bsd-ism or what??  Should we add a similar
 * feature to libsw/malloc.c? 
 */
#ifndef __GLIBC__ /* already in stdlib.h */
extern void *valloc(int);
#endif

static inmeminfo *in_mem;

static int lsv_tag_bug;
static int OnDiskHwm;		/* largest page written to disk */
static int WSETPAGES;
static int fd;
static char tmpfilename[512];
static hash_t *ht;
static hash_t htbl;

int PgTimestamp;
int PgSize;

/* Some accumulated statistics */
int PgNrefcnt;			/* how many are refcnt right now */
int PgReuse;			/* how many locks were in-mem */

/* The actual in-memory working-set. */
void *_PgWset;
int _LGPGSZ;

Timer_t PgInTm, PgOutTm;
Timer_t PgPollTm, PgPollTillDoneTm;
Timer_t PgPollLoopTm;
Timer_t RemoteTm;
Counter_t PgInCnt, PgOutCnt;
Counter_t PgRemoteInCnt, PgRemoteSendCnt;
Counter_t PgPollLoopCnt;

/* How can we really be sure that the OS has delivered the goods on a read?
   We can do a checksum of the first word on every page.  At this point, 
   we know that the data is in OUR memory.  But we don't know where the OS
   got it from.  Perhaps it had it secreted away in a buffer somewhere?
   Bottom line:  we still can't trust that the elapsed time includes a
   disk read.
   */
#define DISKBLOCKSIZE 4096
#ifdef FORCE_IO 
static int
ForceRead(const void *vp, int sz){
  int sum = 0;
  const int *p = vp;
  const int *end = (const int *)((const char *)vp + sz);
  while( p < end ){
    sum ^= *p;
    p += DISKBLOCKSIZE;
  }
  return sum;
}
#endif

static void RemoteRead(PgId pgid, void *addr);
static void RemoteInit(void);
static void PgWriteBack(inmeminfo *imi);
static void PgReadStorage(inmeminfo *imi);
static void RemoteTerm(void);

char *PrintPgId(PgId pgid){
  static char foo[128];
  sprintf(foo, "(pgnum=%d, proc=%d)", pgid.pgnum, pgid.proc);
  return foo;
}

void PgInit(int lgpgsz, int wsetpages, const char *filename){
  int i;

  WSETPAGES = wsetpages;
  _LGPGSZ = lgpgsz;
  PgTimestamp = 0;
#if 0
  lsv_tag_bug = (strstr(MPMY_Transport(), "lsv") != NULL) 
    || (strstr(MPMY_Transport(), "swampi") != NULL);
#else
  lsv_tag_bug = 0;
#endif
  Msgf(("lsv_tag_bug=%d\n", lsv_tag_bug));

  hash_make(&htbl, 3*wsetpages);
  ht = &htbl;

  in_mem = Calloc(WSETPAGES, sizeof(inmeminfo));
  if( filename && strlen(filename) > 0 ){
    strncpy(tmpfilename, filename, sizeof(tmpfilename));
  }else{
    tmpnam(tmpfilename);
  }
  /* Use O_SYNC because when we ask to write a block, we really mean
     to write it NOW.  If we had space for it, we wouldn't have written
     it in the first place.  The Linux manual says that O_SYNC isn't
     supported anyway...*/
  Verify( (fd = open(filename, O_CREAT|O_RDWR|O_SYNC, 0666)) >= 0 );
  Msgf(("Opening temp file: %s\n", tmpfilename));
  /* unlink(tmpfilename);	 */	/* fire and forget  */

  VerifyX( _PgWset = valloc(WSETPAGES <<_LGPGSZ), Msg_do("valloc(%d << %d) returns NULL\n", WSETPAGES, _LGPGSZ) ); 
  /* Map the first WSETPAGES into memory.  */
  for(i=0; i<WSETPAGES; i++){
    inmeminfo *imi;
    hash_entry_t *p;
    imi = &in_mem[i];
    imi->id.pgnum = i;
    imi->id.proc = -1;
    imi->expired = 1;
    imi->dirty = 0;
    imi->refcnt = 0;
    imi->base = _PgWset + (i<<_LGPGSZ);
    p = hash_get(ht, &imi->id);
    assert(p && (p->data == NULL));
    p->data = imi;
    hash_release(ht, p);
  }
  PgSize = 1<<_LGPGSZ;
  OnDiskHwm = -1;
  RemoteInit();
  EnableWCTimer(&PgInTm, "Pagein(sec)");
  EnableWCTimer(&PgOutTm, "Pageout(sec)");
}

void PgTerm(void){
  Free(in_mem);
  free(_PgWset);		/* obtained from valloc, not Malloc! */
  hash_destroy(ht);
  in_mem = NULL;
  _PgWset = NULL;
  RemoteTerm();
}

/* These two are called while holding an exclusive lock 
   on the entire hash table.  They attempt to find the 'oldest', i.e.,
   the most replaceable page in memory. */

static INLINE int Age(inmeminfo *p){
  int age = PgTimestamp - p->timestamp;
  /* Pages that are dirty, i.e., those that need to be written to
     disk, automatically get their age halved, so we are disinclined
     to swap them out. */
  if( p->dirty )
    age >>= 1;
  /* Make remote pages four times less likely to be dumped.  This may
     be too much.  */
  if( p->id.proc != MPMY_Procnum() )
    age >>= 2;
  return age;
}

static hash_entry_t *
hash_find_replaceable(hash_t *tbl){
  /* How long does this take?  It loops WSETPAGES times, which might
     be a few hundred.  There's not too much inside the loop, so
     figure 100 clocks per cycle?  At 10000 cycles it's WAY, WAY less
     than the disk read we are about to execute. */
  hash_entry_t *he;
  int age_oldest;
  hash_entry_t *oldest;
  hash_iterator_t hiter;

  age_oldest = -1;
  oldest = NULL;
  hash_iter_start(&hiter, ht, 0);
  
  while( (he = hash_iter_next(&hiter)) ){
    inmeminfo *imi;
    int age_new;

    imi = he->data;
    /* NULL means it hasn't been set to point to anything yet, i.e.,
       we are probably 'holding a lock' on it with hash_get.  
       A non-zero refcnt means somebody has live pointers to the data */
    if( imi == NULL || imi->refcnt )
      continue;

    if( imi->expired ){		
      oldest = he;		/* We have a winner. */
      break;
    }

    age_new = Age(imi);
    if(age_oldest < age_new){
      age_oldest = age_new;
      oldest = he;
    }
  }

  hash_iter_finish(&hiter);
  Msgf(("hash_find_replaceable returning %p pgid=%s\n", oldest, PrintPgId(oldest->id)));
  return oldest;
}

PgHndl PgRef(PgId pgid){
  inmeminfo *imi;
  hash_entry_t *hp;
  
  hp = hash_get(ht, &pgid);
  if( hp->data == NULL ){
    /* The page is not in memory.  We have to get it from disk. */
    int outpg;
    hash_entry_t *replace;
    int wsetindex;

    Msgf(("Page (%d@%d) is not in memory\n", pgid.pgnum, pgid.proc));

    /* Find a page to replace */
    replace = hash_find_replaceable(ht);
    assert( replace != NULL && replace->data != NULL );
    hp->data = replace->data;
    hash_delete(ht, replace);
    /* Ok.  We've got a replacement page. */

    imi = (inmeminfo *)hp->data;
    wsetindex = imi - in_mem;
    outpg = imi->id.pgnum;
    Msgf(("Replacing wsetind: %d (pg=%d@%d, refcnt=%d, dirty=%d, hits=%d, timestamp=%d)\n", 
	  wsetindex, imi->id.pgnum, imi->id.proc, imi->refcnt, imi->dirty, imi->hits, imi->timestamp));
    /* We may start to do I/O now.  Notice that we are holding a lock
       on hp, which means nobody else should be messing with our page
       in memory or on disk -- right?  We are NOT holding any other
       locks, e.g., the global lock on the hash table itself, so this
       thread is free to block waiting for I/O, and other
       threads will be able to proceed and get something done.  */
    if( imi->dirty ){
      Msgf(("%d is dirty, swapping it out\n", outpg));
      PgWriteBack(imi);
    }

    /* read from disk (local or remote) */
    imi->id = pgid;
    PgReadStorage(imi);
  }else{
    /* It's in memory, but it might be expired... */
    imi = (inmeminfo *)hp->data;
    assert(imi->id.pgnum == pgid.pgnum && imi->id.proc == pgid.proc);
    if( imi->expired )
      PgReadStorage(imi);
    else
      PgReuse++;
  }
  assert(imi->refcnt >= 0);
  Msgf(("Locking wseti=%d (proc=%d, pgnum=%d, refcnt=%d, dirty=%d, hits=%d, ts=%d)\n",
	imi - in_mem, imi->id.proc, imi->id.pgnum, imi->refcnt, imi->dirty, imi->hits, imi->timestamp));
  if( imi->refcnt++ == 0 )
    PgNrefcnt++;
  imi->hits++;
  hash_release(ht, hp);
  PgTimestamp++;
  return imi;
}

void PgUnref(PgHndl hndl, enum PgHygiene_e dirtied){
  inmeminfo *imi = (inmeminfo *)hndl;
  Msgf(("Unlocking wseti=%d (proc=%d, pgnum=%d, refcnt=%d, dirty=%d, hits=%d, ts=%d)\n", 
	imi - in_mem, imi->id.proc, imi->id.pgnum, imi->refcnt, imi->dirty, imi->hits, imi->timestamp));
  imi->dirty |= dirtied;
  assert(imi->refcnt > 0);
  if( --imi->refcnt == 0 )
    PgNrefcnt--;
}

void PgReref(PgHndl hndl){
  inmeminfo *imi = (inmeminfo *)hndl;

  assert(imi->refcnt > 0);
  imi->refcnt++;
}

/* It is assumed that the caller is holding a lock on imi -- or that
   we aren't doing any of the threading stuff. */
static void PgWriteBack(inmeminfo *imi){
  int outpg = imi->id.pgnum;
  int wsetindex = imi - in_mem;

  assert(imi->id.proc == MPMY_Procnum());
  StartTimer(&PgOutTm);
  Verify( lseek(fd, outpg<<_LGPGSZ, SEEK_SET) >= 0 );
  Verify( write(fd, _PgWset + (wsetindex<<_LGPGSZ), 1<<_LGPGSZ) == 1<<_LGPGSZ );
  imi->dirty = 0;
  if( outpg > OnDiskHwm )
    OnDiskHwm = outpg;
  /*  This FORCES the data onto disk */
#ifdef FORCE_IO
  Verify0( fsync(fd) );
#endif
  StopTimer(&PgOutTm);
  IncrCounter(&PgOutCnt);
}

static void
PgReadStorage(inmeminfo *imi){
  PgId pgid = imi->id;
  int wsetindex = imi - in_mem;

  assert(wsetindex >= 0 && wsetindex < WSETPAGES);
  if( pgid.proc == MPMY_Procnum() ){
    int pgnum = pgid.pgnum;
    /* This test is goofy.  We really don't want to read in 'garbage'
       pages, but we don't know which ones they are because we are not
       keeping information about all pages on disk.  That's just an
       optimization, though.  What we must not do is try to read a
       block that's past the end of the disk.  Thus the test here. */
    if( pgnum <= OnDiskHwm ){
      Msgf(("Restoring page %d from disk to wset[%d]\n", pgnum, wsetindex));
      errno = 0;
      StartTimer(&PgInTm);
      Verify( lseek(fd, pgnum<<_LGPGSZ, SEEK_SET) >= 0 );
      Verify( read(fd, _PgWset + (wsetindex<<_LGPGSZ), 1<<_LGPGSZ) == (1<<_LGPGSZ) );
#ifdef FORCE_IO
      ForceRead(_PgWset + wsetindex<<_LGPGSZ, 1<<_LGPGSZ);
#endif
      StopTimer(&PgInTm);
      IncrCounter(&PgInCnt);
    }else{
      Msgf(("Page %d not on disk.  No read\n", pgnum));
    }
  }else{
    /* This will recursively call PgRef.  Are we really reentrant?? 
       Setting the refcnt to 1 is an attempt to avoid a race on
       the page in memory */
    imi->refcnt = 1;
    RemoteRead(pgid, _PgWset + (wsetindex<<_LGPGSZ));
  }
  imi->timestamp = PgTimestamp;
  imi->dirty = 0;
  imi->expired = 0;
  imi->refcnt = 0;
  imi->hits = 0;
}

static void PgFlushHndl(inmeminfo *imi){
  assert( imi->refcnt == 0 );
  if( imi->dirty )
    PgWriteBack(imi);
  Msgf(("PgFlushHndl(proc=%d, pgnum=%d)\n", imi->id.proc, imi->id.pgnum));
  imi->expired = 1;		/* it will be reclaimed the next time we call find_replaceable */
}

/* This is too sweeping.  It will work ok as long as only one 'client'
   is using the paging system.  It also needs some multi-threading work */
void PgFlushRemote(void){
  int i;
  for(i=0; i<WSETPAGES; i++){
    inmeminfo *imi = &in_mem[i];
    if( imi->id.proc>=0 && imi->id.proc != MPMY_Procnum() ){
        PgFlushHndl(imi);
    }
  }
}

int PgStatus(PgId pgid){
  int ret = 0;
  hash_entry_t *hp;
  
  if( pgid.proc != MPMY_Procnum() ) ret |= PG_STAT_REMOTE;
  
  hp = hash_get(ht, &pgid);
  if( (hp->data) == NULL ){
    /* Nothing more to check.  */
  }else{
    inmeminfo *imi = (inmeminfo *)(hp->data);
    ret |= PG_STAT_INMEM;
    if( imi->dirty ) ret |= PG_STAT_DIRTY;
    if( imi->refcnt ) ret |= PG_STAT_REFED;
  }
  hash_release(ht, hp);
  return ret;
}

void
PgStats(){
  Counter_t foo;

  OutputOneCounter(&PgInCnt, singlPrintf);
  OutputOneCounter(&PgOutCnt, singlPrintf);
  OutputTimer(&PgInTm, singlPrintf);
  OutputTimer(&PgOutTm, singlPrintf);

  EnableCounter(&foo, "PgHwm(kB)");
  AddCounter(&foo, (int)((PgHwm<<_LGPGSZ)*1.e-3));
  OutputOneCounter(&foo, singlPrintf);
  DisableCounter(&foo);

  EnableCounter(&foo, "PgHwm");
  AddCounter(&foo, PgHwm);
  OutputOneCounter(&foo, singlPrintf);
  DisableCounter(&foo);

  EnableCounter(&foo, "PgInuse");
  AddCounter(&foo, PgInUse);
  OutputOneCounter(&foo, singlPrintf);
  DisableCounter(&foo);

  EnableCounter(&foo, "PgReuse");
  AddCounter(&foo, PgReuse);
  OutputOneCounter(&foo, singlPrintf);
  DisableCounter(&foo);

  EnableCounter(&foo, "PgNref");
  AddCounter(&foo, PgNrefcnt);
  OutputOneCounter(&foo, singlPrintf);
  DisableCounter(&foo);
}


/* This could all go in a separate file... */
#include "abm.h"

static ABM remoteabm;
struct remote_req{
  PgId id;
  int replytag;
} ;

/* sufficiently random? */
#define REQUEST_TAG 6754
#define REPLY_TAG 6755
#define REMOTE_ABM_TAG 43314
static int remote_abm_tag = REMOTE_ABM_TAG;

static struct remote_req incoming_req;
static MPMY_Comm_request inreq;

static void RemoteInit(void){
  MPMY_Irecv(&incoming_req, sizeof(incoming_req), MPMY_SOURCE_ANY, REQUEST_TAG, &inreq);
  /* Create an ABM solely for the purpose of using it's two-phase shutdown
     code.  Did somebody say overkill??  Shouldn't we be using ABM for
     the data transfer too?  But we don't want the buffering behavior. */
  ABMSetup(&remoteabm, 8, remote_abm_tag, 0, NULL);
}

void PagePoll(void){
  int flag;
  MPMY_Status stat;

  StartTimer(&PgPollTm);
  /* Sigh.  Lsv is broken, and can't receive messages out-of-order, so
     if there's an ABMdone message sitting in our lsv queues, we have to
     clear it before we can get to the interesting stuff.  Yuck! */
  if( lsv_tag_bug )
    ABMPoll(&remoteabm);
  /* There were rumors that the comma operator inside a while() condition
     was broken on T3D.  It's just as easy to split it up, */
  MPMY_Test(inreq, &flag, &stat);
  while( flag ){
    int src;
    PgId id;
    int tag;
    MPMY_Comm_request sndreq;
    PgHndl hndl;

    src = stat.MPMY_SOURCE;
    Msg("Poll", ("RPoll received request for pg %d from %d\n", incoming_req.id.pgnum, src));
    assert(stat.MPMY_TAG == REQUEST_TAG);
    assert(stat.MPMY_COUNT == sizeof(incoming_req));
    id = incoming_req.id;
    assert(id.proc == MPMY_Procnum());
    tag = incoming_req.replytag;
    /* Get an early start on the Iread for the next request... */
    MPMY_Irecv(&incoming_req, sizeof(incoming_req), MPMY_SOURCE_ANY, REQUEST_TAG, &inreq);

    hndl = PgRef(id);
    MPMY_Isend(PgAddr(hndl), 1<<_LGPGSZ, src, tag, &sndreq);
    MPMY_Wait(sndreq, NULL);
    Msg("Poll", ("Completed request for pg %d\n", id.pgnum));
    PgUnref(hndl, PG_CLEAN);
    IncrCounter(&PgRemoteSendCnt);
    MPMY_Test(inreq, &flag, &stat);
  }
  StopTimer(&PgPollTm);
}

/* Read a remote page into the given address */
static void RemoteRead(PgId pgid, void *addr){
  MPMY_Comm_request rdreq, sndreq;
  MPMY_Status rdstat;
  struct remote_req rreq;
  int send_done, recv_done;
  static int reentrant = 0;

  StartTimer(&RemoteTm);
  if( reentrant ){
    Error("RemoteRead is NOT reentrant!\n");
  }else{
    reentrant++;
  }

  MPMY_Irecv(addr, 1<<_LGPGSZ, pgid.proc, REPLY_TAG, &rdreq);
  rreq.id = pgid;
  rreq.replytag = REPLY_TAG;
  Msg("Poll", ("Request page %d from %d\n", pgid.pgnum, pgid.proc));
  MPMY_Isend(&rreq, sizeof(rreq), pgid.proc, REQUEST_TAG, &sndreq);
  send_done = 0;
  recv_done = 0;
  MPMY_Test(sndreq, &send_done, NULL);
  MPMY_Test(rdreq, &recv_done, &rdstat);
  while( !(send_done && recv_done )){
    /* This shouldn't deadlock, but it will busy-wait, diverting cycles
       from the process that might be servicing our request.  Maybe it's
       not a problem in a one-process-per-machine environment.  It's a
       disaster for uni-processor lsv */
    StartTimer(&PgPollLoopTm);
    PagePoll();
    StopTimer(&PgPollLoopTm);
    IncrCounter(&PgPollLoopCnt);
    if( !send_done ){
      MPMY_Test(sndreq, &send_done, NULL);
      if( send_done )
	Msg("Poll", ("request sent\n"));
    }
    if( !recv_done ){
      MPMY_Test(rdreq, &recv_done, &rdstat);
      if( recv_done )
	Msg("Poll", ("reply received.  page %d from %d ready\n", pgid.pgnum, pgid.proc));
    }
  }
  IncrCounter(&PgRemoteInCnt);
  assert( rdstat.MPMY_COUNT == (1<<_LGPGSZ));
  reentrant = 0;
  StopTimer(&RemoteTm);
}

void PagePollTillDone(void){
  StartTimer(&PgPollTillDoneTm);
  ABMIamDone(&remoteabm);
  Msg("Poll", ("PollTillDone - IamDone\n"));
  while(!ABMAllDone(&remoteabm)){
    PagePoll();
    /* ABMPoll is called in PagePoll.  See the comment there.  If
       that becomes unnecessary and is removed, it needs to be called here. */
    if( !lsv_tag_bug )
      ABMPoll(&remoteabm);
  }
  /* This is the easiest 'reset' */
  ABMShutdown(&remoteabm);
  /* Increment the tag, so we don't get burned by messages from the previous
     cycle... I sense something deeply wrong with this, but I don't know 
     what it is. */
  if( ++remote_abm_tag > REMOTE_ABM_TAG + 128 )
    remote_abm_tag = REMOTE_ABM_TAG;
  Msg("Poll", ("Starting up new abm with tag %d\n", remote_abm_tag));
  ABMSetup(&remoteabm, 8, remote_abm_tag, 0, NULL);
  StopTimer(&PgPollTillDoneTm);
}

/* This code looks an awful lot like PagePollTillDone.  The only
   difference is what we do when we are finished polling.  There oughta'
   be a subroutine.  Until then, if one changes, the other probably should
   too... */
static void RemoteTerm(void){
  struct remote_req junk;
  MPMY_Comm_request sndreq;


  /* Clean up everything and leave the system in a state so nobody
     knows we were ever here. -- (yeah, right) */
  Msg("Poll", ("RemoteTerm - IamDone\n"));
  ABMIamDone(&remoteabm);
  while( !ABMAllDone(&remoteabm) ){
    Msg("Poll", ("not Alldone yet, calling PagePoll\n"));
    PagePoll();
    /* ABMPoll is called in PagePoll.  See the comment there.  If
       that gets fixed, it needs to come back here. */
    if( !lsv_tag_bug )
      ABMPoll(&remoteabm);
  }
  ABMShutdown(&remoteabm);
  Msg("Poll", ("ABMShutdown complete\n"));

  /* Balance the pending recv request on inreq.  */
  MPMY_Isend(&junk, sizeof(junk), (MPMY_Procnum()),
	     REQUEST_TAG, &sndreq);
  MPMY_Wait2(sndreq, NULL, inreq, NULL);
}
