#ifdef __OPTIMIZE__
#define NO_MSGS
#endif

#include "pagehash.h"
#include "page.h"
/* We only need a singl-direction linked list, but we've got dll's
   coded up... */
#include "dll.h"
#include "Malloc.h"
#include "Msgs.h"

__inline__ static unsigned int
hash_signature(PgId *id){
  /* A couple of random constants.  Prime is good.  Any other criteria? 
     Is integer multiply too slow?  Should we do shifts and adds
     instead?  
   */
  return id->pgnum * 17 + id->proc * 3;
}

__inline__ static unsigned int
IdEqual(PgId *p1, PgId *p2){
  return p1->proc==p2->proc && p1->pgnum == p2->pgnum;
}

/* Create a table.  */
void hash_make(hash_t *ht, int tblsz){
  int i;
  ht->tblsz = tblsz;
  ht->tbl = Malloc(ht->tblsz*sizeof(Dll));
  DllCreateChn(&ht->chn, sizeof(hash_entry_t), tblsz);
  for(i=0; i<ht->tblsz; i++){
    /* They all share the same chn! */
    DllCreate(&ht->tbl[i], &ht->chn);
  }
}

/* Destroy a hash table */
void hash_destroy(hash_t *ht){
  int i;
  for(i=0; i<ht->tblsz; i++){
    DllTerminate(&ht->tbl[i]);
  }
  ChnTerminate(&ht->chn);
  Free(ht->tbl);
  ht->tblsz = -1;
}

/* Perform internal consistency checks on a hash table */
int hash_check(hash_t *ht){
  return 0;			/* ?? */
}

/* Look for the key, and if not found, return NULL, and also set
 * *dllp to the correct dll, so we don't have to do the signature/mod again.
 */
__inline__
hash_entry_t *_hash_find(hash_t *ht, PgId *pgid, Dll **dllp){
  unsigned int sig = hash_signature(pgid);
  unsigned int idx = sig % ht->tblsz;
  Dll *dll = &ht->tbl[idx];
  Dll_elmt *elmt;

  for(elmt = DllBottom(dll);
      elmt != DllSup(dll);
      elmt = DllUp(elmt)){
    hash_entry_t *he;
    he = DllData(elmt);
    if( IdEqual(&he->id, pgid) )
      return he;
  }
  /* Make it easy to find the table entry again. */
  if( dllp )
    *dllp = dll;
  return NULL;
}

/* Look for the key, and if not found, create a new entry */
hash_entry_t *hash_get(hash_t *ht, PgId *pgid){
  hash_entry_t *ret;
  Dll *dll;

  Msgf(("hash_get pgid=%s\n", PrintPgId(*pgid)));
  ret = _hash_find(ht, pgid, &dll);
  if( ret == NULL ){
    Dll_elmt *new;

    Msgf(("hash_find(%s) returns null, dll=ht->tbl[%ld]\n",
	  PrintPgId(*pgid), (long)(dll-ht->tbl)));
    new = DllInsertAtBottom(dll);
    ret = DllData(new);
    ret->id = *pgid;
    ret->myelmt = new;
    ret->mydll = dll;
    ret->data = NULL;
    Msgf(("hash_get creating new entry at %p, pgid=%s\n", ret, PrintPgId(*pgid)));
  }
  Msgf(("hash_get returning %p in ht->tbl[%ld]\n", ret,
	(long)(ret->mydll - ht->tbl)));
  return ret;
}

/* Look for the key and return NULL if it's not there */
hash_entry_t *hash_find(hash_t *ht, PgId *pgid){
  Msgf(("hash_find pgid=%s\n", PrintPgId(*pgid)));
  return _hash_find(ht, pgid, NULL);
}


/* Delete an entry */
void hash_delete(hash_t *ht, hash_entry_t *he){
  Msgf(("hash_del %p pgid=%s\n", he, he?PrintPgId(he->id):""));
  if( he == NULL )
    return;
  DllDelete(he->mydll, he->myelmt);
}

/* release any 'locks' on the entry.  If we aren't running multi-threaded
   then this does nothing.  There's no harm in keeping the API, though */
void hash_release(hash_t *ht, hash_entry_t*he){
}

/* Start an iteration through the hash table.  By calling hash_iter_next()
   repeatedly, all entries will be returned.

   If multithreaded, hold a lock on the entire table based on the
   value of hold_global_lock.  Use it if you want to be absolutely
   sure nobody is modifying the table while you're looking at it.  If
   a global lock is not held, then entries may be added or deleted by
   other threads while the iteration proceeds.  Consistency is a
   complicated issue.  Do we care?  */
void hash_iter_start(hash_iterator_t *hiter, hash_t *ht, int hold_global_lock){
  hiter->ht = ht;
  hiter->lastdll = &ht->tbl[0];
  hiter->last = DllInf(hiter->lastdll);
}

void hash_iter_finish(hash_iterator_t *hiter){
}

/* Return the next hash entry in the iteration.  Return NULL when
   there are no more entries.

   If multi-threaded, A per-entry lock would be held until the next call to
   hash_iter_next() or hash_iter_delete().  */
hash_entry_t *hash_iter_next(hash_iterator_t *hiter){
  Dll_elmt *next;

  if( hiter->last == NULL ){
    Msg("hiter", ("hinext: last==NULL, returning NULL\n"));
    return NULL;
  }

  next = DllUp(hiter->last);
  if( next == DllSup(hiter->lastdll) ){
    hiter->lastdll++;
    Msg("hiter", ("Iteration moving up to ht->tbl[%ld]\n", 
		  (long)(hiter->lastdll - hiter->ht->tbl)));
    if( hiter->lastdll - hiter->ht->tbl >= hiter->ht->tblsz ){
      hiter->last = NULL;
    }else{
      hiter->last = DllInf(hiter->lastdll);
    }
    /* Call self recursively to look in subsequent tbl entries */
    Msg("hiter", ("hinext:  recursive retry\n"));
    return hash_iter_next(hiter);
  }
  Msg("hiter", ("hinext:  returning he with id=%s\n", 
	PrintPgId( ((hash_entry_t *)DllData(next))->id )));
  hiter->last = next;
  return DllData(next);
}
