#ifdef __OPTIMIZE__
#define NO_MSGS
#else
/* if we aren't optimizing, we might as well do lots of checking.  This
   means there will be a HUGE difference between -O and not.  */
#define DEBUG
#endif

#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include <string.h>

#include "protos.h"
#include "Assert.h"
#include "Msgs.h"
#include "Malloc.h"
#include "tree.h"
#include "key.h"
#include "chn.h"
#include "page.h"
#include "mpmy.h"

#ifdef DEBUG
static void chkSgtmp_(int lineno, tree_t *tp, sg_tmp *sg);
#define chkSgtmp(t, s) chkSgtmp_(__LINE__, t, s)
#else
#define chkSgtmp(t, s)
#endif

#define CHILD_IMPOSSIBLE (1<<MAXNDIM)

static Key_t Key0;

#ifndef NO_MSGS
/* we want our error messages to be grammatically correct */
static const char *ordinal_string[4] = {"th", "st", "nd", "rd"};
#endif

static oocptr nextoocp[KEYBITS];
/* This program was originally written using static vars with file
   scope with the following names.  Subsequently this information was
   encapsulated in the tree_t structure itself, so we can have several
   builds going on simultaneously.  It's just too tedious to add
   t->bs. to every single variable reference in this file.  So we
   cheat and use the preprocessor.  Think of t-> as the C++ 'this' var
   and it all makes sense.  With this trick, it's impossible to 'shadow'
   a tree element with an automatic variable, but such practice is bad
   style anyway */
#define rootsg (t->rootsg)
#define ndim (t->ndim)
#define m (t->m)
#define curr_grp (t->curr_grp)
#define rootsgtmp (t->rootsgtmp)
#define firstsg (t->firstsg)
#define curr_child (t->curr_child)
#define curr_level (t->curr_level)
#define maxlevs (t->maxlevs)
#define next_tc_tmp (t->next_tc_tmp)
#define rshift (t->rshift)
#define ndimmask (t->ndimmask)
#define count (t->count)
#define prevkey (t->prevkey)
#define Bfuncs (t->Bfuncs)
#define norphans (t->norphans)
#define holdidx (t->holdidx)
#define holdingsz (t->holdingsz)
#define holding (t->holding)
#define orphans (t->orphans)
#define sgchn (t->sgchn)
#define CofmChn (t->CofmChn)
#define npages (t->npages)
#define maxpages (t->maxpages)
#define pages (t->pages)

static void MergeParallel(tree_t *t);
static void callAggregate(tree_t *t, sg_tmp *sgtmp);
#ifndef NO_MSGS
static char *PrintTctmp(const tc_tmp *tc);
#endif
static char *PrintOOC(oocptr p);
static char *PrintSgtmp(const sg_tmp *sg);
extern char *PrintTc(const tcell *tc);
extern char *PrintSg(const sib_group *sg);

/* If this is set to non-NULL, and if DEBUG is on in this file, then
   we can do a lot more detailed sanity-checking while tree building */
Key_t (*tree17keyFromTbody)(const _tbody *);

static sg_tmp*
GrpAlloc(tree_t *t){
  sg_tmp *p;

  p = ChnAlloc(&sgchn);
  p->ncell = 0;
  p->ntbody = 0;
  p->flags = 0;
  Msgf(("GrpAlloc(): %p\n", p));
  return p;
}

static oocptr
OOCAlloc(tree_t *t, int level, size_t size){
  oocptr p;

  /* Put the first three levels together.  That's 1 + 1 + 8 + 64 = 74
   sg_tmps, which ought to comfortably fit in any reasonable sized
   page. */
  if( level < 0 )
    level = 0;
  if( size  > PgSize )
    Error("You can't allocate an OOC object size=%ld > PgSize=%d\n",
	  (long int)size, PgSize);
  p = nextoocp[level];

  if( p.offset + size > PgSize ){
    p.pgid = PgAlloc();
    p.offset = 0;
    nextoocp[level] = p;
    Msg("pgcount", ("%d %d TreePage\n", p.pgid.pgnum, level));
    if( npages == maxpages ){
      maxpages += 100;
      pages = Realloc(pages, sizeof(PgId)*(maxpages));
      if( pages == NULL )
	Error("Can't realloc the page table\n");
    }
    pages[npages++] = p.pgid;
  }
  nextoocp[level].offset += size;
  return p;
}

static oocptr
GrpScrunch(tree_t *t, sg_tmp *sg, int level){
  int i;
  oocptr oocfinal;
  PgHndl finalhndl;
  sib_group *final;
  void *to;
  size_t szneed, used;

  assert(sg);
  szneed = offsetof(sib_group, tcells[sg->ncell]);

  for(i=0; i<sg->ntbody; i++){
    szneed += sg->tbodysz[i];
  }
  for(i=0; i<sg->ncell; i++){
    szneed += sg->tc_tmps[i].contents_sz;
  }

  oocfinal = OOCAlloc(t, level, szneed);
  final = OOCFollow(oocfinal, &finalhndl);

  Msgf(("Gs - OOCF k=%s, sg->ncell=%d, sg->ntbody=%d\n", PrintKey(sg->key), sg->ncell, sg->ntbody));
  final->ncell = sg->ncell;
  final->ntbody = sg->ntbody;
  final->key = sg->key;
  to = &final->tcells[sg->ncell];
  Msgf(("to = %p\n", to));
  for(i=0; i<sg->ncell; i++){
    tc_tmp *tc = &sg->tc_tmps[i];
    tcell *tcf = &final->tcells[i];
    Msgf(("Repackaging tc[%d] (%s.%o), to %p\n", 
	  i, PrintTctmp(tc), tc->bits, to));
    used = tc->contents_sz;
    (*Bfuncs->makeFinalCell)(to, tc->contents, tc->contents_sz);
    tcf->bits = tc->bits;
    tcf->offset = (char *)to - (char *)tcf;
    tcf->daughter = tc->oocdaughter;
    Msgf(("Repackaged tcf=%p, offset=%d (%ld)\n", tcf, tcf->offset, 
	  (long)((char*)to- (char*)tcf)));
    to = (char *)to + used;
  }

  /* And finally, the tbodies. */
  final->tbody_offset = (char *)to - (char *)final;
  for(i=0; i<sg->ntbody; i++){
    Msgf(("Copying tbodyptr[%d] from %p to %p, sz=%ld\n",
	  i, sg->tbodyptrs[i], to, (unsigned long) sg->tbodysz[i]));
    memcpy(to, sg->tbodyptrs[i], sg->tbodysz[i]);
    Free(sg->tbodyptrs[i]);
    to = (char *)to + sg->tbodysz[i];
  }

  Msgf(("Unref finalhndl\n"));
  PgUnref(finalhndl, PG_DIRTY);
  if( ((char *)to-(char *)final) != szneed ){
    Error("scrunch sizes do not add up:  to-final=%ld, szneed=%ld!\n", 
		   (long)((char *)to - (char *)final), (long)szneed);
  }

  for(i=0; i<sg->ncell; i++){
    ChnFree(&CofmChn, sg->tc_tmps[i].contents);
  }
  Msgf(("Freed cofm contents\n"));
  if( sg->parent ){
    sg->parent->oocdaughter = oocfinal;
    sg->parent->daughter = NULL;
  }

  ChnFree(&sgchn, sg);
  return oocfinal;
}

/* This is ugly.  It's how we reference into the holding array.
   Prev is usually positive, and we want the "prev'th" previous
   element of the holding array.  This was just a BAD idea.  */
static __inline__ int 
HIDX(tree_t *t, int prev){
  int off = (holdidx - prev) - 1;
  while( off < 0 )
    off += (holdingsz);
  while( off >= holdingsz )
    off -= holdingsz;
  return off;
}

static void callAggregate(tree_t *t, sg_tmp *sgtmp){
  const _cellDataIntermediate *contentsptrs[1<<MAXNDIM];
  int i;

  sgtmp->parent->contents = ChnAlloc(&CofmChn);
  for(i=0; i<sgtmp->ncell; i++){
    contentsptrs[i] = (sgtmp)->tc_tmps[(i)].contents;
  }

  /* Why is the (const _tbody *) cast necessary?? */
  sgtmp->parent->contents_sz = 
    (*Bfuncs->makeIntermediateCell)(sgtmp->key, sgtmp->ntbody, 
				    (const _tbody **)sgtmp->tbodyptrs, 
				    sgtmp->ncell, contentsptrs, 
				    sgtmp->parent->contents);
}

void
TreeBuildStart(tree_t *t, int ndim_arg, int m_arg, BuildFuncs_t *Bfuncs_arg){
  int i;

  ndim = ndim_arg;
  m = m_arg;
  Bfuncs = Bfuncs_arg;

  Msgf(("In BuildTree\n"));
  for(i=0; i<KEYBITS; i++){
    /* nextoocp[i].pgnum = -1; */
    /* setting the offset to PgSize will force a PgAlloc */
    nextoocp[i].offset = PgSize;
  }
  npages = 0;
  maxpages = 0;
  pages = NULL;
  Key0 = KeyInt(0x0);

  holdingsz = m+1;
  holding = Malloc(holdingsz*sizeof(*holding));
  for(i=0; i<(m+1); i++){
    holding[i].tb = NULL;
    holding[i].sz = 0;
  }
  holdidx = 0;
  orphans = Malloc(m*sizeof(*holding));
  norphans = 0;

  ChnInit(&sgchn, sizeof(sg_tmp), 100, Realloc_f);
  ChnInit(&CofmChn, Bfuncs->intermediatesz, 100, Realloc_f);

  maxlevs = (KEYBITS-1)/ndim;
  ndimmask = (1<<ndim) - 1;

  /* The root sg_tmp is special.  It has no parent, and we will never
     call makeIntermediateCell for it.  But we will call 'Scrunch' to free up
     it's only cell's intermediate _cellDataIntermediate* contents,
     and to turn it into an sib_group. */
  rootsgtmp.key = KeyInt(0x0);
  rootsgtmp.ncell = 0;
  rootsgtmp.ntbody = 0;
  rootsgtmp.flags = 0;
  rootsgtmp.parent = NULL;
  curr_grp = &rootsgtmp;

  if( MPMY_Procnum() == 0 ){
    firstsg = &rootsgtmp;
  }else{
    firstsg = NULL;
  }

  curr_child = CHILD_IMPOSSIBLE;
  next_tc_tmp = curr_grp->tc_tmps;
  curr_level = -1;
  count = 0;
  rshift = (maxlevs+1)*ndim;

  prevkey = Key0;
}

/* The Plan: 

   The user scans his 'btab' in whatever format he keeps it.  He calls
   BuildTreeAdd to add bodies to the tree, one at a time.  Keys must
   be monotonically increasing.

   We keep track of the current state of
   the tree using the static variables up above.  What to do depends
   on whether the new body lands:

   - in the current sib-grp or not
   - in the current child or not
   - whether the current child has 'm' bodies already.

   State for TreeBuildAdd is retained in: holding[], curr_grp, curr_child,
   curr_level, count, rshift, next_tc_tmp, and prevkey (used
   only for diagnostics).  In addition, we always have:

   rshift = ndim*(max_level - curr_level)
   curr_level == TreeLevel(curr_grp);
   next_tc_tmp == &curr_grp->tc_tmps[curr_grp->ncell]

  */

static void
TreeBuildAdd_(tree_t *t, int stkdepth){
  Key_t bkey;
  size_t tbodysz;
  Key_t bshift;
  int bchild;
  int i, j;
  int jj;

  jj = HIDX(t, stkdepth);
  bkey = holding[jj].key;
  tbodysz = holding[jj].sz;
  if( Msg_test(__FILE__) ){
    Msg_do("\nTBA_(bkey=%s, tb=%p, sz=%ld)\n",
	   PrintKey(bkey), holding[jj].tb, (long int)tbodysz);
    Msg_do("%s\n", PrintSgtmp(curr_grp));
    Msg_do("curr_level=%d, rshift=%d, count=%d\n", 
	   curr_level, rshift, count);
  }
  assert(curr_level == TreeLevel(curr_grp->key, ndim));
  assert(rshift == ndim*(maxlevs - curr_level));
  assert(&curr_grp->tc_tmps[curr_grp->ncell] == next_tc_tmp);
  if( KeyGT(prevkey, bkey) ){
    Shout("Keys out of order:  prevkey=%s, ", PrintKey(prevkey));
    Shout("bkey=%s\n", PrintKey(bkey));
    Error("Keys out of order\n");
  }
  prevkey = bkey;
  Msgf(("Next key=%s\n", PrintKey(bkey)));
  bshift = KeyRshift(bkey, rshift - ndim);
  bchild = KeyAndInt(bshift, ndimmask);
  Msgf(("bshift: %s\n", PrintKey(bshift)));
  Msgf(("bchild = %d, curr_grp->key = %s\n", bchild, PrintKey(curr_grp->key)));

  if( KeyEQ( KeyRshift(bshift, ndim), curr_grp->key ) ){
    /* if( body is contained in the current grp key ) */
    if( bchild == curr_child ){
      sg_tmp *new_grp;
      /* And it's also within the current child */
      if( count < m ){
	/* And there's space for another body in the current child */
	Msgf(("Body fits in current cell count++=%d\n", count));
	count++;
	return;
      }
      /* Otherwise, we have to create a child */
      Msgf(("Body overflows current cell, creating a child\n"));
      curr_level++;
      rshift -= ndim;
      curr_child = CHILD_IMPOSSIBLE;
      new_grp = GrpAlloc(t);
      if( firstsg )
	new_grp->flags |= COMPLETE_LO;
      new_grp->parent = next_tc_tmp;
      new_grp->key = KeyOrInt(KeyLshift(curr_grp->key, ndim), bchild);
      next_tc_tmp->bits = bchild;
      next_tc_tmp->daughter = new_grp;
      next_tc_tmp->sibs = curr_grp;
      curr_grp->ncell++;
      curr_grp = new_grp;
      next_tc_tmp = curr_grp->tc_tmps;

      /* Start over with the bodies in the current cell at the lower
	 level */
      assert(count == m);
      /* Retry the guys in the 'holding' array */
      prevkey = Key0;		/* don't flag a false out-of-order error! */
      count = 0;
      for(i=0; i<m; i++){
	TreeBuildAdd_(t, stkdepth + (m-i));
      }
      Msgf(("Recursive retry of current:\n"));
      TreeBuildAdd_(t, stkdepth);
      return;
    }else{
      /* The new body begins a different child.  Copy the 'count'
	 children to the tbody area. */
      Msgf(("Body begins a new child (%d), count=%d\n", bchild, count));
      if(firstsg == NULL && count > 0 ){
	Msg("merge", ("%d true orphans created\n", count));
	for(i=0; i<count; i++){ /* send current body too! */
	  j = HIDX(t, stkdepth + count - i);
	  orphans[norphans++] = holding[j];
	}
	firstsg = curr_grp;
	Msg("merge", ("Discovered firstsg after %d orphans, bodykey=%s\n", 
	      norphans, PrintKey(bkey)));
	Msg("merge", ("firstsg = %s\n", PrintSgtmp(firstsg)));
	for(i=0; i<norphans; i++){
	  Msg("merge", ("orphans[%d].key = %s\n", i, PrintKey(orphans[i].key)));
	}
      }else{
	for(i=0; i<count; i++){
	  j = HIDX(t, stkdepth + count-i);
	  Msgf(("i=%d, curr_grp->tbodyptrs[%d] = holding[%d].tb = %p\n",
		i, curr_grp->ntbody, j, holding[j].tb));
	  curr_grp->tbodyptrs[curr_grp->ntbody] = holding[j].tb;
	  curr_grp->tbodysz[curr_grp->ntbody] = holding[j].sz;
	  curr_grp->ntbody++;
	}

	if( count > 0 )
	  Msgf(("Copied %d%s body to curr_grp: %s\n", curr_grp->ntbody,
		(curr_grp->ntbody>3)? "th" : ordinal_string[curr_grp->ntbody],
		PrintKey(curr_grp->key)));
      }
      curr_child = bchild;
      count = 1;
      return;
    }
  }else{
    /* Body falls outside the current group. */
    Msgf(("Body begins a new sib_grp.  Peeling back\n"));
    assert(count <= m);
    if( firstsg == NULL ){
      Shout("DEAD CODE:  firstsg==NULL when beginning a new grp??\n");
      for(i=0; i<count; i++){
	j = HIDX(t, stkdepth + count - i);
	orphans[norphans++] = holding[j];
      }
      Msg("merge", ("firstsg discovered for new grp: %s\n", PrintSgtmp(curr_grp)));
      Msg("merge", ("count=%d, bodykey = %s\n", count, PrintKey(bkey)));
      firstsg = curr_grp;
      Msgf(("final norphans = %d\n", norphans));
      for(i=0; i<norphans; i++){
	Msg("merge", ("orphans[%d].key = %s\n", i, PrintKey(orphans[i].key)));
      }
    }else{
      for(i=0; i<count; i++){
	j = HIDX(t, stkdepth + count-i);
	Msgf(("curr_grp->tbodyptrs[%d] = %p\n",
	      curr_grp->ntbody, holding[i].tb));
	curr_grp->tbodyptrs[curr_grp->ntbody] = holding[j].tb;
	curr_grp->tbodysz[curr_grp->ntbody] = holding[j].sz;
	curr_grp->ntbody++;
      }
    }

    /* Now pop back up to the parent level */
    next_tc_tmp = curr_grp->parent;
    curr_grp->flags |= COMPLETE_HI;
    if( Completep(curr_grp) ){
      callAggregate(t, curr_grp);
      Msgf(("curr_grp=%p, curr_grp->parent=%p\n", curr_grp, curr_grp->parent));
      GrpScrunch(t, curr_grp, curr_level);
    }else{
      Msgf(("Deferring scrunch of grp key=%s, flags=%d\n",
	    PrintKey(curr_grp->key), curr_grp->flags));
      Msgf(("Sgtmp: %s\n", PrintSgtmp(curr_grp)));
    }
    curr_level--; 
    rshift += ndim; 
    curr_child = CHILD_IMPOSSIBLE;
    curr_grp = next_tc_tmp->sibs;
    next_tc_tmp++;
    /* And try this body again at the higher level */
    count = 0;
    TreeBuildAdd_(t, stkdepth);
    return;
  }
}

void
TreeBuildAdd(tree_t *t, Key_t bkey, _tbody *tb, size_t tbodysz){
  holding[holdidx].key = bkey;
  Msgf(("\nTreeBuildAdd(bkey=%s, tb=%p, sz=%ld)",
	PrintKey(bkey), tb, (long int)tbodysz));
  holding[holdidx].tb = Malloc(tbodysz); /* Freed by GrpScrunch */
  memcpy(holding[holdidx].tb, tb, tbodysz);
  Msgf(("-> holding[%d].tb = %p\n",
	holdidx, holding[holdidx].tb));

  holding[holdidx].sz = tbodysz;
  /* Make it a circular buffer. */
  if( ++holdidx >= holdingsz )
    holdidx = 0;

  TreeBuildAdd_(t, 0);	/* top of stack */
}

oocptr
TreeBuildFinish(tree_t *t){
  int i, j;

  /* We now have our local tree built.  But some of the sgs are
     incomplete, lacking either _HI or _LO flags (or maybe both).  On
     a uni-processor, we are actually done at this point.  In
     parallel, we have to sew these trees together at the edges. */
  MergeParallel(t);

  assert(norphans <= m);

  /* Attach any remaining bodies to the last group we were looking at */
  Msgf(("Last processor, all count=%d remaining bodies to last cell\n%s\n",
	count, PrintSgtmp(curr_grp)));
  for(i=0; i<count; i++){
    j = HIDX(t, (count-1)-i);
    curr_grp->tbodyptrs[curr_grp->ntbody] = holding[j].tb;
    curr_grp->tbodysz[curr_grp->ntbody] = holding[j].sz;
    curr_grp->ntbody++;
  }
  count = 0;
  /* recurse upwards finishing off all the cells and declaring
     them COMPLETE_HI */
  Msgf(("Loop up to root\n"));
  while( curr_level >= 0 ){
    sg_tmp *up;
    up = curr_grp->parent->sibs;
    curr_grp->flags |= COMPLETE_HI;
    if( Completep(curr_grp) ){
      Msgf(("Scrunching grp at level %d, key=%s\n", curr_level, 
	    PrintKey(curr_grp->key)));
      callAggregate(t, curr_grp);
      GrpScrunch(t, curr_grp, curr_level);
    }else{
      Msgf(("Deferring scrunch of grp key=%s\n", PrintSgtmp(curr_grp)));
    }
    curr_grp = up;
    --curr_level;
  }

  Msgf(("Done scanning bodies, count=%d bodies left unattached\n", count));
  Msgf(("MergeParallel Finished curr_grp = %s\n", PrintSgtmp(curr_grp)));
  /* Don't call makeIntermediateCell for the root.  We don't want to
   allocate another aggregate structure!  This Scrunch will delete the
   only one remaining in rootsgtmp. */
  rootsg = GrpScrunch(t, &rootsgtmp, -1);
  Msgf(("After GrpScrunch(&rootsgtmp)\n"));
  ChnTerminate(&sgchn);
  ChnTerminate(&CofmChn);
  Free(holding);
  Free(orphans);
  return rootsg;
}

oocptr
TreeRoot(tree_t *t){
  return rootsg;
}

void
TreeFree(tree_t *t){
  int i;

  for(i=0; i<npages; i++){
    PgFree(pages[i]);
  }
  Free(pages);
  PgFlushRemote();
  npages = 0;
  maxpages = 0;
  pages = NULL;
}

/* Now for the code that takes a bunch of trees built in parallel, and
   merges them so they have consistent upper levels. */

/* static buffers are ok because they are live only during MergeTrees. */
static char inbuf[65536]; 
static char outbuf[65536]; 
static char *inptr, *outptr, *inend;

static void
Enqueue(tree_t *t, sg_tmp *sg){
  int i;
  sg_tmp *outsg;

  outsg = (sg_tmp *)outptr;	/* alignment problems?? */
  memcpy(outptr, sg, offsetof(sg_tmp, tc_tmps));
  outptr += offsetof(sg_tmp, tc_tmps);
  Msg("merge", ("Enqueuing %s\n", PrintSgtmp(sg)));
  for(i=0; i<sg->ncell; i++){
    tc_tmp *tc;
    tc = &sg->tc_tmps[i];
    if( tc->daughter == NULL ){
      assert(tc->contents);
      Msg("merge", ("cell[%d]@%p, bits=%d, daughter=%p, sibs=%p, oocd=%s\n", 
	    i, tc, tc->bits, tc->daughter,
	    tc->sibs, PrintOOC(tc->oocdaughter)));
      memcpy(outptr, tc, sizeof(tc_tmp));
      outptr += sizeof(tc_tmp);
      memcpy(outptr, tc->contents, Bfuncs->intermediatesz);
      Msg("merge", ("enqueueing tcell[%d] bits=%d\n", 
		    i, tc->bits));
      outptr += Bfuncs->intermediatesz;
    }else{
      Msg("merge", ("Skipping daughter @%p - key=%s\n", 
	    tc->daughter, PrintKey(tc->daughter->key)));
      outsg->ncell--;
    }
  }
  
  for(i=0; i<sg->ntbody; i++){
    Msg("merge", ("copying tbodyptf[%d]=%p, sz=%ld\n",
		  i, sg->tbodyptrs[i], (long)sg->tbodysz[i]));
    *(int *)outptr = sg->tbodysz[i];
    outptr += sizeof(int);
    memcpy(outptr, sg->tbodyptrs[i], sg->tbodysz[i]);
    outptr += sg->tbodysz[i];
  }
  assert(outptr - outbuf < sizeof(outbuf));
  Msg("merge", ("Enqueue returning\n"));
}

static sg_tmp *
Dequeue(tree_t *t){
  sg_tmp *sg;
  int i;

  if(inptr == inend)
    return NULL;

  sg = ChnAlloc(&sgchn);
  memcpy(sg, inptr, offsetof(sg_tmp, tc_tmps));
  inptr += offsetof(sg_tmp, tc_tmps);
  Msg("merge", ("Dequeuing sg->key=%s, ncell=%d\n",
	PrintKey(sg->key), sg->ncell));
  for(i=0; i<sg->ncell; i++){
    tc_tmp *tc;
    tc = &sg->tc_tmps[i];
    memcpy(tc, inptr, sizeof(tc_tmp));
    inptr += sizeof(tc_tmp);
    tc->contents = ChnAlloc(&CofmChn);
    tc->sibs = sg;
    memcpy(tc->contents, inptr, Bfuncs->intermediatesz);
    inptr += Bfuncs->intermediatesz;
    Msg("merge", ("cell[%d], (%s.%o), daughter=%p, sibs=%p, oocd=%s\n", 
		  i, PrintKey(sg->key), tc->bits, tc->daughter,
		  tc->sibs, PrintOOC(tc->oocdaughter)));
  }

  Msg("merge", ("Dequeuing %d tbodies\n", sg->ntbody));
  for(i=0; i<sg->ntbody; i++){
    _tbody *bp;
    int sz;
    sz = *(int *)inptr;
    inptr += sizeof(int);
    bp = Malloc(sz);
    memcpy(bp, inptr, sz);
    Msg("merge", ("Dequeued tbody[%d]@%p sz=%d\n",
	  i, bp, sz));
    sg->tbodyptrs[i] = bp;
    sg->tbodysz[i] = sz;
    /* Alignment? */
    inptr += sz;
  }
  assert(inptr <= inend);
  return sg;
}

static void
Clearqueue(void){
  inend = inptr = inbuf;
  outptr = outbuf;
}

static void
Exchg(int otherproc){
  int inlen;
  int outlen;

  if( otherproc < MPMY_Nproc() ){
    outlen = outptr - outbuf;
    Msg("merge", ("Exchg with %d, outlen=%d\n", otherproc, outlen));
    MPMY_Shift(otherproc, &inlen, sizeof(inlen), &outlen, sizeof(outlen), NULL);
    assert(inlen <= sizeof(inbuf));
    Msg("merge", ("Exchg inlen=%d\n", inlen));
    MPMY_Shift(otherproc, inbuf, inlen, outbuf, outlen, NULL);
    inend = inbuf + inlen;
  }
}

static void
MergeGroups(tree_t *t, sg_tmp *tsg, sg_tmp *sg, int IamHi){
  int i;
  tc_tmp *fromtc, *totc;
  int totbindx;

  Msg("merge", ("Mergegroups with key=%s\n", PrintKey(tsg->key)));
  assert(KeyEQ(tsg->key, sg->key));
  sg->flags |= tsg->flags;
  if( IamHi ){
    /* We need to move the current contents of sg to make room for
       the incoming stuff. */
    i = sg->ncell;
    while( i-- > 0 ){
      tc_tmp *tctmp;

      Msg("merge", ("Copying tc_tmps[%d] -> tc_tmps[%d]\n", i, i+tsg->ncell));
      tctmp = &sg->tc_tmps[i+tsg->ncell];
      *tctmp = sg->tc_tmps[i];
      if( tctmp->daughter ){
	tctmp->daughter->parent = tctmp;
      }
    }
    i = sg->ntbody;
    while( i-- > 0 ){
      Msg("merge", ("Copying bodyptr[%d](%p) -> bodyptr[%d]\n", i, sg->tbodyptrs[i], i+tsg->ntbody));
      sg->tbodyptrs[i+tsg->ntbody] = sg->tbodyptrs[i];
      sg->tbodysz[i+tsg->ntbody] = sg->tbodysz[i];
    }
    totc = &sg->tc_tmps[0];
    totbindx = 0;
  }else{
    totc = &sg->tc_tmps[sg->ncell];
    totbindx = sg->ntbody;
  }
  Msg("merge", ("Merge %d tcells with %d already present\n", tsg->ncell, sg->ncell));
  fromtc = &tsg->tc_tmps[0];
  for(i=0; i<tsg->ncell; i++){
    *totc = *fromtc;
    totc->sibs = sg;
    totc++;
    fromtc++;
  }
  sg->ncell += tsg->ncell;
  assert(sg->ncell <= (1<<ndim));
  Msg("merge", ("Final ncell=%d\n", sg->ncell));

  Msg("merge", ("Merge %d tbodies with %d already present\n", tsg->ntbody, sg->ntbody));
  for(i=0; i<tsg->ntbody; i++){
    sg->tbodyptrs[totbindx] = tsg->tbodyptrs[i];
    sg->tbodysz[totbindx] = tsg->tbodysz[i];
    totbindx++;
  }
  sg->ntbody += tsg->ntbody;
  if( sg->ntbody > (sizeof(tsg->tbodyptrs)/sizeof(*tsg->tbodyptrs)) ){
    Error("Tbody overflow while merging groups!");
  }
  ChnFree(&sgchn, tsg);
  chkSgtmp(t, sg);
  Msg("merge", ("MergeGroups finished\n"));
}

static void
RecvHolding(tree_t *t, int partner){
  int i;

  Msg("merge", ("Receiving %d holding bodies from partner\n", count));
  MPMY_recvn(&count, sizeof(int), partner, 300);
  for(i=0; i<count; i++){
    MPMY_recvn(&holding[holdidx], sizeof(*holding), partner, 301);
    holding[holdidx].tb = Malloc(holding[i].sz);
    MPMY_recvn(holding[holdidx].tb, holding[i].sz, partner, 302);
    prevkey = holding[holdidx].key;
    if( ++holdidx >= holdingsz )
      holdidx = 0;
  }
  if( count == 0 ){
    if( norphans > 0 ){
      prevkey = orphans[norphans-1].key;
    }else{
      prevkey = KeyInt(0x0);
    }
    Msg("merge", ("Mlo: count=%d, norphans=%d, setting prevkey = %s\n",
		  count, norphans, PrintKey(prevkey)));
  }
}

static void
SendHolding(tree_t *t, int partner){
  int i;

  Msg("merge", ("Send count=%d holding to partner\n", count));
  MPMY_send(&count, sizeof(int), partner, 300);
  for(i=0; i<count; i++){
    int jj;
    jj = HIDX(t, count-1-i);
    MPMY_send(&holding[jj], sizeof(*holding), partner, 301);
    MPMY_send(holding[jj].tb, holding[jj].sz, partner, 302);
  }
}

static void
mergeHi(tree_t *t, int partner){
  int i, j;
  tc_tmp *tc;
  sg_tmp *outersg, *innersg;
  sg_tmp *sg, *sgin, *next;
  sg_tmp *outersgs[ KEYBITS ];
  sg_tmp *incomingouterparent;
  sg_tmp *lastoutersend;
  int nout;
  int bit;

  if( firstsg == NULL) {
    innersg = &rootsgtmp;
    assert(curr_grp = &rootsgtmp);
    for(i=0; i<count; i++){
      j = HIDX(t, (count-1)-i);
      orphans[norphans++] = holding[j];
    }
    assert(norphans <= m );
    Msg("merge", ("firstsg == NULL in mergeHi, copied %d held bods orphans\n",
		  count));
    Msg("merge", ("norphans = %d\n", norphans));
    count = 0;
  }else{
    innersg = firstsg;
  }

  /* send orphans to partner */
  Msg("merge", ("Sending norphans=%d\n", norphans));
  MPMY_send(&norphans, sizeof(int), partner, 123);
  MPMY_send(orphans, (norphans)*sizeof(*orphans), partner, 124);
  for(i=0; i<norphans; i++){
    Msg("merge", ("Send orphan with key %s\n", PrintKey(orphans[i].key)));
    MPMY_send(orphans[i].tb, orphans[i].sz, partner, 125+i);
  }
  /* Tell partner our inner key, so he can adjust to match it. */
  MPMY_send(&innersg->key, sizeof(Key_t), partner, 125+i);

  outersg = curr_grp;
  Msg("merge", ("mHi, outersg=curr_grp=%s\n", PrintSgtmp(curr_grp)));
  if( firstsg == NULL ){
    /* We have nothing at all.  We must create enough of a framework
       that we can incorporate the cells that will be sent from our
       partner.  These will become our upper boundary.  Don't forget
       about 'holding' */
    Key_t hisinnerkey;
    int hislevel;
    int flag, bit;


    /* If we represent all processors through MPMY_Nproc()-1, then
       set COMPLETE_HI */
    bit = partner ^ MPMY_Procnum();
    flag = ((MPMY_Procnum()+bit)>=MPMY_Nproc()) ? COMPLETE_HI : 0;
    Msg("merge", ("firsgsg==NULL in mergeHi, drilling down\n"));

    MPMY_recvn(&hisinnerkey, sizeof(Key_t), partner, 140);
    hislevel = TreeLevel(hisinnerkey, ndim);
    Msg("merge", ("curr_level=%d, hisinnerkey=%s, lev=%d\n", 
		  curr_level, PrintKey(hisinnerkey), hislevel));
    while( curr_level++ < hislevel ){
      sg_tmp *new;

      new = GrpAlloc(t);
      rshift -= ndim;
      new->key = KeyRshift(hisinnerkey, (hislevel-curr_level)*ndim);
      tc = &curr_grp->tc_tmps[0];
      tc->sibs = curr_grp;
      tc->daughter = new;
      new->parent = tc;
      tc->bits = KeyAndInt(new->key, ndimmask);
      new->flags = 0;		/* flag? */
      curr_grp->ncell++;
      next_tc_tmp = &curr_grp->tc_tmps[1];
      Msg("merge", ("created new daughter of %s\n", PrintSgtmp(curr_grp)));
      chkSgtmp(t, curr_grp);
      curr_grp = new;
      Msg("merge", ("We now have a firstsg %s\n", PrintSgtmp(curr_grp)));
      firstsg = curr_grp;
    }
    --curr_level;
    prevkey = KeyLshift( hisinnerkey, ndim * (maxlevs - curr_level));
    Msg("merge", ("Setting prevkey=%s\n", PrintKey(prevkey)));
    innersg = curr_grp;
    Msg("merge", ("New innerkey = %s\n", PrintSgtmp(curr_grp)));
    
    RecvHolding(t, partner);
    curr_child = KeyAndInt( KeyRshift(prevkey, rshift-ndim), ndimmask );
  }
    
  Msg("merge", ("myinnerkey=%s, ", PrintKey(innersg->key)));
    
  sg = innersg;
  Clearqueue();
  Msg("merge", ("Clearq done\n"));
  while(sg->flags){
    Enqueue(t, sg);
    sg = sg->parent->sibs;
  }
  /* Extra enqueue to gather up completed daughters */
  Msg("merge", ("Enqueue last sg\n"));
  Enqueue(t, sg);
  lastoutersend = sg;

  Exchg(partner);
  /* Merge the incoming sibgroups with those on our own inner boundary. */
  next = innersg;
  while((sgin = Dequeue(t))){
    sg = next;
    next = (sg->parent)? sg->parent->sibs : sg;
    Msg("merge", ("Call MergeGroups(from=%s, ", PrintSgtmp(sgin)));
    Msg("merge", ("%s)\n", PrintSgtmp(sg)));
    MergeGroups(t, sgin, sg, 1);

    if( Completep(sg) ){
      callAggregate(t, sg);
      GrpScrunch(t, sg, /*???*/0);  /* trashes sg! */
      if( curr_grp == sg ){
	Msg("merge", ("Bumping outersg up to %s\n", PrintSgtmp(next)));
	curr_grp = next;
	curr_level--;
      }
    }else{
      Msg("merge", ("Defer Merge flags=%d\n", sg->flags));
    }
  }
  incomingouterparent = sg;

    /* Then enqueue the 'outer' boundary.  It's easier to locate them
     in bottom-to-top order, but we really want them in top-to-bottom. */
  Msg("merge", ("Starting descent of outer branches.\n"));
  Msg("merge", ("curr_grp: %s\n", PrintSgtmp(curr_grp)));
  Msg("merge", ("incomingouterparent: %s\n", PrintSgtmp(incomingouterparent)));
  nout = 0;
  sg = curr_grp;
  while( sg != lastoutersend ){
    outersgs[nout++] = sg;
    sg = sg->parent->sibs;
  }
  Clearqueue();
  /* Enqueue them from top to bottom */
  while(nout--){
    Enqueue(t, outersgs[nout]);
  }
  Exchg(partner);
  /* Now insert the groups that make up the new outside border */
  sg = incomingouterparent;
  while( (sgin=Dequeue(t)) ){
    if( !KeyEQ(sg->key, KeyRshift(sgin->key, ndim)) ){
      Msg_do("sg->key: %s\n", PrintKey(sg->key));
      Msg_do("sgin->key: %s\n", PrintKey(sgin->key));
      Error( "Failed assertion KeyEQ(sg->key, KeyRshif(sgin->key, ndim))");
    }

    if( sg == &rootsgtmp ){
      Msg("merge", ("Don't merge another daugher of 0x0!\n"));
      sg = sg->tc_tmps[0].daughter;
      continue;
    }

    Msg("merge", ("Shifting %d tcells upwards in %s\n", sg->ncell, 
		  PrintSgtmp(sg)));
    for(i=sg->ncell; i>0; --i){
      tc_tmp *totc = &sg->tc_tmps[i];
      *totc = *(totc-1);
      if( totc->daughter )
	totc->daughter->parent = totc;
      Msg("merge", ("tc_tmps[%d].bits = %d\n", i, totc->bits));
    }

    sg->ncell++;
    tc = &sg->tc_tmps[0];
    sgin->parent = tc;
    tc->sibs = sg;
    tc->daughter = sgin;
    tc->bits = KeyAndInt(sgin->key, ndimmask);
    sg = sgin;
  }

  bit = MPMY_Procnum()^partner;
  if( sg != &rootsgtmp ){
    Msg("merge", ("New firstsg on lower boundary, %s\n",
		  PrintSgtmp(sg)));
    firstsg = sg;
  }else if( (MPMY_Procnum()- 2*bit ) <= 0 ){
    Msg("merge", ("Setting firstsg=rootsgtmp, procnum=%d, bit=%d\n",
		  MPMY_Procnum(), bit));
    firstsg = &rootsgtmp;
  }

  /* Get orphans from partner */
  Msg("merge", ("mHI: done with outer branch, firstsg = %s\n", PrintSgtmp(firstsg)));
  MPMY_recvn(&norphans, sizeof(int), partner, 200);
  Msg("merge", ("receiving norphans=%d\n", norphans));
  MPMY_recvn(orphans, norphans*sizeof(*orphans), partner, 201);
  Msg("merge", ("received orphans=%d\n", norphans));
  for(i=0; i<norphans; i++){
    Msg("merge", ("orphans[%d].sz=%ld, key=%s\n", 
		  i, (long)orphans[i].sz, PrintKey(orphans[i].key) ));
    orphans[i].tb = Malloc(orphans[i].sz);
    MPMY_recvn(orphans[i].tb, orphans[i].sz, partner, 202+i);
  }
  if( firstsg == NULL ){
    prevkey = orphans[norphans-1].key;
    Msg("merge", ("firstsg==NULL in mHI, setting prevkey from last orphan: %s\n",
	PrintKey(prevkey)));
  }
  /* Send holding to partner */
  SendHolding(t, partner);

  /* Fix up state vars. */
  next_tc_tmp = &curr_grp->tc_tmps[curr_grp->ncell];
  Msg("merge", ("Finishing MergeHi with count=%d, norphans=%d\n", count, norphans));
  Msg("merge", ("firstsg=%s\n", PrintSgtmp(firstsg)));
  Msg("merge", ("curr_grp=%s\n", PrintSgtmp(curr_grp)));
}


static void
mergeLo(tree_t *t, int partner){
  int i;
  tc_tmp *tc;
  sg_tmp *outersg, *innersg;
  sg_tmp *sg, *sgin, *next;
  sg_tmp *outersgs[ KEYBITS ];
  sg_tmp *incomingouterparent;
  sg_tmp *lastoutersend;
  int nout;
  Key_t hisinnerkey;
  int newnorphans;
  struct body_info *neworphans;
  int hislevel, hisnewlevel; 

  /* Receive orphans from partner */

  MPMY_recvn(&newnorphans, sizeof(int), partner, 123);
  neworphans = Malloc((newnorphans)*sizeof(*neworphans));
  MPMY_recvn(neworphans, (newnorphans)*sizeof(*neworphans), partner, 124);
  Msg("merge", ("receiving %d orphans\n", newnorphans));
  for(i=0; i<newnorphans; i++){
    neworphans[i].tb = Malloc(neworphans[i].sz);
    MPMY_recvn(neworphans[i].tb, neworphans[i].sz, partner, 125+i);
  }
  MPMY_recvn(&hisinnerkey, sizeof(Key_t), partner, 125+i);

  /* Call TreeAddBody_ on the incoming data */
  for(i=0; i<newnorphans; i++){
    holding[holdidx] = neworphans[i];
    if( ++holdidx >= holdingsz )
      holdidx = 0;
    Msg("merge", ("\nMerging orphan %d, k=%s\n", i, PrintKey(neworphans[i].key)));
    TreeBuildAdd_(t, 0);
  }

  Free(neworphans);		/* alloca? */
  innersg = curr_grp;
    
  Msg("merge", ("innersg->key=%s, ", PrintKey(innersg->key)));
  Msg("merge", ("hisinnerkey=%s\n", PrintKey(hisinnerkey)));
  /* Now make sure my inner key and my partners are the same. 
     If I am deeper, then I can finish off cells and 'peel' back the stack.
     If I am shallower, then I need to create empty cells that will be
       filled in by the MergeTrees sequence below. */
  if( KeyEQ( hisinnerkey, KeyInt(0x0) )){
    Msg("merge", ("firstsg apparently unset in partner\n"));
    Msg("merge", ("setting hislevel=-1, will adjust later\n"));
    sg = curr_grp;
    hislevel = -1;
  }else{
    hislevel = TreeLevel(hisinnerkey, ndim);
  }

  while( hislevel > curr_level ){
    /* push down */
    sg_tmp *new_grp;
    int bchild, oldcount;
    
    Msg("merge", ("Level mismatch:  hislevel=%d, curr_level=%d, count=%d -- push\n", hislevel, curr_level, count));
    /* Code lifted from TreeBuildAdd_() */
    bchild = KeyAndInt( KeyRshift(hisinnerkey, ndim*((hislevel-curr_level)-1)), ndimmask);
    Msg("merge", ("create new child with bchild=%d in %s\n", bchild, 
		  PrintSgtmp(curr_grp)));
    curr_level++;
    rshift -= ndim;
    curr_child = CHILD_IMPOSSIBLE;
    new_grp = GrpAlloc(t);
    if( firstsg )
      new_grp->flags |= COMPLETE_LO;
    new_grp->parent = next_tc_tmp;
    new_grp->key = KeyOrInt(KeyLshift(curr_grp->key, ndim), bchild);
    next_tc_tmp->bits = bchild;
    next_tc_tmp->daughter = new_grp;
    next_tc_tmp->sibs = curr_grp;
    curr_grp->ncell++;
    chkSgtmp(t, curr_grp);
    curr_grp = new_grp;
    next_tc_tmp = curr_grp->tc_tmps;

    /* Retry the guys in the 'holding' array */
    if( count > 0 ){
      prevkey = Key0;		/* don't flag a false out-of-order error! */
      oldcount = count;
      count = 0;
      Msg("merge", ("attempt to store holding bodies into %s\n",
		    PrintSgtmp(curr_grp)));
      for(i=0; i<oldcount; i++){
	TreeBuildAdd_(t, (oldcount-i)-1);
      }
    }
    chkSgtmp(t, curr_grp);
  }

  if( hislevel < 0 ){
    /* Our partner has no bodies at all!  He will create an empty scaffold 
       that we can merge with */
    MPMY_send(&curr_grp->key, sizeof(Key_t), partner, 140);
    SendHolding(t, partner);
    hisnewlevel = curr_level;
    Msg("merge", ("Sending curr_grp->key=%s to partner, his new level is %d\n",
		  PrintKey(curr_grp->key), curr_level));
  }else{
    hisnewlevel = hislevel;
  }

  if( hisnewlevel < curr_level ){
    Msg("merge", ("\nLevel mismatch: hisnewlevel=%d, curr_level=%d\n", hisnewlevel, curr_level));
    if( firstsg == NULL ){
      Msg("merge", ("firstsg NULL in mergeLo.  %d more orphans\n", count));
      for(i=0; i<count; i++){
	int j;
	j = HIDX(t, (count-1)-i);
	orphans[norphans++] = holding[j];
      }
      assert(norphans <= m);
      Msg("merge", ("norphans = %d\n", norphans));
    }else{
      Msg("merge", ("Mlo: Attaching %d remaining bods to curr_grp\n", count));
      chkSgtmp(t, curr_grp);
      for(i=0; i<count; i++){
	int j;
	j = HIDX(t, (count-1)-i);
	curr_grp->tbodyptrs[curr_grp->ntbody] = holding[j].tb;
	curr_grp->tbodysz[curr_grp->ntbody] = holding[j].sz;
	curr_grp->ntbody++;
      }
    }
    chkSgtmp(t, curr_grp);
    count = 0;
  }

  while( hisnewlevel < curr_level ){
    sg_tmp *up;
    up = curr_grp->parent->sibs;
    curr_grp->flags |= COMPLETE_HI;
    chkSgtmp(t, curr_grp);
    if( Completep(curr_grp) ){
      Msgf(("Scrunching grp at level %d, key=%s\n", curr_level, 
	    PrintKey(curr_grp->key)));
      callAggregate(t, curr_grp);
      GrpScrunch(t, curr_grp, curr_level);
    }else{
      Msgf(("Deferring scrunch of grp key=%s, flags=%d\n",
	    PrintKey(curr_grp->key), curr_grp->flags));
    }
    curr_grp = up;
    --curr_level;
  }

  /* Our partner has a bona fide 'firstsg', so we can safely
     attach the holding bodies to the curr_grp. */
  if( hislevel >= 0 ){
    if( firstsg == NULL ) {
      for(i=0; i<count; i++){
	int j;
	j = HIDX(t, (count-1)-i);
	orphans[norphans++] = holding[j];
      }
      assert(norphans <= m);
      firstsg = curr_grp;
      Msg("merge", ("firstsg NULL in mergeLo, now %s.  %d more orphans\n", PrintSgtmp(firstsg), count));
      Msg("merge", ("norphans = %d\n", norphans));
    }else{
      Msg("merge", ("Mlo: Attaching %d remaining bods to curr_grp: %s\n", 
		    count, PrintSgtmp(curr_grp)));
      chkSgtmp(t, curr_grp);
      for(i=0; i<count; i++){
	int j;
	j = HIDX(t, (count-1)-i);
	curr_grp->tbodyptrs[curr_grp->ntbody] = holding[j].tb;
	curr_grp->tbodysz[curr_grp->ntbody] = holding[j].sz;
	curr_grp->ntbody++;
      }
      chkSgtmp(t, curr_grp);
    }
    count = 0;
  }

  innersg = curr_grp;
  chkSgtmp(t, curr_grp);

  sg = innersg;
  Clearqueue();
  Msg("merge", ("Clearq done\n"));
  while(sg->flags){
    Enqueue(t, sg);
    sg = sg->parent->sibs;
  }
  /* Extra enqueue to gather up completed daughters */
  Enqueue(t, sg);
  lastoutersend = sg;

  if( firstsg == NULL ){
    Msg("merge", ("firstsg==NULL, setting outersg = curr_grp in mergeLo\n"));
    outersg = curr_grp;	/* ?? */
  }else{
    outersg = firstsg;
  }

  Exchg(partner);
  /* Merge the incoming sibgroups with those on our own inner boundary. */
  next = innersg;
  while((sgin = Dequeue(t))){
    sg = next;
    next = (sg->parent)? sg->parent->sibs : NULL;
    Msg("merge", ("Call MergeGroups(from=%s, ", PrintSgtmp(sgin)));
    Msg("merge", ("to=%s)\n", PrintSgtmp(sg)));
    MergeGroups(t, sgin, sg, 0);

    if( Completep(sg) ){
      callAggregate(t, sg);
      GrpScrunch(t, sg, /*???*/0);  /* trashes sg! */
      if( outersg == sg ){
	Msg("merge", ("Bumping outersg up to %s\n", PrintSgtmp(next)));
	outersg = next;
      }
    }else{
      Msg("merge", ("Defer Merge flags=%d\n", sg->flags));
    }
  }
  incomingouterparent = sg;

    /* Then enqueue the 'outer' boundary.  It's easier to locate them
     in bottom-to-top order, but we really want them in top-to-bottom. */
  Msg("merge", ("Starting descent of outer branches.\n"));
  Msg("merge", ("outersg: %s\n", PrintSgtmp(outersg)));
  Msg("merge", ("incomingouterparent: %s\n", PrintSgtmp(incomingouterparent)));
  Msg("merge", ("lastoutersend: %s\n", PrintSgtmp(lastoutersend)));
  nout = 0;
  sg = outersg;
  while( sg != lastoutersend ){
    outersgs[nout++] = sg;
    sg = sg->parent->sibs;
  }
  Clearqueue();
  /* Enqueue them from top to bottom */
  Msg("merge", ("Enqueuing %d outer sgs\n", nout));
  while(nout--){
    Enqueue(t, outersgs[nout]);
  }
  Exchg(partner);
  /* Now insert the groups that make up the new outside border */
  sg = incomingouterparent;
  while( (sgin=Dequeue(t)) ){
    int insert;
    if( !KeyEQ(sg->key, KeyRshift(sgin->key, ndim)) ){
      Msg_do("sg->key: %s\n", PrintKey(sg->key));
      Msg_do("sgin->key: %s\n", PrintKey(sgin->key));
      Error( "Failed assertion KeyEQ(sg->key, KeyRshif(sgin->key, ndim))");
    }

    if( sg == &rootsgtmp ){
      Msg("merge", ("Don't merge another daugher of 0x0!\n"));
      sg = sg->tc_tmps[0].daughter;
      continue;
    }
    insert = sg->ncell;
    sg->ncell++;
    tc = &sg->tc_tmps[insert];
    sgin->parent = tc;
    tc->sibs = sg;
    tc->daughter = sgin;
    tc->bits = KeyAndInt(sgin->key, ndimmask);
    sg = sgin;
  }

  /* Send orphans to partner */
  Msg("merge", ("Sending %d orphans to partner\n", norphans));
  MPMY_send(&norphans, sizeof(int), partner, 200);
  MPMY_send(orphans, norphans*sizeof(*orphans), partner, 201);
  for(i=0; i<norphans; i++){
    MPMY_send(orphans[i].tb, orphans[i].sz, partner, 202+i);
  }

  /* Get 'holding' from partner */
  RecvHolding(t, partner);

  curr_grp = sg;
  Msg("merge", ("Fixing state variables\n"));
  Msg("merge", ("New prevkey from holding array: %s\n", PrintKey(prevkey)));
  Msg("merge", ("fixing state vars\nprevkey=%s ", PrintKey(prevkey)));
  Msg("merge", ("curr_grp=%s\n", PrintSgtmp(curr_grp)));
  Msg("merge", ("curr_level = %d\n", curr_level));
  Msg("merge", ("rshift=%d, prevkey>>rshift = %s\n", rshift,
		PrintKey(KeyRshift(prevkey, rshift))));
  chkSgtmp(t, curr_grp);
  curr_level = TreeLevel(curr_grp->key, ndim);
  rshift = ndim * (maxlevs - curr_level);
  next_tc_tmp = &curr_grp->tc_tmps[curr_grp->ncell];
  curr_child =  KeyAndInt( KeyRshift(prevkey, rshift-ndim), ndimmask );

  Msg("merge", ("Finishing MergeLo with count=%d, norphans=%d\n", count, norphans));
  Msg("merge", ("firstsg=%s\n", PrintSgtmp(firstsg)));
  Msg("merge", ("curr_grp=%s\n", PrintSgtmp(curr_grp)));
}

static void
MergeParallel(tree_t *t){
  int bit, IamHi, partner;

  Msg("merge", ("MergeParallel:\n"));

  for(bit=1; bit<MPMY_Nproc(); bit<<=1){
    IamHi = bit & MPMY_Procnum();
    partner = bit ^ MPMY_Procnum();
    Msg("merge", ("Bit=%d\nfirstsg=%s\n", bit, PrintSgtmp(firstsg)));
    Msg("merge", ("curr_grp=%s\n", PrintSgtmp(curr_grp)));
    Msg("merge", ("count=%d, IamHi=%d, norphans=%d\n", count, IamHi, norphans));
    if( IamHi ){
      mergeHi(t, partner);
    }else{
      mergeLo(t, partner);
    }
  }
  Msg("merge", ("Returning from MergeParallel...Are we done!?\n"));
}

#ifdef DEBUG

static void
chkSgtmp_(int lineno, tree_t *t, sg_tmp *sg){
  Key_t *bkeys = alloca(sg->ntbody * sizeof(Key_t));
  Key_t ckey = sg->key;
  int lobitslast;
  int nlobits;
  int lev, rsh, i, j;

  lev = TreeLevel(ckey, ndim);
  if( lev < 0 && KeyNEQ(ckey, KeyInt(0x0) )){
    SeriousWarning("bad key in chkSgtmp(sg=%s)\n", PrintSgtmp(sg));
    return;
  }

  if( lev>=0 && KeyNEQ( KeyRshift(ckey, lev*ndim) , KeyInt(0x1)) ){
    SeriousWarning("bad key in lev=%d, chkSgtmp(sg=%s)\n",
		   lev, PrintSgtmp(sg));
    return;
  }
  
  if( tree17keyFromTbody ){
    for(i=0; i<sg->ntbody; i++){
      bkeys[i] = (*tree17keyFromTbody)(sg->tbodyptrs[i]);
    }
  
    rsh = ndim * ( ((KEYBITS-1)/(ndim)) - lev );
    rsh -= ndim;		/* (leave the low ndim bits) */
    nlobits = 0;
    lobitslast = -1;
    Msgf(("Line %d chkSgtmp(sg=%s)\n", lineno, PrintSgtmp(sg)));

    for(i=0; i<sg->ntbody; i++){
      Key_t bkey;
      unsigned int lobits;

      bkey = KeyRshift(bkeys[i], rsh);
      lobits = KeyAndInt(bkey, (1<<ndim)-1);

      if( lobits == lobitslast ){
	nlobits++;
      }else{
	nlobits = 0;
      }
      if( nlobits > m ){
	SeriousWarning("FAILED chkSgtmp(sg=%s)\n", PrintSgtmp(sg));
	Shout("Too many keys with same lo bits, i=%d\n", i);
      }

      bkey = KeyRshift(bkey, ndim);
      if( !KeyEQ(ckey, bkey) ){
	SeriousWarning("FAILED chkSgtmp(sg=%s)\n", PrintSgtmp(sg));
	Shout("bkey[%d] %s outside\n", i, PrintKey(bkeys[i]));
      }

      if( i>0 && KeyLE(bkeys[i], bkeys[i-1]) ){
	SeriousWarning("FAILED chkSgtmp(sg=%s)\n", PrintSgtmp(sg));
	Shout("bkey[%d] %s out of order\n", i, PrintKey(bkeys[i]));
      }

      for(j=0; j<sg->ncell; j++){
	if( sg->tc_tmps[j].bits == lobits ){
	  SeriousWarning("FAILED chkSgtmp(sg=%s)\n", PrintSgtmp(sg));
	  Shout("bkey[%d] %s should be in tcells[%d]\n", 
		i, PrintKey(bkeys[i]), j);
	}
      }
    }
  }

  for(j=0; j<sg->ncell; j++){
    unsigned int bits;
    bits = sg->tc_tmps[j].bits;
    if( j>0 && bits <= sg->tc_tmps[j-1].bits ){
      SeriousWarning("FAILED chkSgtmp(sg=%s)\n", PrintSgtmp(sg));
      Shout("tcells bits[%d]=%d out of order (prev=%d)\n", 
	    j, bits, sg->tc_tmps[j-1].bits);
    }
    if( bits >= (1<<ndim) ){
      SeriousWarning("FAILED chkSgtmp(sg=%s)\n", PrintSgtmp(sg));
      Shout("tcells bits[%d] = %d out of range\n", j, bits);
    }
  }
}

#endif /* DEBUG */

#undef rootsg
#undef ndim
#undef m
#undef curr_child
#undef curr_grp
#undef rootsgtmp
#undef firstsg
#undef curr_level
#undef maxlevs
#undef next_tc_tmp
#undef rshift
#undef ndimmask
#undef count
#undef prevkey
#undef Bfuncs
#undef norphans
#undef holdidx
#undef holdingsz
#undef holding
#undef orphans
#undef sgchn
#undef CofmChn
#undef npages
#undef maxpages
#undef pages

/* Some diagnostic tools */
#ifndef NO_MSGS
static char *PrintTctmp(const tc_tmp *tc){
  static char ret[512];
  if( tc==NULL ){
    sprintf(ret, "%p", tc);
  }else{
    sprintf(ret, "%p: contents=%p, key=%s.%o, sibs=%p, daughter=%p",
	    tc, tc->contents, 
	    (tc->sibs)? PrintKey(tc->sibs->key):"(null)", 
	    tc->bits, tc->sibs, tc->daughter);
  }
  return ret;
}
#endif

static char *PrintOOC(oocptr p){
  static char statbuf[128];
  sprintf(statbuf, "(%s+%d)", PrintPgId(p.pgid), p.offset);
  return statbuf;
}

static char *PrintSgtmp(const sg_tmp *sg){
  static char ret[512];
  if( sg != NULL ){
    sprintf(ret, "sg@%p:  key=%s, ncells=%d, nbodies=%d, flags=%d",
	    sg, PrintKey(sg->key), sg->ncell, sg->ntbody, sg->flags);
  }else{
    sprintf(ret, "sg@0x0");
  }
  return ret;
}

char *PrintTc(const tcell *tc){
  static char ret[512];
  if( tc==NULL ){
    sprintf(ret, "%p", tc);
  }else{
    sprintf(ret, "%p: cell=%p, bits=%o, daughter=%s",
	    tc, Contents(tc), 
	    tc->bits, PrintOOC(tc->daughter));
  }
  return ret;
}

char *PrintSg(const sib_group *sg){
  static char ret[512];
  if( sg != NULL ){
    sprintf(ret, "sg@%p:  key=%s, ncells=%d, nbodies=%d",
	    sg, PrintKey(sg->key), sg->ncell, sg->ntbody);
  }else{
    sprintf(ret, "sg@0x0");
  }
  return ret;
}

void
TreeChkSibGroup(const sib_group *sg, const Key_t *bkeys, int ndim_, int m_){
  Key_t ckey = sg->key;
  int lobitslast;
  int nlobits;
  int lev, rshift_, i, j;

  lev = TreeLevel(ckey, ndim_);
  rshift_ = ndim_ * ( ((KEYBITS-1)/(ndim_)) - lev );
  rshift_ -= ndim_;		/* (leave the low ndim_ bits) */
  
  nlobits = 0;
  lobitslast = -1;
  Msgf(("chkVisitBods(sg->key=%s)\n", PrintKey(ckey)));

  for(i=0; i<sg->ntbody; i++){
    Key_t bkey;
    unsigned int lobits;

    bkey = KeyRshift(bkeys[i], rshift_);
    lobits = KeyAndInt(bkey, (1<<ndim_)-1);

    if( lobits == lobitslast ){
      nlobits++;
    }else{
      nlobits = 0;
    }
    if( nlobits > m_ ){
      Shout("FAILED chkVisitBods(sg->key=%s)\n", PrintKey(ckey));
      Error("Too many keys with same lo bits, i=%d\n", i);
    }

    bkey = KeyRshift(bkey, ndim_);
    if( !KeyEQ(ckey, bkey) ){
      Shout("FAILED chkVisitBods(sg->key=%s)\n", PrintKey(ckey));
      Error("bkey[%d] %s outside\n", i, PrintKey(bkeys[i]));
    }

    if( i>0 && KeyLE(bkeys[i], bkeys[i-1]) ){
      Shout("FAILED chkVisitBods(sg->key=%s)\n", PrintKey(ckey));
      Error("bkey[%d] %s out of order\n", i, PrintKey(bkeys[i]));
    }

    for(j=0; j<sg->ncell; j++){
      if( sg->tcells[j].bits == lobits ){
	Shout("FAILED chkVisitBods(sg->key=%s)\n", PrintKey(ckey));
	Error("bkey[%d] %s should be in tcells[%d]\n", 
	      i, PrintKey(bkeys[i]), j);
      }
    }
  }

  for(j=0; j<sg->ncell; j++){
    unsigned int bits;
    bits = sg->tcells[j].bits;

    if( j>0 && bits < sg->tcells[j-1].bits ){
      Shout("FAILED chkVisitBods(sg->key=%s)\n", PrintKey(ckey));
      Error("tcells bits[%d]=%d out of order, prev=%d\n", 
	    j, bits, sg->tcells[j-1].bits);
    }
    if( bits >= (1<<ndim_) ){
      SeriousWarning("FAILED chkSgtmp(sg->key=%s)\n", PrintKey(ckey));
      Error("tcells bits[%d]=%d out of range\n", j, bits);
    }
  }
}

