/* A non-plug-compatible replacement for physics_generic.[ch].  Use with
   care, but then, you were using physics_generic.c with care, weren't
   you? */

#include <stddef.h>
#include <float.h>
#include "protos.h"
#include "mpmy.h"
#include "Msgs.h"
#include "Assert.h"
#include "key.h"
#include "keycvt.h"

#ifndef FLT_MAX
/* I wonder what they put in float.h, anyway */
#define FLT_MAX 1.e38
#endif

int KeyOutOfBounds;

void MakeBbox(tbbox *bb, int ndim, float *min, float *max){
  int d;
  bb->ndim = ndim;
  for(d=0; d<ndim; d++){
    bb->rmin[d] = min[d];
    bb->sz[d] = max[d] - min[d];
  }
}

void CornersBbox(const tbbox *bb, float *min, float *max){
  int d;
  for(d=0; d<bb->ndim; d++){
    min[d] = bb->rmin[d];
    max[d] = bb->sz[d] + bb->rmin[d];
  }
}  

/* This gives a tight bounding box around the list of positions.  Note
   that the result has rmin <= pos[i] and rmax > pos[i].  The equality
   in rmin shouldn't cause any headaches for conversions, but if we
   allowed equality on the high end, we'd have problems with float-to-int
   conversions.
*/
void
TightBbox(const float *pstart, int nobj, int pstride, int ndim, tbbox *bb){
    int d;
    const float *pos = pstart;
    float rmin[MAXNDIMKU];
    float rmax[MAXNDIMKU];

    assert(ndim < MAXNDIMKU);

    for(d=0; d<ndim; d++){
	rmin[d] = FLT_MAX;
	rmax[d] = -FLT_MAX;
    }
    bb->ndim = ndim;
    while(nobj--){
	for(d=0; d<ndim; d++){
	    float pd = pos[d];
	    if( rmin[d] > pd ) rmin[d] = pd;
	    if( rmax[d] < pd ) rmax[d] = pd;
	}
	pos = (float *)(pstride + (char *)pos);
    }
    MakeBbox(bb, ndim, rmin, rmax);
    for(d=0; d<ndim; d++){
      bb->sz[d] *= (1.F + FLT_EPSILON);
    }
}

void
MPMYCombineBbox(tbbox *bb){
  MPMY_Comm_request req;
  float rmin[MAXNDIMKU];
  float rmax[MAXNDIMKU];
  int d;

  MPMY_ICombine_Init(&req);
  for(d=0; d<bb->ndim; d++){
    rmin[d] = bb->rmin[d];
    rmax[d] = bb->rmin[d]+bb->sz[d];
  }
  MPMY_ICombine(rmin, bb->rmin, bb->ndim, MPMY_FLOAT, MPMY_MIN, req);
  MPMY_ICombine(rmax, bb->sz, bb->ndim, MPMY_FLOAT, MPMY_MAX, req);
  for(d=0; d<bb->ndim; d++){
    bb->sz[d] -= bb->rmin[d];
  }
  MPMY_ICombine_Wait(req);
}

void 
CenterBbox(const tbbox *bb, float *center, float *sz){
    int d;
    for(d=0; d<bb->ndim; d++){
	center[d] = bb->rmin[d] + 0.5F*bb->sz[d];
	sz[d] = bb->sz[d];
    }
}

int 
ContainsBbox(const tbbox *bb1, const tbbox *bb2){
    /* Return 1 if bb1 completely contains bb2 */
    int d;

    for(d=0; d<bb1->ndim; d++){
	if( bb1->rmin[d] > bb2->rmin[d] ||
	   bb1->rmin[d] + bb1->sz[d] < bb2->rmin[d]+bb2->sz[d] )
	    return 0;
    }
    return 1;
}

/* Construct bbu, the 'union' of bb1 and bb2 */
void
UnionBbox(const tbbox *bb1, const tbbox *bb2, tbbox *bbu){
    int d;

    bbu->ndim = bb1->ndim;
    for(d=0; d<bb1->ndim; d++){
	float max1, max2;
	float min1, min2;

	/* Read these first in case bbu is equal to bb1 or bb2! */
	min1 = bb1->rmin[d];
	min2 = bb2->rmin[d];
	bbu->rmin[d] = (min1 < min2) ? min1 : min2;
	max1 = min1 + bb1->sz[d];
	max2 = min2 + bb2->sz[d];
	bbu->sz[d] = (max1 > max2) ? max1 - bbu->rmin[d] : max2 - bbu->rmin[d];
    }
}

/* Make the bbox a cube by expanding the smaller dimensions. */
void
CubeBbox(tbbox *bb){
    int d;
    float maxs, halfmaxs;
    float center[MAXNDIMKU];

    maxs = 0.;
    for(d=0; d<bb->ndim; d++) {
	center[d] = bb->rmin[d] + 0.5 * bb->sz[d];
	if( maxs < bb->sz[d] ) 
	    maxs = bb->sz[d];
    }
    halfmaxs = 0.5*maxs;
    for( d=0; d<bb->ndim; d++){
	bb->rmin[d] = center[d] - halfmaxs;
	bb->sz[d] = maxs;
    }
}

/* Increase the linear dimension by 'factor' on all sides */
void
InflateBbox(tbbox *bb, float factor){
    float center;
    int d;
    for( d=0; d<bb->ndim; d++){
	center = 0.5*bb->sz[d] + bb->rmin[d];
	bb->sz[d] *= factor;
	bb->rmin[d] = center - 0.5*bb->sz[d];
    }
}

int
GenerateKeys(const float *pstart, int nobj, int pstride, tbbox *bb, 
	     Key_t *kstart, int kstride, enum KeyOrder_e order_type){
    int i, d;
    float keyfactor[MAXNDIMKU];
    unsigned int ik[MAXNDIMKU];
    unsigned int chubits;
    Key_t *kp;
    const float *pos;
    unsigned int maxikey;
    float fmaxikey;
    Key_t (*KfI)(unsigned int *xp, int ndim, int nbits);
    int ret;

    chubits = ((KEYBITS-1) / bb->ndim );
    maxikey = (1U<<chubits) - 1;
    fmaxikey = (float)maxikey;
    for(d=0; d<bb->ndim; d++){
	/* Divide by 0 ?? */
	keyfactor[d] = (1U<<chubits)/(bb->sz[d]);
    }
    pos = pstart;
    kp = kstart;
    switch( order_type ){
    case MORTON_ORDER:
      KfI = &KeyFromInts;
      break;
    case PH_ORDER:
      KfI = &PHKeyFromInts;
      break;
    default:
      Error("Unrecognized order_type in GenerateKeys\n");
    }

    ret = 0;
    for(i=0; i<nobj; i++){
	for(d=0; d<bb->ndim; d++){
	    float fk = keyfactor[d] * (pos[d] - bb->rmin[d]);
	    if( fk < 0. ){
	      ret = -1;
	      ik[d] = 0;
	    }else if( fk > fmaxikey ){
	      ik[d] = maxikey;
	      ret = -1;
	    }else
	      ik[d] = (unsigned int)fk;
	}
	*kp = (*KfI)(ik, bb->ndim, chubits);
	kp = (Key_t *) ( kstride + (char *)kp);
	pos = (float *) ( pstride + (char *)pos);
    }
    return ret;
}

Key_t
KeyFromFloats(const float *pos, const tbbox *universe, int level, enum KeyOrder_e order_type){
    unsigned int ik[MAXNDIMKU];
    Key_t (*KfI)(unsigned int *xp, int ndim, int nbits);
    int ok;

    switch( order_type ){
    case MORTON_ORDER:
      KfI = &KeyFromInts;
      break;
    case PH_ORDER:
      KfI = &PHKeyFromInts;
      break;
    default:
      Error("Unrecognized order_type in GenerateKeys\n");
    }

    ok = IntsFromFloats(pos, ik, universe, level);
    if( ok < 0 )
      return KeyInt(0x0);

    return (*KfI)(ik, universe->ndim, level);
}

void 
CellBBFromKey(Key_t key, tbbox *bb, tbbox *cellbb, enum KeyOrder_e order_type)
{
    unsigned int icorner[MAXNDIMKU];
    unsigned int iscale;
    float factor;
    int d;

    switch(order_type){
    case MORTON_ORDER:
      iscale = (1<<IntsFromKey(key, icorner, bb->ndim));
      break;
    case PH_ORDER:
      iscale = (1<<IntsFromPHKey(key, icorner, bb->ndim));
      break;
    default:
      Error("Unrecognized ordering in CellBBFromKey\n");
    }
      
    /* Now scale it back to "physical" units */
    cellbb->ndim = bb->ndim;
    for(d=0; d<bb->ndim; d++){
	factor = (bb->sz[d])/iscale;
	cellbb->rmin[d] = bb->rmin[d] + factor*icorner[d];
	cellbb->sz[d] = factor;
    }
}

int
IntsFromFloats(const float *x, unsigned int *ix, const tbbox *bb, int nbits){
    int d;
    unsigned int iscale = (1<<nbits);
    float fmaxikey = (float)iscale;
    int bad;

    bad = 0;
    for(d=0; d<bb->ndim; d++){
      float fk = fmaxikey * (x[d] - bb->rmin[d]) / (bb->sz[d]);
      if( fk < 0. || fk >= fmaxikey ){
	bad = -1;
	break;
      }
      ix[d] = (unsigned int)fk;
    }
    return bad;
}

void
FloatsFromInts(const int *ix, float *x, const tbbox *bb, int nbits){
    unsigned int iscale = 1<<nbits;
    int d;
    for(d=0; d<bb->ndim; d++){
	x[d] = ix[d] * ((bb->sz[d])/iscale);
    }
}

/* These two names conflict with physics_generic.c.  Good.  It will
   keep me from using physics_generic.c accidentally. */
Key_t 
KeyFromInts(unsigned int *xp, int ndim, int nbits){
    Key_t key;
    unsigned int rshift, bits, dim;

    /* Set first bit. Important! */
    key = KeyInt(1);
    for(rshift=nbits; rshift; ){
	rshift--;
	bits = 0;
	for(dim = 0; dim < ndim; dim++ )
	    bits |= ((xp[dim]>>rshift)&1) << dim;
	key = KeyOrInt(KeyLshift(key, ndim), bits);
    }
    return(key);
}

/* Notice that the return value is DIFFERENT FROM physics_generic.c 
 Here, we return the number of bits.  It's easier for the caller to do
 left-shift than ilog2. */
extern Key_t _PHKeyFromInts3d(int *ikey, int depth, int start, int type);
extern int _IntsFromPHKey3d(Key_t key, int *ikey, int depth, int start, int type);
extern Key_t _PHKeyFromInts2d(int *ikey, int depth, int start, int type);
extern int _IntsFromPHKey2d(Key_t key, int *ikey, int depth, int start, int type);

int
IntsFromKey(Key_t key, unsigned int *ip, int ndim){
    unsigned int iscale = 1;
    unsigned int lev = 0;
    int i;

    for(i=0; i<ndim; i++){
	ip[i] = 0;
    }
    while(KeyGT(key, KeyInt(1))) {
	for(i=0; i<ndim; i++){
	    if( KeyAndInt(key, (1<<i)) )
	      ip[i] |= iscale;
	}
	key = KeyRshift(key, ndim);
	iscale <<= 1;
	lev++;
    }
    return lev;
}

Key_t PHKeyFromInts(unsigned int *ikey, int ndim, int depth){
  switch(ndim){
  case 3:
    return _PHKeyFromInts3d(ikey, depth, 0, 0);
  case 2:
    return _PHKeyFromInts2d(ikey, depth, 0, 0);
  case 1:
    return KeyFromInts(ikey, 1, depth);
  default:
    Error("Peano-Hilbert keys only supported for 1<=ndim<=3\n");
  }
}

int IntsFromPHKey(Key_t key, unsigned int *ikey, int ndim){
  switch(ndim){
  case 3:
    return _IntsFromPHKey3d(key, ikey, TreeLevel(key, ndim), 0, 0);
  case 2:
    return _IntsFromPHKey2d(key, ikey, TreeLevel(key, ndim), 0, 0);
  case 1:
    return IntsFromKey(key, ikey, ndim);
  default:
    Error("Peano-Hilbert keys only supported for 1<=ndim<=3\n");
  }
}
