/* This version has a PrintMemfileVia0 that actually sends the data
   to node0.  That's not really necessary if we just wait for the go-ahead,
   we shouldn't tickle NORMA... */
#include <stdio.h>
#include <string.h>
#include "Malloc.h"
#include "Msgs.h"
#include "error.h"
#include "protos.h"
#include "mpmy.h"

/* Here we try to implement a cicular memory buffer which we can use as */
/* the vfprintf-like arg to Msg_init */

#if 1
static char *memfile;
static int memfile_offset;
static int memfile_bufsz;

#ifdef __PARAGON__
static void HrecvSetup(int);
#endif

void
memfile_init(int sz)
{
    memfile = Malloc(sz);
    memfile_offset = 0;
    memfile_bufsz = sz-1;
    memfile[sz-1] = 0;   /* final null to make wrapped output cleaner */
#ifdef __PARAGON__ 
    HrecvSetup(sz);
#endif
}

void
memfile_delete(void)
{
    Free(memfile);
    memfile_offset = 0;
    memfile_bufsz = 0;
}
#else
/* IPD will not output the buffer in a reasonable manner using code above */
char memfile[65536];
int memfile_offset;
int memfile_bufsz = 65535;
void memfile_init(){}
void memfile_delete(){}
#endif

#define BUFSZ 1024

void
memfile_vfprintf(void *junk, const char *fmt, va_list args)
{
    char tbuf[BUFSZ];	/* This might overflow, but msgs should be small */
    int i, len;

    vsprintf(tbuf, fmt, args);
    len = strlen(tbuf);
    if (len >= BUFSZ) Error("Buffer overflowed in mem_vfprintf\n");
    for (i = 0; i <= len; i++) {
	memfile[(memfile_offset+i)%memfile_bufsz] = tbuf[i];
    }
    memfile_offset += len;
}

void
PrintMemfile(void)
{
    if (memfile_offset == 0) return;
    printf("----- Messages from procnum %d -----\n", MPMY_Procnum());
    if (memfile_offset < memfile_bufsz) {
	printf("%s\n", memfile);
    } else {
	printf("Buffer has wrapped\n");
	printf("%s\n%s\n", memfile+(memfile_offset+1)%memfile_bufsz,
	       memfile);
    }
    fflush(stdout);
}

#ifndef __PARAGON__
/* I used lots of paragon-specific code in the implementation below, so
   non-paragon callers won't actually go 'via' node 0. */

void PrintMemfileVia0(void)
{
    PrintMemfile();
}

#else

/* The paragon will crash if all nodes do PrintMemfile simultaneously.
   The dreaded NORMA bug.  So what to do?  */
#include <nx.h>

/* 'random' numbers */
#define PRTMEMTYPE1 0x39847
#define PRTMEMTYPE2 0x56473
#define PRTMEMTYPE3 0x29384

char *memfile2;
int req_pair[2];

/* This is the hrecv handler set up only in node 0 for PRTMEMTYPE1 messages.
   It handshakes with the requesting node and then dumps the buffer to stdout
   on its behalf */
static void
PrintMemfileHndlr(long type, long count, long node, long ptype){
    int reqbytes, offset, start;

    if( count != 2*sizeof(int) || type != PRTMEMTYPE1 || reqbytes < 0 )
	/* Returning will cause the sender to hang.  But what else
	   can we do.  His alarm clock will wake him... */
	return;
    reqbytes = req_pair[0];
    offset = req_pair[1];
    printf("PMH: node=%d, reqbytes=%d, offset=%d\n", node, reqbytes, offset);
    if( reqbytes > memfile_bufsz )
	reqbytes = memfile_bufsz;
    csend(PRTMEMTYPE2, &reqbytes, sizeof(int), node, 0);
    crecv(PRTMEMTYPE3, memfile2, reqbytes);
    printf("PMH: node=%d, handshake done!\n", node);
    
    printf("Memfile buffer from node %d\n", node);
    if (offset < reqbytes) {
	fwrite(memfile2, 1, offset, stdout);
    } else {
	printf("Buffer has wrapped\n");
	start = (offset+1)%reqbytes;
	fwrite(memfile2+start, 1, reqbytes-start, stdout); 
	printf("-- WRAP --");
	start = offset%reqbytes;
	fwrite(memfile, 1, start, stdout);
	printf("%s\n%s\n", memfile+(memfile_offset+1)%memfile_bufsz,
	       memfile);
    }
    putchar('\n');
    fflush(stdout);
}

/* This is the alternative version of PrintMemfile.  Perhaps it should
   simply replace the other one?  */
void
PrintMemfileVia0(void){
    int reqpair[2], reply;
    long smid, rmid;
    int rdone, sdone;
    double give_up_time;

    /* Bail out in 60 sec. */
    give_up_time = dclock() + 60.0;

    printf("Printmemfilevia0 on proc %d, give up at %g\n", mynode(), give_up_time);
    reqpair[0] = memfile_bufsz;
    reqpair[1] = memfile_offset;
    smid = isend(PRTMEMTYPE1, reqpair, sizeof(int), 0, 0);
    rmid = irecv(PRTMEMTYPE2, &reply, sizeof(int));
    sdone = rdone = 0;
    do{
	if(!sdone) sdone = msgdone(smid);
	if(!rdone) rdone = msgdone(rmid);
    }while(dclock() < give_up_time && (!sdone || !rdone));
    printf("Printmemfilevia0 on proc %d, sdone=%d, rdone=%d, time=%g\n", mynode(),
	   sdone, rdone, dclock());
    if( sdone && rdone )
	csend(PRTMEMTYPE3, memfile, reply, 0, 0);
}

static void
HrecvSetup(int sz){
    if( mynode() == 0 ){
	memfile2 = Malloc(sz);
	hrecv(PRTMEMTYPE1, req_pair, sizeof(req_pair), PrintMemfileHndlr);
    }
    gsync();
}

#endif /* __PARAGON__ */
