/* gk9.c - Gravity Kernel benchmark, version 9.  
   Copyright 1993-1998, Michael S. Warren and John K. Salmon.  All rights
   reserved. */

/* This file contains a benchmark of the most important loop in our
   gravitational N-body computation.  It is completely self-contained.
   The intent is to explore various optimization settings of the
   compiler, assembly-language inclusions or other options to obtain
   the fastest possible computation of the necessary functionality.
   The problem is to repeatedly (NLOOP times) compute Newton's law of
   gravitation (both potential energy and acceleration) for the
   interaction of a single particle (mass0, pos0[0], pos0[1],
   pos0[2]), with a vector of NVEC particles (btab).  Any and all
   optimizations are allowed, including requiring specific alignments,
   assembly language hacks, rewriting code, etc.

   VERSIONS:
   2 Feb 97 - first rewrite of gk8.c.  Verified that it works on
   hppa(anubis), sgi(dis), sunos(sampson), paragon(raptor), 
   x86linux(penguin).

   3 Feb 97 - added do_grav_split.  Added correctness check of results.

   7 Feb 97 - Lots of small changes in the cheb routines.  Compute
   x2=x*x and x3=x2*x as explicit temporaries.  Combine the u[] and
   t[] tables into ut[] so we get both values on the same cache line.
   Avoid the ut[] table lookup if the exponent is the same as last
   time.  Initialize the local accumulators to zero and then add them
   to the argument-list only at the very end (this should improve
   overall accuracy without affecting speed).  Unrolled split into
   split2 (which seems to be a bad idea).  Created an explicit loop
   variable, i, in most of the loops, since apparently some compilers
   don't grok the *p++ iterator.  Write dr2 = eps2 + Dotx(...)  to
   give the compiler more freedom to reorder the additions.  Modified
   the 'nosqrt' versions so we do a simple multiply in lieu of the
   sqrt.  Thus, it is now a likely upper-bound on speed, rather than
   an wild guess based on the assumption that 1./sqrt(x) costs 8 flops.

   COMPILATION NOTES: 

   - Writing an assembly-language or other alternative: Arrange,
   however you like, to link with a file that defines another function
   with the same 'signature' as all the rest of the grav_func
   typedefs.  When compiling gk9.c, include something like: 
     -D DO_GRAV_EXTERNAL=new_function_name 

   on the command line.  The new function will be timed along with the
   rest.  The Makefile has an example for the i860 assembly language
   implementation of 'do_grav_fast', and an hppa asm version too.

   - We use cpu times obtained with 'clock()' to measure time.
   Unfortunately, despite numerous standards, it is still not possible
   to portably measure time.  System dependent timing stuff should be
   added between the TIMER UGLINESS comments as necessary.  Currently,
   it is possible to pre-define, e.g., cc -DSEC_PER_TICK=1.0e-6 if the
   macro CLOCKS_PER_SEC is not defined in time.h as it should be.

   - The code assumes that <math.h> and libm.a has an appropriate definition
   for float sqrtf(float).  If not, it will be necessary to add
   -DNO_SQRTF to the compile command line.  This can be handled on a
   system-dependent basis within the SQRT_HACKS comments.

   - Often, there is some non-standard, non-portable hack that gives a
   fast, single-precision (or less) inverse-sqrt.  See the code
   between the SQRT HACKS comments for various system-dependent
   possibilities.  Add more as needed, preferably protected by an
   appropriate #ifdef.  The sqrt that is ultimately used should
   provide 16 or more accurate bits in the mantissa.  That is, we
   specifically DO NOT REQUIRE an IEEE conformant sqrt, but it is
   also unacceptable to stop after only five or eight bits of accuracy
   have been obtained.  See the functions *_cheb_* for examples of
   alternative inverse sqrt (actually x^{-3/2}) computations.

   - We assume an ANSI compiler.  If you don't have an ANSI compiler, we
   don't want to talk to you.  

   - There is some remarkably dense pre-processor madness between the
   comments VOP.H.  These are just shorthand that we tend to use a
   lot.  If they somehow introduce inefficiencies, they can be
   avoided, but we want to know why.  They are specifically designed to
   be 'hard' on the preprocessor, but 'easy' on the compiler.

*/

#define _POSIX_SOURCE
#include <math.h>
#include <stdio.h>
#include <float.h>
#include <values.h>
#ifndef FLT_MAX
#define FLT_MAX 3.40282347e+38F
#endif
#ifndef FLT_MIN
#define FLT_MIN 1.17549435e-38F
#endif
#ifndef UINT_MAX
#define UINT_MAX (~0U)
#endif

/*********************** STANDARD PARAMETERS ************************/
/* Parameters to change to test different vector lengths, etc.  These are
   the 'fiducial' ones since around 1993 */
#define NLOOP 10000		/* how many times to call do_*  */
#define NVEC (501)		/* how many to do per do_* call */
#define NEXTRA (5)		/* padding at the end of the arrays */
#define NELEM (NDIM+1)		/* mass and 3 positions */
#define LAST (NVEC*NELEM)	/* index of last float in btab */
#define LOOPCNT(nfloat, unroll)  (((nfloat)/NELEM + (unroll-1))/unroll)
#define NOT_SQRT(x) (0.00001F*(x))/* used in the nosqrt routines */
#define VERSION_STRING "7feb97"
/********************** END OF STANDARD PARAMETERS ***********************/

/***********************TIMER UGLINESS*****************************/
/* Intel Inside */
#if defined(__PARAGON__) || defined(__DELTA__)
#define clock mclock
#ifdef __DELTA__
typedef long clock_t;
#endif
#define SEC_PER_TICK ((float)1e-3)
#define NO_TIME_H
#endif /* __INTEL_SSD__ */

#ifndef NO_TIME_H
#include <time.h>
#endif /* NO_TIME_H */

#ifndef SEC_PER_TICK
#ifdef CLOCKS_PER_SEC
#define SEC_PER_TICK (1.F/CLOCKS_PER_SEC)
#else  /* SEC_PER_TICK */
#if defined(sparc) /* probably others... */
#define SEC_PER_TICK (1.0e-6)
#else
 # error Neither CLOCKS_PER_SEC nor SEC_PER_TICK defined
#endif 
#endif /* CLOCKS_PER_SEC */
#endif /* SEC_PER_TICK */
/****************END OF TIMER UGLINESS ************************/

/************************* SQRT HACKS ***************************/
#if defined (sparc) 
#if defined(__GNUC__) && !defined(__INSIGHT__)
#undef sqrtf_fast
#define sqrtf_fast(x) \
     ({ float __value, __arg = (x);   \
        asm ("fsqrts %1,%0": "=f" (__value): "f" (__arg));  \
        __value; })
#else /* GNUC */
     /* sqrtf is in the library, but not in acc's math.h */
extern float sqrtf(float);
#endif /* GNUC */
#endif /* sparc */

#if defined(mips) 
#undef sqrtf_fast
#if defined(__GNUC__)
/* I'm not positive the sqrt.s instruction is right (msw) */
#define sqrtf_fast(x) \
     ({ float __value, __arg = (x);   \
        asm ("sqrt.s %1, %0" : "=f" (__value): "f" (__arg));  \
        __value; })
#else
/* Compilation with -OPT:fast_sqrt -OPT:IEEE_arithmetic=3 seems to
   make a big difference on an R8000.  The first enables use of the fast
   sqrt instruction, and the second enables the compiler to play fast-and-
   loose with expressions involving division.  -- johns Oct 9, 1995 */
#define sqrtf_fast sqrtf
#endif
#endif

#if defined(__alpha) && !defined(__linux)
#undef sqrtf_fast
#define sqrtf_fast(x) F_sqrtf(x)
extern float F_sqrtf(float);
#endif /* __alpha */

/* Is this the same as -ffast-math -mcpu=rios2 ? */
#if defined(_ARCH_PWR2) && defined(__GNUC__)
#undef sqrtf_fast
#define sqrtf_fast(x) \
  ({ float __value, __arg = (x); \
     asm ("fsqrt %0,%1" : "=f" (__value) : "f" (__arg)); \
     __value; })
#endif

#if defined(_ARCH_PPC64) && defined(__GNUC__)
#undef sqrtf_fast

#define sqrtf_fast(x) \
  ({ float __value, __arg = (x); \
     asm ("fsqrts %0,%1" : "=f" (__value) : "f" (__arg)); \
     __value; })
#endif

#if defined(__hppa)
#undef sqrtf_fast
#ifdef __GNUC__
#define sqrtf_fast(x) \
  ({ float __value, __arg = (x); \
     asm ("fsqrt,sgl %1,%0" : "=fx" (__value) : "fx" (__arg)); \
     __value; })
#else  /* __GNUC__ */
     /* there are two choices here.  We can get the double-precision
	instruction inlined automatically by calling sqrt(), or we can
	make a function-call go sqrtf() which will then use the
	single- precision instruction.  We decide by looking at
	NO_SQRTF.  Since sqrtf is not in math.h so we need to mention
	it explicitly. */
#ifndef NO_SQRTF
extern float sqrtf(float);
#endif /* NO_SQRTF */
#endif /* __GNUC__ */
#endif /* __hppa */

#if defined(__linux)
/* at least not on alpha or x86 */
#define NO_SQRTF
#if defined(__alpha)
   #define PREFER_DOUBLE_SQRT
#endif
#endif

/* Unfortunately, there is no standard test to determine if libm has
   single-precision intrinsics, so we rely on the pp-symbol NO_SQRTF,
   which might be set from the command line, or in one of the system-specific
   blocks above.  If it is not set, we assume that sqrtf is properly
   defined and will be available at link time.  */
#ifndef sqrtf_fast
#ifndef NO_SQRTF
#define sqrtf_fast(x) sqrtf(x)
#else
#define sqrtf_fast(x) (float)sqrt(x)
#endif /* NO_SQRTF */
#endif /* sqrtf_fast */

#ifndef recipsqrtf
#define recipsqrtf(x) (1.0F/sqrtf_fast(x))
#endif

/*************************END OF SQRT HACKS ******************/

/******************** FROM VOP.H **********************/
/* From vop.h, with NDIM==3 */
/* Some of these are pretty gross, but they represent what the compiler
   will probably see if we compile our 'normal' code.  Any reasonable
   compiler will immediately detect that the switch/branch disappears
   after constant-elimination? */
#define NDIM 3
#ifdef __STDC__
#define GLUE(a,b) a##b
#else
#define GLUE(a,b) a/**/b
#endif
#define VVinfix(a, b, op) a[0] b[0] op a[1] b[1] op a[2] b[2]
#define Vxinfix(a, op) GLUE(a,0) op GLUE(a,1) op GLUE(a,2)
#define VxVxinfix(a, b, op) GLUE(a,0) GLUE(b,0) op GLUE(a,1) GLUE(b,1) op GLUE(a,2) GLUE(b,2)
#define Dotx(a, b) (VxVxinfix(a, *b, +))
#define Dot(a, b) (VVinfix(a, *b, +))
#define Vxd(a) Vxinfix(a,;)
#define VxS(a, b) do{ switch(NDIM){\
	case 3: GLUE(a,2) b; \
	case 2: GLUE(a,1) b; \
	case 1: GLUE(a,0) b; \
      }}while(0) 
#define VxV(a, b) do{ switch(NDIM){\
        case 3: GLUE(a,2) b[2]; \
        case 2: GLUE(a,1) b[1]; \
        case 1: GLUE(a,0) b[0]; \
      }}while(0) 
#define VxVx(a, b) do{ switch(NDIM){\
        case 3: GLUE(a,2) GLUE(b,2); \
        case 2: GLUE(a,1) GLUE(b,1); \
        case 1: GLUE(a,0) GLUE(b,0); \
      }}while(0) 
#define VVx(a, b) do{ switch(NDIM){\
        case 3: a[2] GLUE(b,2); \
        case 2: a[1] GLUE(b,1); \
        case 1: a[0] GLUE(b,0); \
      }}while(0) 
#define VS(a, b) do{ switch(NDIM){ \
  case 3: a[2] b; \
  case 2: a[1] b; \
  case 1: a[0] b; \
      }}while(0)
#define VV(a, b) do{ switch(NDIM) { \
  case 3: a[2] b[2]; \
  case 2: a[1] b[1]; \
  case 1: a[0] b[0]; \
      }}while(0)

/************************ END OF VOP.H EXTRACT *****************/

float e2;
int ncut, nterms, nquad_terms;
float phi0, mass0;
float mass0;
float pos0[NDIM];
float acc0[NDIM];
float accexact[NDIM];
float phiexact;

double de2;
double dphi0, dmass0;
double dmass0;
double dpos0[NDIM];
double dacc0[NDIM];

typedef void grav_func(const float *p, const float *end, 
			  const float *pos0, float *mass0, 
			  float *acc0, float *phi0, 
			  const float *eps2p, int *ncut);

typedef void dgrav_func(const double *p, const double *end, 
			  const double *pos0, double *mass0, 
			  double *acc0, double*phi0, 
			  const double *eps2p, int *ncut);

grav_func do_grav, do_grav_rsqt, do_grav_nosq, 
  do_grav_unroll3, do_grav_unroll2,
  do_grav_unroll4, do_grav_unro_nosq, do_grav_cheb, 
  do_grav_cheb_unroll3, do_grav_cheb_unroll2, do_grav_cheb_unroll4;

dgrav_func do_grav_cheb_dbl;

#ifdef DO_GRAV_EXTERNAL
grav_func DO_GRAV_EXTERNAL;
#endif

float btab[LAST+NEXTRA*NELEM];		/* must be 8 byte aligned */

void do_time(grav_func *funcp, const char *label){
  clock_t t;
  int i;
  float phierr, accerr;
  
  phi0 = 0.F;
  VS(acc0, = 0.F);
  mass0 = 0.F;
  (*funcp)(btab, btab+LAST, pos0, &mass0, acc0, &phi0, &e2, &ncut);

  phierr = fabs(phi0 - phiexact)/fabs(phiexact);
  VV(acc0, -= accexact);
  accerr = sqrt( Dot(acc0, acc0) / Dot(accexact, accexact));

  t = clock();
  for(i = 0; i < NLOOP; i++)
    (*funcp)(btab, btab+LAST, pos0, &mass0, acc0, &phi0, &e2, &ncut);
  t = clock() - t;
  /* We should test against a known correct answer here! */
  printf("%s: %.2f MNewtons/sec (%.1e phi error, %.1e acc error)\n", label, 
	 NLOOP*NVEC*1.e-6/(t*SEC_PER_TICK),
	 phierr, accerr);
}

double dbtab[LAST+NEXTRA*NELEM];		/* must be 8 byte aligned */

void do_double_time(dgrav_func *funcp, const char *label){
  clock_t t;
  int i;
  double dphierr, daccerr;
  
  dphi0 = 0.;
  VS(dacc0, = 0.);
  dmass0 = 0.;
  (*funcp)(dbtab, dbtab+LAST, dpos0, &dmass0, dacc0, &dphi0, &de2, &ncut);

  dphierr = fabs(dphi0 - phiexact)/fabs(phiexact);
  VV(dacc0, -= accexact);
  daccerr = sqrt( Dot(dacc0, dacc0) / Dot(accexact, accexact));

  t = clock();
  for(i = 0; i < NLOOP; i++)
    (*funcp)(dbtab, dbtab+LAST, dpos0, &dmass0, dacc0, &dphi0, &de2, &ncut);
  t = clock() - t;
  /* We should test against a known correct answer here! */
  printf("%s: %.2f dpMNewtons/sec (%.1e phi error, %.1e acc error)\n", label, 
	 NLOOP*NVEC*1.e-6/(t*SEC_PER_TICK),
	 dphierr, daccerr);
}

/* This is the fiducial routine.  Measure everything against this one. */
void 
do_grav(const float *p, const float *end, const float *pos0, float *mass0, float *acc0,
	float *phi0, const float *eps2p, int *ncut)
{
    float dr2;
    Vxd(float r);
    float phii, mor3, mass;
    Vxd(float ppos);
    Vxd(float a);
    float phi = 0.F;
    float total_mass = 0.F;
    float eps2 = *eps2p;

    VxS(a, = 0.F);
    VxV(ppos, = pos0);

    while (p < end) {
	mass = *p++;
	r0 = *p++;
	r1 = *p++;
	r2 = *p++;
	VxVx(r, -= ppos);	/* 3 flops */

	dr2 = eps2 + Dotx(r, r);	/* 6 flops */

	phii = 1.0F/(float)sqrt(dr2);	/* 1 inverse-sqrt */
	
	mor3 = phii * phii;	/* 5 flops */
	phii *= mass;
	total_mass += mass;
	mor3 *= phii;
	phi -= phii;

	VxVx(a, += mor3 * r);	/* 6 flops */
    }
    VVx(acc0, += a);
    *mass0 += total_mass;
    *phi0 += phi;
}

void 
do_grav_rsqt(const float *p, const float *end, const float *pos0, float *mass0, 
	      float *acc0, float *phi0, const float *eps2p, int *ncut)
{
    float dr2;
    Vxd(float r);
    float phii, mor3, mass;
    Vxd(float ppos);
    Vxd(float a);
    float phi = 0.F;
    float total_mass = 0.F;
    float eps2 = *eps2p;

    VxS(a, = 0.F);
    VxV(ppos, = pos0);

    while (p < end) {
	mass = *p++;
	r0 = *p++;
	r1 = *p++;
	r2 = *p++;
	VxVx(r, -= ppos);	/* 3 flops */

	dr2 = eps2 + Dotx(r, r);	/* 6 flops */

	phii = recipsqrtf(dr2);	/* 8 flops */
	
	mor3 = phii * phii;	/* 5 flops */
	phii *= mass;
	total_mass += mass;
	mor3 *= phii;
	phi -= phii;

	VxVx(a, += mor3 * r);	/* 6 flops */
    }
    VVx(acc0, += a);
    *mass0 += total_mass;
    *phi0 += phi;
}
 
void 
do_grav_split(const float *p, const float *end, const float *pos0, float *mass0, 
	      float *acc0, float *phi0, const float *eps2p, int *ncut)
{
    float dra2, drb2;
    Vxd(float ra);
    Vxd(float rb);
    float phii, mor3, massa, massb;
    Vxd(float ppos);
    Vxd(float a);
    float phi = 0.F;
    float total_mass = 0.F;
    float eps2 = *eps2p;
    int n = LOOPCNT(end - p, 1);
    int i;

    VxS(a, = 0.F);
    VxV(ppos, = pos0);

    massa = *p++;
    ra0 = *p++;
    ra1 = *p++;
    ra2 = *p++;
    VxVx(ra, -= ppos);	/* 3 flops */

    dra2 = eps2 + Dotx(ra, ra);	/* 6 flops */
    
    for(i=1; i<n; i++){
      phii = recipsqrtf(dra2);	/* 8 flops */
      /* dr2, phii are for distance to btab[i-1] */
	
      /* determine distance to btab[i] */
      massb = *p++;
      rb0 = *p++;
      rb1 = *p++;
      rb2 = *p++;
      VxVx(rb, -= ppos);	/* 3 flops */

      drb2 = eps2 + Dotx(rb, rb);	/* 6 flops */

      /* accumulate force/potential for btab[i-1] */
      total_mass += massa;	/* 5 flops */
      mor3 = phii * phii;	
      phii *= massa;
      mor3 *= phii;
      phi -= phii;
      
      VxVx(a, += mor3 * ra);	/* 6 flops */
      VxVx(ra, = rb);
      dra2 = drb2;
      massa = massb;
    }

    phii = recipsqrtf(dra2);	/* 8 flops */
	
    mor3 = phii * phii;	/* 5 flops */
    phii *= massa;
    total_mass += massa;
    mor3 *= phii;
    phi -= phii;
    VxVx(a, += mor3 * ra);	/* 6 flops */

    VVx(acc0, += a);
    *mass0 += total_mass;
    *phi0 += phi;
}
 
void 
do_grav_split2(const float *p, const float *end, const float *pos0, float *mass0, 
	      float *acc0, float *phi0, const float *eps2p, int *ncut)
{
  /* A and B are the unrolled variables.  
     a and b are the 'previous' and 'current' loop iterations */
    float draA2, drbA2;
    float draB2, drbB2;
    Vxd(float raA);
    Vxd(float raB);
    Vxd(float rbA);
    Vxd(float rbB);
    float phiiA, mor3A, massaA, massbA;
    float phiiB, mor3B, massaB, massbB;
    Vxd(float ppos);
    Vxd(float aA);
    Vxd(float aB);
    float phiA = 0.F;
    float phiB = 0.F;
    float total_massA = 0.F;
    float total_massB = 0.F;
    float eps2 = *eps2p;
    int n = LOOPCNT(end - p, 2);
    int i;

    VxS(aA, = 0.F);
    VxS(aB, = 0.F);
    VxV(ppos, = pos0);

    massaA = *p++;
    raA0 = *p++;
    raA1 = *p++;
    raA2 = *p++;

    massaB = *p++;
    raB0 = *p++;
    raB1 = *p++;
    raB2 = *p++;
    
    VxVx(raA, -= ppos);
    VxVx(raB, -= ppos);

    draA2 = eps2 + Dotx(raA, raA);
    draB2 = eps2 + Dotx(raB, raB);
    
    for(i=1; i<n; i++){
      phiiA = recipsqrtf(draA2);
      phiiB = recipsqrtf(draB2);
	
      /* determine distance to btab[i] */
      massbA = *p++;
      rbA0 = *p++;
      rbA1 = *p++;
      rbA2 = *p++;
      massbB = *p++;
      rbB0 = *p++;
      rbB1 = *p++;
      rbB2 = *p++;

      VxVx(rbA, -= ppos);
      VxVx(rbB, -= ppos);

      drbA2 = eps2 + Dotx(rbA, rbA);
      drbB2 = eps2 + Dotx(rbB, rbB);

      total_massA += massaA;
      total_massB += massaB;

      mor3A = phiiA * phiiA;	
      mor3B = phiiB * phiiB;

      phiiA *= massaA;
      phiiB *= massaB;

      mor3A *= phiiA;
      mor3B *= phiiB;
      phiA -= phiiA;
      phiB -= phiiB;
      
      VxVx(aA, += mor3A * raA);
      VxVx(aB, += mor3B * raB);

      VxVx(raA, = rbA);
      VxVx(raB, = rbB);

      draA2 = drbA2;
      draB2 = drbB2;
      massaA = massbA;
      massaB = massbB;
    }

    phiiA = recipsqrtf(draA2);
    phiiB = recipsqrtf(draB2);
	
    mor3A = phiiA * phiiA;
    mor3B = phiiB * phiiB;

    phiiA *= massaA;
    phiiB *= massaB;

    total_massA += massaA;
    total_massB += massaB;

    mor3A *= phiiA;
    mor3B *= phiiB;

    phiA -= phiiA;
    phiB -= phiiB;
    VxVx(aA, += mor3A * raA);
    VxVx(aB, += mor3B * raB);

    VxVx(aA, += aB);
    VVx(acc0, += aA);
    *mass0 += (total_massA + total_massB);
    *phi0 += (phiA + phiB);
}
 
void 
do_grav_unroll2(const float *p, const float *end, const float *pos0, 
	       float *mass0, float *acc0, float *phi0, const float 
	       *eps2p, int *ncut)
{
    float dr2a;
    float dr2b;
    Vxd(float ra);
    Vxd(float rb);
    float phiia;
    float phiib;
    float massa;
    float massb;
    float mor3a;
    float mor3b;
    Vxd(float ppos);
    Vxd(float a);
    float phi = 0.F;
    float total_mass = 0.F;
    const float eps2 = *eps2p;
    const int n = LOOPCNT(end-p, 2);
    int i;

    VxS(a, = 0.F);
    VxV(ppos, = pos0);

    for (i = 0; i < n; i++) {
	massa = *p++;
	ra0 = *p++;
	ra1 = *p++;
	ra2 = *p++;

	VxVx(ra, -= ppos);
	dr2a = eps2 + Dotx(ra, ra);

	massb = *p++;
	rb0 = *p++;
	rb1 = *p++;
	rb2 = *p++;

	VxVx(rb, -= ppos);
	dr2b = eps2 + Dotx(rb, rb);

	phiia = recipsqrtf(dr2a);
	phiib = recipsqrtf(dr2b);
	mor3a = phiia * phiia;
	mor3b = phiib * phiib;
	phiia *= massa;
	phiib *= massb;

	total_mass += massa + massb;

	mor3a *= phiia;
	mor3b *= phiib;

	phi -= phiia;
	phi -= phiib;

	VxVx(a, += mor3a * ra);
	VxVx(a, += mor3b * rb);
    }
    VVx(acc0, += a);
    *mass0 += total_mass;
    *phi0 += phi;
}

void 
do_grav_unroll3(const float *p, const float *end, const float *pos0, 
	       float *mass0, float *acc0, float *phi0, const float 
	       *eps2p, int *ncut)
{
    float dr2a;
    float dr2b;
    float dr2c;
    Vxd(float ra);
    Vxd(float rb);
    Vxd(float rc);
    float phiia;
    float phiib;
    float phiic;
    float massa;
    float massb;
    float massc;
    float mor3a;
    float mor3b;
    float mor3c;
    Vxd(float ppos);
    Vxd(float a);
    float phi = 0.F;
    float total_mass = 0.F;
    const float eps2 = *eps2p;
    const int n = LOOPCNT(end-p, 3);
    int i;

    VxS(a, = 0.F);
    VxV(ppos, = pos0);

    for (i = 0; i < n; i++) {
	massa = *p++;
	ra0 = *p++;
	ra1 = *p++;
	ra2 = *p++;

	massb = *p++;
	rb0 = *p++;
	rb1 = *p++;
	rb2 = *p++;

	massc = *p++;
	rc0 = *p++;
	rc1 = *p++;
	rc2 = *p++;
	VxVx(ra, -= ppos);
	VxVx(rb, -= ppos);
	VxVx(rc, -= ppos);

	dr2a = eps2 + Dotx(ra, ra);
	dr2b = eps2 + Dotx(rb, rb);
	dr2c = eps2 + Dotx(rc, rc);

	phiia = recipsqrtf(dr2a);
	mor3a = phiia * phiia;
	phiia *= massa;
	phiib = recipsqrtf(dr2b);
	mor3b = phiib * phiib;
	phiib *= massb;
	phiic = recipsqrtf(dr2c);
	mor3c = phiic * phiic;
	phiic *= massc;

	total_mass += massa + massb + massc;

	mor3a *= phiia;
	mor3b *= phiib;
	mor3c *= phiic;

	phi -= phiia;
	phi -= phiib;
	phi -= phiic;

	VxVx(a, += mor3a * ra);
	VxVx(a, += mor3b * rb);
	VxVx(a, += mor3c * rc);
    }
    VVx(acc0, += a);
    *mass0 += total_mass;
    *phi0 += phi;
}

void 
do_grav_unroll4(const float *p, const float *end, const float *pos0, 
	       float *mass0, float *acc0, float *phi0, const float 
	       *eps2p, int *ncut)
{
    float dr2a;
    float dr2b;
    float dr2c;
    float dr2d;
    Vxd(float ra);
    Vxd(float rb);
    Vxd(float rc);
    Vxd(float rd);
    float phiia;
    float phiib;
    float phiic;
    float phiid;
    float massa;
    float massb;
    float massc;
    float massd;
    float mor3a;
    float mor3b;
    float mor3c;
    float mor3d;
    Vxd(float ppos);
    Vxd(float a);
    float phi = 0.F;
    float total_mass = 0.F;
    const float eps2 = *eps2p;
    const int n = LOOPCNT(end-p, 4);
    int i;

    VxS(a, = 0.F);
    VxV(ppos, = pos0);

    for (i = 0; i < n; i++) {
	massa = *p++;
	ra0 = *p++;
	ra1 = *p++;
	ra2 = *p++;

	massb = *p++;
	rb0 = *p++;
	rb1 = *p++;
	rb2 = *p++;

	massc = *p++;
	rc0 = *p++;
	rc1 = *p++;
	rc2 = *p++;

	massd = *p++;
	rd0 = *p++;
	rd1 = *p++;
	rd2 = *p++;

	VxVx(ra, -= ppos);
	VxVx(rb, -= ppos);
	VxVx(rc, -= ppos);
	VxVx(rd, -= ppos);

	dr2a = eps2 + Dotx(ra, ra);
	dr2b = eps2 + Dotx(rb, rb);
	dr2c = eps2 + Dotx(rc, rc);
	dr2d = eps2 + Dotx(rd, rd);

	phiia = recipsqrtf(dr2a);
	mor3a = phiia * phiia;
	phiia *= massa;
	phiib = recipsqrtf(dr2b);
	mor3b = phiib * phiib;
	phiib *= massb;
	phiic = recipsqrtf(dr2c);
	mor3c = phiic * phiic;
	phiic *= massc;
	phiid = recipsqrtf(dr2d);
	mor3d = phiid * phiid;
	phiid *= massd;

	total_mass += massa + massb + massc + massd;

	mor3a *= phiia;
	mor3b *= phiib;
	mor3c *= phiic;
	mor3d *= phiid;

	phi -= phiia;
	phi -= phiib;
	phi -= phiic;
	phi -= phiid;

	VxVx(a, += mor3a * ra);
	VxVx(a, += mor3b * rb);
	VxVx(a, += mor3c * rc);
	VxVx(a, += mor3d * rd);
    }
    VVx(acc0, += a);
    *mass0 += total_mass;
    *phi0 += phi;
}

void 
do_grav_nosq(const float *p, const float *end, const float *pos0, float *mass0, 
	      float *acc0, float *phi0, const float *eps2p, int *ncut)
{
    float dr2;
    Vxd(float r);
    float phii, mor3, mass;
    Vxd(float ppos);
    Vxd(float a);
    float phi = 0.F;
    float total_mass = 0.F;
    float eps2 = *eps2p;

    VxS(a, = 0.F);
    VxV(ppos, = pos0);

    while (p < end) {
	mass = *p++;
	r0 = *p++;
	r1 = *p++;
	r2 = *p++;
	VxVx(r, -= ppos);	/* 3 flops */

	dr2 = eps2 + Dotx(r, r);	/* 6 flops */

	phii = NOT_SQRT(dr2);
	
	mor3 = phii * phii;	/* 5 flops */
	phii *= mass;
	total_mass += mass;
	mor3 *= phii;
	phi -= phii;

	VxVx(a, += mor3 * r);	/* 6 flops */
    }
    VVx(acc0, += a);
    *mass0 += total_mass;
    *phi0 += phi;
}

void 
do_grav_unro_nosq(const float *p, const float *end, const float *pos0, 
	       float *mass0, float *acc0, float *phi0, const float 
	       *eps2p, int *ncut)
{
    float dr2a;
    float dr2b;
    float dr2c;
    Vxd(float ra);
    Vxd(float rb);
    Vxd(float rc);
    float phiia;
    float phiib;
    float phiic;
    float massa;
    float massb;
    float massc;
    float mor3a;
    float mor3b;
    float mor3c;
    Vxd(float ppos);
    Vxd(float a);
    float phi = 0.F;
    float total_mass = 0.F;
    const float eps2 = *eps2p;
    const int n = LOOPCNT(end-p, 3);
    int i;

    VxS(a, = 0.F);
    VxV(ppos, = pos0);

    for (i = 0; i < n; i++) {
	massa = *p++;
	ra0 = *p++;
	ra1 = *p++;
	ra2 = *p++;

	massb = *p++;
	rb0 = *p++;
	rb1 = *p++;
	rb2 = *p++;

	massc = *p++;
	rc0 = *p++;
	rc1 = *p++;
	rc2 = *p++;
	VxVx(ra, -= ppos);
	VxVx(rb, -= ppos);
	VxVx(rc, -= ppos);

	dr2a = eps2 + Dotx(ra, ra);
	dr2b = eps2 + Dotx(rb, rb);
	dr2c = eps2 + Dotx(rc, rc);

	phiia = NOT_SQRT(dr2a);
	mor3a = phiia * phiia;
	phiia *= massa;
	phiib = NOT_SQRT(dr2b);
	mor3b = phiib * phiib;
	phiib *= massb;
	phiic = NOT_SQRT(dr2c);
	mor3c = phiic * phiic;
	phiic *= massc;

	total_mass += massa + massb + massc;

	mor3a *= phiia;
	mor3b *= phiib;
	mor3c *= phiic;

	phi -= phiia;
	phi -= phiib;
	phi -= phiic;

	VxVx(a, += mor3a * ra);
	VxVx(a, += mor3b * rb);
	VxVx(a, += mor3c * rc);
    }
    VVx(acc0, += a);
    *mass0 += total_mass;
    *phi0 += phi;
}


/* from karp.h */
#ifdef __CRAY__
#define FLOAT_PUN short
#else
#define FLOAT_PUN int
#endif

/* Table to reduce domain of input for Chebychev polynomial approx.
   The first column is u = (1/2)^k.  The second column is t=u^3/2 */

static const float ut[2*256] = {
  1.70141e+38,         FLT_MAX,  
8.50706e+37,	     FLT_MAX,    
4.25353e+37,	     FLT_MAX,    
2.12676e+37,	     FLT_MAX,    
1.06338e+37,	     FLT_MAX,    
5.31691e+36,	     FLT_MAX,    
  2.65846e+36,	      FLT_MAX,   
1.32923e+36,	     FLT_MAX,    
6.64614e+35,	     FLT_MAX,    
3.32307e+35,	     FLT_MAX,    
1.66153e+35,	     FLT_MAX,    
8.30767e+34,	     FLT_MAX,    
  4.15384e+34,	       FLT_MAX,  
2.07692e+34,	     FLT_MAX,    
1.03846e+34,	     FLT_MAX,    
5.19230e+33,	     FLT_MAX,    
2.59615e+33,	     FLT_MAX,    
1.29807e+33,	     FLT_MAX,    
  6.49037e+32,	       FLT_MAX,  
3.24519e+32,	     FLT_MAX,    
1.62259e+32,	     FLT_MAX,    
8.11296e+31,	     FLT_MAX,    
4.05648e+31,	     FLT_MAX,    
2.02824e+31,	     FLT_MAX,    
  1.01412e+31,	       FLT_MAX,  
5.07060e+30,	     FLT_MAX,    
2.53530e+30,	     FLT_MAX,    
1.26765e+30,	     FLT_MAX,    
6.33825e+29,	     FLT_MAX,    
3.16913e+29,	     FLT_MAX,    
  1.58456e+29,	       FLT_MAX,  
7.92282e+28,	     FLT_MAX,    
3.96141e+28,	     FLT_MAX,    
1.98070e+28,	     FLT_MAX,    
9.90352e+27,	     FLT_MAX,    
4.95176e+27,	     FLT_MAX,    
  2.47588e+27,	       FLT_MAX,  
1.23794e+27,	     FLT_MAX,    
6.18970e+26,	     FLT_MAX,    
3.09485e+26,	     FLT_MAX,    
1.54743e+26,	     FLT_MAX,    
7.73713e+25,	     FLT_MAX,    
  3.86856e+25,	   2.40616e+38,  
1.93428e+25,	 8.50706e+37,    
9.67141e+24,	 3.00770e+37,    
4.83570e+24,	 1.06338e+37,    
2.41785e+24,	 3.75962e+36,    
1.20893e+24,	 1.32923e+36,    
  6.04463e+23,	   4.69953e+35,  
3.02231e+23,	 1.66153e+35,    
1.51116e+23,	 5.87441e+34,    
7.55579e+22,	 2.07692e+34,    
3.77789e+22,	 7.34302e+33,    
1.88895e+22,	 2.59615e+33,    
  9.44473e+21,	   9.17877e+32,  
4.72237e+21,	 3.24519e+32,    
2.36118e+21,	 1.14735e+32,    
1.18059e+21,	 4.05648e+31,    
5.90296e+20,	 1.43418e+31,    
2.95148e+20,	 5.07060e+30,    
  1.47574e+20,	   1.79273e+30,  
7.37870e+19,	 6.33825e+29,    
3.68935e+19,	 2.24091e+29,    
1.84467e+19,	 7.92282e+28,    
9.22337e+18,	 2.80114e+28,    
4.61169e+18,	 9.90352e+27,    
  2.30584e+18,	   3.50142e+27,  
1.15292e+18,	 1.23794e+27,    
5.76461e+17,	 4.37678e+26,    
2.88230e+17,	 1.54743e+26,    
1.44115e+17,	 5.47097e+25,    
7.20576e+16,	 1.93428e+25,    
  3.60288e+16,	   6.83872e+24,  
1.80144e+16,	 2.41785e+24,    
9.00720e+15,	 8.54840e+23,    
4.50360e+15,	 3.02231e+23,    
2.25180e+15,	 1.06855e+23,    
1.12590e+15,	 3.77789e+22,    
  5.62950e+14,	   1.33569e+22,  
2.81475e+14,	 4.72237e+21,    
1.40737e+14,	 1.66961e+21,    
7.03687e+13,	 5.90296e+20,    
3.51844e+13,	 2.08701e+20,    
1.75922e+13,	 7.37870e+19,    
  8.79609e+12,	   2.60876e+19,  
4.39805e+12,	 9.22337e+18,    
2.19902e+12,	 3.26095e+18,    
1.09951e+12,	 1.15292e+18,    
5.49756e+11,	 4.07619e+17,    
2.74878e+11,	 1.44115e+17,    
  1.37439e+11,	   5.09524e+16,  
6.87195e+10,	 1.80144e+16,    
3.43597e+10,	 6.36905e+15,    
1.71799e+10,	 2.25180e+15,    
8.58993e+09,	 7.96131e+14,    
4.29497e+09,	 2.81475e+14,    
  2.14748e+09,	   9.95164e+13,  
1.07374e+09,	 3.51844e+13,    
5.36871e+08,	 1.24396e+13,    
2.68435e+08,	 4.39805e+12,    
1.34218e+08,	 1.55494e+12,    
6.71089e+07,	 5.49756e+11,    
  3.35544e+07,	   1.94368e+11,  
1.67772e+07,	 6.87195e+10,    
8.38861e+06,	 2.42960e+10,    
4.19430e+06,	 8.58993e+09,    
2.09715e+06,	 3.03700e+09,    
1.04858e+06,	 1.07374e+09,    
      524288.,	   3.79625e+08,  
    262144.,	 1.34218e+08,    
    131072.,	 4.74531e+07,    
    65536.0,	 1.67772e+07,    
    32768.0,	 5.93164e+06,    
    16384.0,	 2.09715e+06,    
      8192.00,	       741455.,  
    4096.00,	     262144.,    
    2048.00,	     92681.9,    
    1024.00,	     32768.0,    
    512.000,	     11585.2,    
    256.000,	     4096.00,    
      128.000,	       1448.15,  
    64.0000,	     512.000,    
    32.0000,	     181.019,    
    16.0000,	     64.0000,    
    8.00000,	     22.6274,    
    4.00000,	     8.00000,    
      2.00000,	       2.82843,  
    1.00000,	     1.00000,    
   0.500000,	    0.353553,    
   0.250000,	    0.125000,    
   0.125000,	   0.0441942,    
  0.0625000,	   0.0156250,    
    0.0312500,	    0.00552427,  
  0.0156250,	  0.00195312,    
 0.00781250,	 0.000690534,    
 0.00390625,	 0.000244141,    
 0.00195312,	 8.63167e-05,    
0.000976562,	 3.05176e-05,    
  0.000488281,	   1.07896e-05,  
0.000244141,	 3.81470e-06,    
0.000122070,	 1.34870e-06,    
6.10352e-05,	 4.76837e-07,    
3.05176e-05,	 1.68587e-07,    
1.52588e-05,	 5.96046e-08,    
  7.62939e-06,	   2.10734e-08,  
3.81470e-06,	 7.45058e-09,    
1.90735e-06,	 2.63418e-09,    
9.53674e-07,	 9.31323e-10,    
4.76837e-07,	 3.29272e-10,    
2.38419e-07,	 1.16415e-10,    
  1.19209e-07,	   4.11590e-11,  
5.96046e-08,	 1.45519e-11,    
2.98023e-08,	 5.14488e-12,    
1.49012e-08,	 1.81899e-12,    
7.45058e-09,	 6.43110e-13,    
3.72529e-09,	 2.27374e-13,    
  1.86265e-09,	   8.03887e-14,  
9.31323e-10,	 2.84217e-14,    
4.65661e-10,	 1.00486e-14,    
2.32831e-10,	 3.55271e-15,    
1.16415e-10,	 1.25607e-15,    
5.82077e-11,	 4.44089e-16,    
  2.91038e-11,	   1.57009e-16,  
1.45519e-11,	 5.55112e-17,    
7.27596e-12,	 1.96262e-17,    
3.63798e-12,	 6.93889e-18,    
1.81899e-12,	 2.45327e-18,    
9.09495e-13,	 8.67362e-19,    
  4.54747e-13,	   3.06659e-19,  
2.27374e-13,	 1.08420e-19,    
1.13687e-13,	 3.83323e-20,    
5.68434e-14,	 1.35525e-20,    
2.84217e-14,	 4.79154e-21,    
1.42109e-14,	 1.69407e-21,    
  7.10543e-15,	   5.98943e-22,  
3.55271e-15,	 2.11758e-22,    
1.77636e-15,	 7.48678e-23,    
8.88178e-16,	 2.64698e-23,    
4.44089e-16,	 9.35848e-24,    
2.22045e-16,	 3.30872e-24,    
  1.11022e-16,	   1.16981e-24,  
5.55112e-17,	 4.13590e-25,    
2.77556e-17,	 1.46226e-25,    
1.38778e-17,	 5.16988e-26,    
6.93889e-18,	 1.82783e-26,    
3.46945e-18,	 6.46235e-27,    
  1.73472e-18,	   2.28479e-27,  
8.67362e-19,	 8.07794e-28,    
4.33681e-19,	 2.85598e-28,    
2.16840e-19,	 1.00974e-28,    
1.08420e-19,	 3.56998e-29,    
5.42101e-20,	 1.26218e-29,    
  2.71051e-20,	   4.46247e-30,  
1.35525e-20,	 1.57772e-30,    
6.77626e-21,	 5.57809e-31,    
3.38813e-21,	 1.97215e-31,    
1.69407e-21,	 6.97261e-32,    
8.47033e-22,	 2.46519e-32,    
  4.23516e-22,	   8.71576e-33,  
2.11758e-22,	 3.08149e-33,    
1.05879e-22,	 1.08947e-33,    
5.29396e-23,	 3.85186e-34,    
2.64698e-23,	 1.36184e-34,    
1.32349e-23,	 4.81482e-35,    
  6.61744e-24,	   1.70230e-35,  
3.30872e-24,	 6.01853e-36,    
1.65436e-24,	 2.12787e-36,    
8.27181e-25,	 7.52316e-37,    
4.13590e-25,	 2.65984e-37,    
2.06795e-25,	 9.40395e-38,    
  0.0000000,     0.00000000};	/* let the compiler extend with 0. */

/* Chebychev coefficients of the monomials on [1,2] for minimax fit.  */

/* These should be in registers for the duration of do_grav_cheb.  If not,
   they should be copied into automatic variables */
static const float g0 =   7.05452470f,
		   g1 = -14.85088557f,
		   g2 =  14.70832310f,
		   g3 =  -7.83703555f,
		   g4 =   2.17094576f,
		   g5 =  -0.24598174f;
/* end of karp.h */

void 
do_grav_cheb(const float *p, const float *end, const float *pos0, float *mass0, 
	      float *acc0, float *phi0, const float *eps2p, int *ncut)
{
    float dr2;
    float idr2;
    Vxd(float r);
    float mor3, mass;
    Vxd(float ppos);
    Vxd(float a);
    float phi = 0.F;
    float total_mass = 0.F;
    float eps2 = *eps2p;
    float x, x2, x3;
    unsigned int it;
    int i, n;
    static unsigned int itlast = UINT_MAX;
    static float uval, tval;

    VxS(a, = 0.F);
    VxV(ppos, = pos0);

    n = LOOPCNT(end-p, 1);
    for(i=0; i<n; i++){
	mass = *p++;
	r0 = *p++;
	r1 = *p++;
	r2 = *p++;
	VxVx(r, -= ppos);	 /* (1) */

	dr2 = eps2 + Dotx(r, r); /* (2) dep (1) */

	/* 20 flops (including one flt-int type-pun) */
	idr2 = dr2;
	it = (*(int *)&idr2) >> 23;  /* Pun float to int */
	if( it != itlast ){
	  uval = ut[2*it];
	  tval = ut[2*it+1];
	  itlast = it;
	}
	x = dr2*uval;		/* (4) dep 2 */
	x2 = x*x;		/* (5) dep 4 */
	x3 = x2*x;		/* (6) dep 5 */
	mor3 = g0 + x*g1 + x2*g2 + x3*(g3 + x*g4 + x2*g5); /* (7) dep 4,5 */
	mor3 *= 1.5f - 0.5f*x3*mor3*mor3;  /* (8) dep 7,6 */
	mor3 *= tval*mass;  /* (9) dep 8 */
	total_mass += mass;	/* (10) dep -  */
	phi -= dr2*mor3;	/* (11) dep 9 */

	VxVx(a, += mor3 * r);	/* (12) dep 9 */
    }
    VVx(acc0, += a);
    *mass0 += total_mass;
    *phi0 += phi;
}

void
do_grav_cheb_unroll2(const float *p, const float *end, const float *pos0, 
	       float *mass0, float *acc0, float *phi0, const float 
	       *eps2p, int *ncut)
{
    float dr2a;
    float dr2b;
    float idr2;
    Vxd(float ra);
    Vxd(float rb);
    float massa;
    float massb;
    float mor3a;
    float mor3b;
    Vxd(float ppos);
    Vxd(float aa);
    Vxd(float ab);
    float phia = 0.F;
    float phib = 0.F;
    float total_massa = 0.F;
    float total_massb = 0.F;
    const float eps2 = *eps2p;
    const int n = LOOPCNT(end-p, 2);
    int i;
    float xa, xb;
    float xa2, xb2;
    float xa3, xb3;
    float ua, ub, ulast;
    float ta, tb, tlast;
    unsigned int ita, itb, itlast = UINT_MAX;
    unsigned int it;
    unsigned int itlasta = UINT_MAX;
    unsigned int itlastb = UINT_MAX;

    VxS(aa, = 0.F);
    VxS(ab, = 0.F);
    VxV(ppos, = pos0);

    for (i = 0; i < n; i++) {
	massa = *p++;
	ra0 = *p++;
	ra1 = *p++;
	ra2 = *p++;

	massb = *p++;
	rb0 = *p++;
	rb1 = *p++;
	rb2 = *p++;

	VxVx(ra, -= ppos);
	VxVx(rb, -= ppos);

	dr2a = eps2 + Dotx(ra, ra);
	dr2b = eps2 + Dotx(rb, rb);

#if 0 /* separate 'itlast' values for each of the unrolls */
	ita = (*(int *)&dr2a) >> 23;  /* Pun float to int */
	if( ita != itlasta ){
	  ua = ut[2*ita];
	  ta = ut[2*ita+1];
	  itlasta =ita;
	}
	itb = (*(int *)&dr2b) >> 23;  /* Pun float to int */
	if( itb != itlastb ){
	  ub = ut[2*itb];
	  tb = ut[2*itb+1];
	  itlastb = itb;
	}
#else /* a single itlast carried through from one unroll to the next */
	idr2 = dr2a;
	it = (*(int *)&idr2) >> 23;  /* Pun float to int */
	if( it != itlast ){
	  ulast = ut[2*it];
	  tlast = ut[2*it+1];
	  itlast = it;
	}
	ua = ulast;
	ta = tlast;
	idr2 = dr2b;
	it = (*(int *)&idr2) >> 23;  /* Pun float to int */
	if( it != itlast ){
	  ulast = ut[2*it];
	  tlast = ut[2*it+1];
	  itlast = it;
	}
	ub = ulast;
	tb = tlast;
#endif

	xa = dr2a*ua;  /* Scale input domain to [1,2] */
	xb = dr2b*ub;  /* Scale input domain to [1,2] */

	xa2 = xa*xa;
	xb2 = xb*xb;

	xa3 = xa2*xa;
	xb3 = xb2*xb;

	mor3a = g0 + xa*g1 + xa2*g2 + xa3*(g3 + xa*g4 + xa2*g5);
	mor3b = g0 + xb*g1 + xb2*g2 + xb3*(g3 + xb*g4 + xb2*g5);

	mor3a *= 1.5f - 0.5f*xa3*mor3a*mor3a;  /* N-R iteration */
	mor3a *= ta*massa;  /* Restore result scaling */

	mor3b *= 1.5f - 0.5f*xb3*mor3b*mor3b;  /* N-R iteration */
	mor3b *= tb*massb;  /* Restore result scaling */

	total_massa += massa;
	total_massb += massb;

	phia -= dr2a*mor3a;
	phib -= dr2b*mor3b;

	VxVx(aa, += mor3a * ra);
	VxVx(ab, += mor3b * rb);
    }
    VxVx(aa, += ab);
    VVx(acc0, += aa);
    *mass0 += total_massa + total_massb;
    *phi0 += (phia + phib);
}

void
do_grav_cheb_unroll3(const float *p, const float *end, const float *pos0, 
	       float *mass0, float *acc0, float *phi0, const float 
	       *eps2p, int *ncut)
{
    float dr2a;
    float dr2b;
    float dr2c;
    float idr2;
    Vxd(float ra);
    Vxd(float rb);
    Vxd(float rc);
    float massa;
    float massb;
    float massc;
    float mor3a;
    float mor3b;
    float mor3c;
    Vxd(float ppos);
    Vxd(float aa);
    Vxd(float ab);
    Vxd(float ac);
    float phia = 0.F;
    float phib = 0.F;
    float phic = 0.F;
    float total_massa = 0.F;
    float total_massb = 0.F;
    float total_massc = 0.F;
    const float eps2 = *eps2p;
    const int n = LOOPCNT(end-p, 3);
    int i;
    float xa, xb, xc;
    float xa3, xb3, xc3;
    unsigned int ita, itb, itc;
    float ulast, tlast;
    float ua, ub, uc;
    float ta, tb, tc;
    unsigned int it;
    unsigned int itlast = UINT_MAX;
    unsigned int italast = UINT_MAX;
    unsigned int itblast = UINT_MAX;
    unsigned int itclast = UINT_MAX;

    VxS(aa, = 0.F);
    VxS(ab, = 0.F);
    VxS(ac, = 0.F);
    VxV(ppos, = pos0);

    for (i = 0; i < n; i++) {
	massa = *p++;
	ra0 = *p++;
	ra1 = *p++;
	ra2 = *p++;

	massb = *p++;
	rb0 = *p++;
	rb1 = *p++;
	rb2 = *p++;

	massc = *p++;
	rc0 = *p++;
	rc1 = *p++;
	rc2 = *p++;
	VxVx(ra, -= ppos);
	VxVx(rb, -= ppos);
	VxVx(rc, -= ppos);

	dr2a = eps2 + Dotx(ra, ra);
	dr2b = eps2 + Dotx(rb, rb);
	dr2c = eps2 + Dotx(rc, rc);

#if 0
	ita = (*(int *)&dr2a) >> 23;  /* Pun float to int */
	if( ita != italast ){
	  ua = ut[2*ita];
	  ta = ut[2*ita+1];
	  italast = ita;
	}
	it = (*(int *)&dr2b) >> 23;  /* Pun float to int */
	if( itb != itblast ){
	  ub = ut[2*itb];
	  tb = ut[2*itb+1];
	  itblast = itb;
	}
	itc = (*(int *)&dr2c) >> 23;  /* Pun float to int */
	if( itc != itclast ){
	  uc = ut[2*itc];
	  tc = ut[2*itc+1];
	  itclast = itc;
	}
#else
	idr2 = dr2a;
	it = (*(int *)&idr2) >> 23;  /* Pun float to int */
	if( it != itlast ){
	  ulast = ut[2*it];
	  tlast = ut[2*it+1];
	  itlast = it;
	}
	ua = ulast;
	ta = tlast;
	idr2 = dr2b;
	it = (*(int *)&idr2) >> 23;  /* Pun float to int */
	if( it != itlast ){
	  ulast = ut[2*it];
	  tlast = ut[2*it+1];
	  itlast = it;
	}
	ub = ulast;
	tb = tlast;
	idr2 = dr2c;
	it = (*(int *)&idr2) >> 23;  /* Pun float to int */
	if( it != itlast ){
	  ulast = ut[2*it];
	  tlast = ut[2*it+1];
	  itlast = it;
	}
	uc = ulast;
	tc = tlast;
#endif

	xa = dr2a*ua;  /* Scale input domain to [1,2] */
	xb = dr2b*ub;  /* Scale input domain to [1,2] */
	xc = dr2c*uc;  /* Scale input domain to [1,2] */

	xa3 = xa*xa*xa;
	xb3 = xb*xb*xb;
	xc3 = xc*xc*xc;

	mor3a = g0 + xa*(g1 + xa*(g2)) + xa3*(g3 + xa*(g4 + xa*g5));
	mor3b = g0 + xb*(g1 + xb*(g2)) + xb3*(g3 + xb*(g4 + xb*g5));
	mor3c = g0 + xc*(g1 + xc*(g2)) + xc3*(g3 + xc*(g4 + xc*g5));

	mor3a *= 1.5f - 0.5f*xa3*mor3a*mor3a;  /* N-R iteration */
	mor3a *= ta*massa;  /* Restore result scaling */

	mor3b *= 1.5f - 0.5f*xb3*mor3b*mor3b;  /* N-R iteration */
	mor3b *= tb*massb;  /* Restore result scaling */

	mor3c *= 1.5f - 0.5f*xc3*mor3c*mor3c;  /* N-R iteration */
	mor3c *= tc*massc;  /* Restore result scaling */
	
	total_massa += massa;
	total_massb += massb;
	total_massc += massc;

	phia -= dr2a*mor3a;
	phib -= dr2b*mor3b;
	phic -= dr2c*mor3c;

	VxVx(aa, += mor3a * ra);
	VxVx(ab, += mor3b * rb);
	VxVx(ac, += mor3c * rc);
    }
    VxVx(aa, += ab);
    VxVx(aa, += ac);
    VVx(acc0, += aa);
    *mass0 += total_massa + total_massb + total_massc;
    *phi0 += (phia + phib + phic);
}

int
main(int argc, char **argv)
{
    int i;
    struct body {
	float mass;
	float pos[3];
    } b1;

    b1.mass = 765.43;
    b1.pos[0] = .123456;
    b1.pos[1] = 1.987654;
    b1.pos[2] = 4.334654;
    
    for (i = 0; i < LAST;) {
	dbtab[i] = btab[i] = b1.mass+(i*10.);
	i++;
	dbtab[i] = btab[i] = b1.pos[0]+i;
	i++;
	dbtab[i] = btab[i] = b1.pos[1]+i;
	i++;
	dbtab[i] = btab[i] = b1.pos[2]+i;
	i++;
    }

    dpos0[0] = pos0[0] = -.2134545;
    dpos0[1] = pos0[1] = -.3423456;
    dpos0[2] = pos0[2] = -7.134246;
    
    for (i = LAST; i < LAST + NEXTRA*NELEM; i++)
      dbtab[i] = btab[i] = 0.0;

    e2 = 0.F;
    de2 = 0.;
    do_grav(btab, btab+LAST, pos0, &mass0, accexact, &phiexact, &e2, &ncut);

    printf("# Gravity kernel version: %s \n", VERSION_STRING);
    printf("# 20 flops, one inverse-sqrt, 4 float loads per Newton \n");
    printf("# Cheb does 18 flops and two loads for each inverse-sqrt\n");
    printf("# A vector of %d Newton interactions is computed %d times in each test\n",
	   NVEC, NLOOP);
    do_time(&do_grav, "do_grav");
    do_time(&do_grav_rsqt, "do_grav_rsqt");
    do_time(&do_grav_split, "do_grav_split");
    do_time(&do_grav_split2, "do_grav_split2");
#ifdef DO_GRAV_EXTERNAL
    do_time(DO_GRAV_EXTERNAL, "external");
#endif 
    do_time(&do_grav_unroll2, "do_grav_unroll2");
    do_time(&do_grav_unroll3, "do_grav_unroll3");
    do_time(&do_grav_unroll4, "do_grav_unroll4");
    do_time(&do_grav_nosq, "do_grav(nosqrt)");
    do_time(&do_grav_unro_nosq, "do_grav_unro(nosqrt)");
    do_time(&do_grav_cheb, "do_grav_cheb");
    do_double_time(&do_grav_cheb_dbl, "do_grav_cheb_dbl");
    do_time(&do_grav_cheb_unroll2, "do_grav_cheb_unroll2");
    do_time(&do_grav_cheb_unroll3, "do_grav_cheb_unroll3");
    exit(0);
}

static const double dut[2*256] = {
  1.70141e+38,         DBL_MAX,  
8.50706e+37,	     DBL_MAX,    
4.25353e+37,	     DBL_MAX,    
2.12676e+37,	     DBL_MAX,    
1.06338e+37,	     DBL_MAX,    
5.31691e+36,	     DBL_MAX,    
  2.65846e+36,	      DBL_MAX,   
1.32923e+36,	     DBL_MAX,    
6.64614e+35,	     DBL_MAX,    
3.32307e+35,	     DBL_MAX,    
1.66153e+35,	     DBL_MAX,    
8.30767e+34,	     DBL_MAX,    
  4.15384e+34,	       DBL_MAX,  
2.07692e+34,	     DBL_MAX,    
1.03846e+34,	     DBL_MAX,    
5.19230e+33,	     DBL_MAX,    
2.59615e+33,	     DBL_MAX,    
1.29807e+33,	     DBL_MAX,    
  6.49037e+32,	       DBL_MAX,  
3.24519e+32,	     DBL_MAX,    
1.62259e+32,	     DBL_MAX,    
8.11296e+31,	     DBL_MAX,    
4.05648e+31,	     DBL_MAX,    
2.02824e+31,	     DBL_MAX,    
  1.01412e+31,	       DBL_MAX,  
5.07060e+30,	     DBL_MAX,    
2.53530e+30,	     DBL_MAX,    
1.26765e+30,	     DBL_MAX,    
6.33825e+29,	     DBL_MAX,    
3.16913e+29,	     DBL_MAX,    
  1.58456e+29,	       DBL_MAX,  
7.92282e+28,	     DBL_MAX,    
3.96141e+28,	     DBL_MAX,    
1.98070e+28,	     DBL_MAX,    
9.90352e+27,	     DBL_MAX,    
4.95176e+27,	     DBL_MAX,    
  2.47588e+27,	       DBL_MAX,  
1.23794e+27,	     DBL_MAX,    
6.18970e+26,	     DBL_MAX,    
3.09485e+26,	     DBL_MAX,    
1.54743e+26,	     DBL_MAX,    
7.73713e+25,	     DBL_MAX,    
  3.86856e+25,	   2.40616e+38,  
1.93428e+25,	 8.50706e+37,    
9.67141e+24,	 3.00770e+37,    
4.83570e+24,	 1.06338e+37,    
2.41785e+24,	 3.75962e+36,    
1.20893e+24,	 1.32923e+36,    
  6.04463e+23,	   4.69953e+35,  
3.02231e+23,	 1.66153e+35,    
1.51116e+23,	 5.87441e+34,    
7.55579e+22,	 2.07692e+34,    
3.77789e+22,	 7.34302e+33,    
1.88895e+22,	 2.59615e+33,    
  9.44473e+21,	   9.17877e+32,  
4.72237e+21,	 3.24519e+32,    
2.36118e+21,	 1.14735e+32,    
1.18059e+21,	 4.05648e+31,    
5.90296e+20,	 1.43418e+31,    
2.95148e+20,	 5.07060e+30,    
  1.47574e+20,	   1.79273e+30,  
7.37870e+19,	 6.33825e+29,    
3.68935e+19,	 2.24091e+29,    
1.84467e+19,	 7.92282e+28,    
9.22337e+18,	 2.80114e+28,    
4.61169e+18,	 9.90352e+27,    
  2.30584e+18,	   3.50142e+27,  
1.15292e+18,	 1.23794e+27,    
5.76461e+17,	 4.37678e+26,    
2.88230e+17,	 1.54743e+26,    
1.44115e+17,	 5.47097e+25,    
7.20576e+16,	 1.93428e+25,    
  3.60288e+16,	   6.83872e+24,  
1.80144e+16,	 2.41785e+24,    
9.00720e+15,	 8.54840e+23,    
4.50360e+15,	 3.02231e+23,    
2.25180e+15,	 1.06855e+23,    
1.12590e+15,	 3.77789e+22,    
  5.62950e+14,	   1.33569e+22,  
2.81475e+14,	 4.72237e+21,    
1.40737e+14,	 1.66961e+21,    
7.03687e+13,	 5.90296e+20,    
3.51844e+13,	 2.08701e+20,    
1.75922e+13,	 7.37870e+19,    
  8.79609e+12,	   2.60876e+19,  
4.39805e+12,	 9.22337e+18,    
2.19902e+12,	 3.26095e+18,    
1.09951e+12,	 1.15292e+18,    
5.49756e+11,	 4.07619e+17,    
2.74878e+11,	 1.44115e+17,    
  1.37439e+11,	   5.09524e+16,  
6.87195e+10,	 1.80144e+16,    
3.43597e+10,	 6.36905e+15,    
1.71799e+10,	 2.25180e+15,    
8.58993e+09,	 7.96131e+14,    
4.29497e+09,	 2.81475e+14,    
  2.14748e+09,	   9.95164e+13,  
1.07374e+09,	 3.51844e+13,    
5.36871e+08,	 1.24396e+13,    
2.68435e+08,	 4.39805e+12,    
1.34218e+08,	 1.55494e+12,    
6.71089e+07,	 5.49756e+11,    
  3.35544e+07,	   1.94368e+11,  
1.67772e+07,	 6.87195e+10,    
8.38861e+06,	 2.42960e+10,    
4.19430e+06,	 8.58993e+09,    
2.09715e+06,	 3.03700e+09,    
1.04858e+06,	 1.07374e+09,    
      524288.,	   3.79625e+08,  
    262144.,	 1.34218e+08,    
    131072.,	 4.74531e+07,    
    65536.0,	 1.67772e+07,    
    32768.0,	 5.93164e+06,    
    16384.0,	 2.09715e+06,    
      8192.00,	       741455.,  
    4096.00,	     262144.,    
    2048.00,	     92681.9,    
    1024.00,	     32768.0,    
    512.000,	     11585.2,    
    256.000,	     4096.00,    
      128.000,	       1448.15,  
    64.0000,	     512.000,    
    32.0000,	     181.019,    
    16.0000,	     64.0000,    
    8.00000,	     22.6274,    
    4.00000,	     8.00000,    
      2.00000,	       2.82843,  
    1.00000,	     1.00000,    
   0.500000,	    0.353553,    
   0.250000,	    0.125000,    
   0.125000,	   0.0441942,    
  0.0625000,	   0.0156250,    
    0.0312500,	    0.00552427,  
  0.0156250,	  0.00195312,    
 0.00781250,	 0.000690534,    
 0.00390625,	 0.000244141,    
 0.00195312,	 8.63167e-05,    
0.000976562,	 3.05176e-05,    
  0.000488281,	   1.07896e-05,  
0.000244141,	 3.81470e-06,    
0.000122070,	 1.34870e-06,    
6.10352e-05,	 4.76837e-07,    
3.05176e-05,	 1.68587e-07,    
1.52588e-05,	 5.96046e-08,    
  7.62939e-06,	   2.10734e-08,  
3.81470e-06,	 7.45058e-09,    
1.90735e-06,	 2.63418e-09,    
9.53674e-07,	 9.31323e-10,    
4.76837e-07,	 3.29272e-10,    
2.38419e-07,	 1.16415e-10,    
  1.19209e-07,	   4.11590e-11,  
5.96046e-08,	 1.45519e-11,    
2.98023e-08,	 5.14488e-12,    
1.49012e-08,	 1.81899e-12,    
7.45058e-09,	 6.43110e-13,    
3.72529e-09,	 2.27374e-13,    
  1.86265e-09,	   8.03887e-14,  
9.31323e-10,	 2.84217e-14,    
4.65661e-10,	 1.00486e-14,    
2.32831e-10,	 3.55271e-15,    
1.16415e-10,	 1.25607e-15,    
5.82077e-11,	 4.44089e-16,    
  2.91038e-11,	   1.57009e-16,  
1.45519e-11,	 5.55112e-17,    
7.27596e-12,	 1.96262e-17,    
3.63798e-12,	 6.93889e-18,    
1.81899e-12,	 2.45327e-18,    
9.09495e-13,	 8.67362e-19,    
  4.54747e-13,	   3.06659e-19,  
2.27374e-13,	 1.08420e-19,    
1.13687e-13,	 3.83323e-20,    
5.68434e-14,	 1.35525e-20,    
2.84217e-14,	 4.79154e-21,    
1.42109e-14,	 1.69407e-21,    
  7.10543e-15,	   5.98943e-22,  
3.55271e-15,	 2.11758e-22,    
1.77636e-15,	 7.48678e-23,    
8.88178e-16,	 2.64698e-23,    
4.44089e-16,	 9.35848e-24,    
2.22045e-16,	 3.30872e-24,    
  1.11022e-16,	   1.16981e-24,  
5.55112e-17,	 4.13590e-25,    
2.77556e-17,	 1.46226e-25,    
1.38778e-17,	 5.16988e-26,    
6.93889e-18,	 1.82783e-26,    
3.46945e-18,	 6.46235e-27,    
  1.73472e-18,	   2.28479e-27,  
8.67362e-19,	 8.07794e-28,    
4.33681e-19,	 2.85598e-28,    
2.16840e-19,	 1.00974e-28,    
1.08420e-19,	 3.56998e-29,    
5.42101e-20,	 1.26218e-29,    
  2.71051e-20,	   4.46247e-30,  
1.35525e-20,	 1.57772e-30,    
6.77626e-21,	 5.57809e-31,    
3.38813e-21,	 1.97215e-31,    
1.69407e-21,	 6.97261e-32,    
8.47033e-22,	 2.46519e-32,    
  4.23516e-22,	   8.71576e-33,  
2.11758e-22,	 3.08149e-33,    
1.05879e-22,	 1.08947e-33,    
5.29396e-23,	 3.85186e-34,    
2.64698e-23,	 1.36184e-34,    
1.32349e-23,	 4.81482e-35,    
  6.61744e-24,	   1.70230e-35,  
3.30872e-24,	 6.01853e-36,    
1.65436e-24,	 2.12787e-36,    
8.27181e-25,	 7.52316e-37,    
4.13590e-25,	 2.65984e-37,    
2.06795e-25,	 9.40395e-38,    
  0.0000000,     0.00000000};	/* let the compiler extend with 0. */

static const double dg0 =   7.05452470f,
		   dg1 = -14.85088557f,
		   dg2 =  14.70832310f,
		   dg3 =  -7.83703555f,
		   dg4 =   2.17094576f,
		   dg5 =  -0.24598174f;
void 
do_grav_cheb_dbl(const double *p, const double *end, const double *pos0, double *mass0, 
	      double *acc0, double *phi0, const double *eps2p, int *ncut)
{
  double dr2;
    double idr2;
    Vxd(double r);
    double mor3, mass;
    Vxd(double ppos);
    Vxd(double a);
    double phi = 0.;
    double total_mass = 0.;
    double eps2 = *eps2p;
    double x, x2, x3;
    unsigned int it;
    int i, n;
    static unsigned int itlast = UINT_MAX;
    static double uval, tval;

    VxS(a, = 0.);
    VxV(ppos, = dpos0);

    n = LOOPCNT(end-p, 1);
    for(i=0; i<n; i++){
	mass = *p++;
	r0 = *p++;
	r1 = *p++;
	r2 = *p++;
	VxVx(r, -= ppos);	 /* (1) */

	dr2 = eps2 + Dotx(r, r); /* (2) dep (1) */

	/* 20 flops (including one flt-int type-pun) */
	idr2 = dr2;
	/* This is a real headache to write completely portably.
	   Let's just assume that doubles are 64bits and ints are 32.
	   Then we just have to deal with endianness */
#ifdef BIGENDIAN
	it = ((((int *)&idr2)[0]) >> 20) + (127-1023);
#else
	it = ((((int *)&idr2)[1]) >> 20) + (127-1023);
#endif       
	if( it != itlast ){
	  uval = dut[2*it];
	  tval = dut[2*it+1];
	  itlast = it;
	}
	x = dr2*uval;		/* (4) dep 2 */
	x2 = x*x;		/* (5) dep 4 */
	x3 = x2*x;		/* (6) dep 5 */
	mor3 = dg0 + x*dg1 + x2*dg2 + x3*(dg3 + x*dg4 + x2*dg5); /* (7) dep 4,5 */
	mor3 *= 1.5 - 0.5*x3*mor3*mor3;  /* (8) dep 7,6 */
	mor3 *= tval*mass;  /* (9) dep 8 */
	total_mass += mass;	/* (10) dep -  */
	phi -= dr2*mor3;	/* (11) dep 9 */

	VxVx(a, += mor3 * r);	/* (12) dep 9 */
    }
    VVx(acc0, += a);
    *mass0 += total_mass;
    *phi0 += phi;
}
