/* A first pass at converting an SDF file to an AVS field. */
/* 
   We use AVS' ability to do "indirection" and parse data descriptions.
   I don't really want to figure out a command line syntax that
   can handle all the possible permutation of coordinates and vectors
   and ndim and nspace and so forth.  Thus, we look for "x", "y" and "z"
   to be the coordinates in our irregular 1-dim field.  More functionality
   can be added as necessary.  When SDF becomes really capable of handling
   multi-dim arrays this will become a real mess...

   This would work oh-so-much better if AVS could deal with byte-swapping!

   And isn't it special that the AVS parser can't deal with spaces
   around the '='.
*/

#include <stdio.h>
#include "Assert.h"
#include "SDF.h"
#include "protos.h"

extern int getopt(int, char **, const char *);
extern char *optarg;
extern int optind, opterr;

static void 
do_filespec(const char *name);

static char *fname;
static SDF *sdfp;
static int nrec;

int main(int argc, char **argv){
    char *hdr;
    char *name;
    int verbose;
    int c;
    enum SDF_type_enum t;
    int nspace;
    int i;

    hdr = NULL;
    verbose = 0;
    while((c=getopt(argc, argv, "h:v")) != -1)
	switch(c){
	case 'h':
	    hdr = optarg;
	    break;
	case 'v':
	    verbose = 1;
	    break;
	case '?':
	    fprintf(stderr, "Usage: %s [-h hdrfile] sdffile name ...\n", argv[0]);
	    exit(0);
	}

    fname = argv[optind];
    optind++;
    sdfp = SDFopen(hdr, fname);
    if( sdfp == NULL ){
	Error("Could not SDFopen(%s, %s): %s\n", hdr, fname, SDFerrstring);
    }
    
    printf("# AVS\n");
    if( SDFhasname("z", sdfp) ){
	nspace = 3;
	nrec = SDFnrecs("z", sdfp);
#ifndef NO_STRINGIFICATION
	assert(SDFhasname("y", sdfp) && SDFhasname("x", sdfp));
	assert(SDFnrecs("y", sdfp) == nrec && SDFnrecs("x", sdfp)==nrec);
#endif
    }else if( SDFhasname("y", sdfp) ){
	nspace = 2;
	nrec = SDFnrecs("y", sdfp);
#ifndef NO_STRINGIFICATION
	assert(SDFhasname("x", sdfp));
	assert(SDFnrecs("x", sdfp)==nrec);
#endif
    }else if( SDFhasname("x", sdfp) ){
	nrec = SDFnrecs("x", sdfp);
	nspace = 1;
    }else{
	Error("Can't find coordinates in file!\n");
    }
    t = SDFtype("x", sdfp);
    if( t != SDF_FLOAT ){
	Error("We can only deal with floats for now.\n");
    }

    printf("ndim=1\n");
    printf("field=irregular\n");
    printf("nspace=%d\n", nspace);
    printf("data=float\n");
    printf("dim1=%d\n", nrec);
    printf("veclen=%d\n", argc-optind);
    printf("coord 1 ");
    do_filespec("x");
    if( nspace > 1 ){
	printf("coord 2 ");
	do_filespec("y");
    }
    if( nspace > 2 ){
	printf("coord 3 ");
	do_filespec("z");
    }


    i = 1;
    printf("#\n");
    for(; optind < argc; optind++){
	name = argv[optind];
	printf("variable %d ", i++);
	do_filespec(name);
	printf("label=%s\n#\n", name);
    }

    SDFclose(sdfp);
    exit(0);
}

static void 
do_filespec(const char *name){
     int off, str;
     enum SDF_type_enum t;

     off = SDFfileoffset(name, sdfp);
     if( off == -1 ){
	 Error("No offset for %s in %s: %s\n", name, fname, SDFerrstring);
     }
     t = SDFtype(name, sdfp);
     if( t != SDF_FLOAT ){
	 Error("%s is not a float value!\n", name);
     }
     str = SDFfilestride(name, sdfp);
     if( str % sizeof(float) ){
	 Error("The SDF record size must be a multiple of sizeof(float)\n");
     }
     printf("file=%s filetype=binary skip=%d stride=%d # %s\n", 
	    fname, off, str/sizeof(float), name);
 }
     
