/* Write out six panels projecting the particle onto the faces of a cube */
/* An icosahedron would incur less edge-to-edge distortion, but would require
   more mucking around... */

#include <math.h>
#include <stdio.h>
#include "protos.h"
#include "Malloc.h"
#include "physics_n.h"
#include "fastflpt.h"
#include "vop.h"
#include "mpmy.h"

#if NDIM != 3
  # error this code only makes sense for 3-d simulations
#endif

#define NORTH 0
#define SOUTH 1
#define EAST  2
#define WEST  3
#define UP    4
#define DOWN  5

struct accumulator{
  float m;
  float bright;			/* mass/r^2 */
  float vaway;			/* brightness-weighted recession-velocity */
} *faces[6];

static void Accumulate(int npix);
static void Output(const char *base, int npix);

void
cubeimage(int npix, int nobj, body *btab, const char *base){
  int f;
  body *bp, *btabend;
  float npix2;

  for(f=0; f<6; f++){
    faces[f] = Calloc(npix*npix, sizeof(struct accumulator));
  }
  
  npix2 = npix/2.000001;
  btabend = btab + nobj;
  for(bp=btab; bp<btabend; bp++){
    float absx, absy, absz;
    float u, v;
    int i, j;
    int idx;
    float r2;

    absx = fabs(bp->pos[0]);
    absy = fabs(bp->pos[1]);
    absz = fabs(bp->pos[2]);

    /* Figure out which face we project onto by finding the largest
       absolute coordinate.  Then project intu (u,v) \in ([-1,1], [-1,1]) */
    if( absx > absy ){
      if( absx > absz ){
	f = (bp->pos[0] > 0)? EAST : WEST;
	u = bp->pos[1]/absx;
	v = bp->pos[2]/absx;
      }else{
	f = (bp->pos[2] > 0)? UP : DOWN;
	u = bp->pos[0]/absz;
	v = bp->pos[1]/absz;
      }
    }else{
      if( absy > absz ){
	f = (bp->pos[1] > 0)? NORTH : SOUTH;
	u = bp->pos[0]/absy;
	v = bp->pos[2]/absy;
      }else{
	f = (bp->pos[2] > 0)? UP : DOWN;
	u = bp->pos[0]/absz;
	v = bp->pos[1]/absz;
      }
    }	

    /* Figure out which pixel to increment. */
    i = (u+1.0F)*npix2;
    j = (v+1.0F)*npix2;
    idx = j*npix + i;

    /* Now accumulate whatever it is we are trying to sum about
       each pixel. */
    r2 = Dot(bp->pos, bp->pos);
    faces[f][idx].m += bp->mass;
    faces[f][idx].bright += bp->mass/r2;
    faces[f][idx].vaway += Dot(bp->pos, bp->vel) * bp->mass / (r2*sqrtf_fast(r2));
  }

  Accumulate(npix);
  Output(base, npix);

  for(f=0; f<6; f++){
    Free(faces[f]);
  }
}

static void 
Accumulate(int npix){
  /* This is the most naive possible implementation.  It would be
     MUCH better to do something like Run-length-encode all the empty
     values in the faces array, and communicate only the reduced data. */
  int f;

  for(f=0; f<6; f++){
    int p;
    
    MPMY_Combine(faces[f], faces[f], npix*npix*(sizeof(struct accumulator)/sizeof(float)), MPMY_FLOAT, MPMY_SUM);
    for(p=0; p<npix*npix; p++){
      if( faces[f][p].bright > 0. ){
	faces[f][p].vaway /= faces[f][p].bright;
      }
    }
  }
}

/* How much processing do we want to do on output?  Probably not much.
   Fiddling with color maps, orientations, etc. is probably best left to
   post-processing. */
static void 
Output(const char *base, int npix){
  char *ext[6];
  int f;

  if( MPMY_Procnum() != 0 )
    return;

  ext[EAST] = "e";
  ext[SOUTH] = "s";
  ext[NORTH] = "n";
  ext[WEST] = "w";
  ext[UP] = "u";
  ext[DOWN] = "d";
    
  for(f=0; f<6; f++){
    FILE *fp;
    char name[128];
    sprintf(name, "%s.%s", base, ext[f]);
    fp = fopen(name, "w");
    fwrite(faces[f], sizeof(struct accumulator), npix*npix, fp);
    fclose(fp);
  }
}      
