/* Read an 'equal-area' file, and write out a ppm file.
   This requires lots of fiddling to get a nice-looking image.
   It really ought to be interactive, or at least cmd-line driven.
   For now, we just edit the code. */

#include <math.h>
#include <stdio.h>
#include <stddef.h>
#include <getopt.h>
#include <stdlib.h>
#include "protos.h"
#include "SDF.h"
#include "error.h"
#include "readbodies.h"

float mweightmax, mweightmin;
float bweightmax, bweightmin;
float vrmmax, vrmmin;
float vrbmax, vrbmin;
float logbratio;

FILE *outfp;

/* This doesn't have to be the same as in eaimage.h, but it saves thinking */
struct accumulator{
  float mweight;
  float bweight;		/* mass/r^2 */
  float vrm;			/* mass-weighted radial-velocity */
  float vrb;			/* brightness-weighted radial-velocity */
};

static void 
do_Zpixel(const struct accumulator *pix){
  if( pix->bweight == 0. ){
    putc('\0', outfp);
  }else{
    /* Map the radial-velocity to hue in (5/6,0), m->b->c->g->y->r */
    if( pix->vrb < vrbmin )
      putc( 2, outfp);
    else if( pix->vrb > vrbmax )
      putc( 254, outfp);
    else{
      float ratio = (pix->vrb - vrbmin) / (vrbmax - vrbmin);
      /* Save 0 for black 1 for white and 255 for white also */
      putc((char)( 2. + 252.*ratio), outfp);
    }
  }
}

static void 
do_Lpixel(const struct accumulator *pix){
  float v;

  if( pix->bweight == 0. ){
    putc('\0', outfp);
  }else{
    /* Map the 'brightness' weight logarithmically to 'value',
       i.e., gray-scale, and then gamma-correct.  It is also
       worth trying to map it to hue, possibly multi-wrapped.*/
    if( pix->bweight < bweightmin )
      putc( 2, outfp);
    else if( pix->bweight > bweightmax )
      putc( 254, outfp);
    else{
      v = log( pix->bweight / bweightmin ) / logbratio;
      putc((char)(2. + 252.*v), outfp);
    }
  }
}

static float
magnitude(float bweight){
  return -10.28 - 2.5*log10(bweight);
}

struct name_address_s na_arr[8];

/* Here's a neat little macro that might be helpful in constructing
   a name_address array called na_arr. */
#define Map(i, Name, member) do{\
  na_arr[i].name = Name;\
  na_arr[i].offset = offsetof(struct accumulator, member);\
}while(0)

#define BWMAX_OPT 1001
#define BWMIN_OPT 1002
#define VRBMAX_OPT 1003
#define VRBMIN_OPT 1004

struct option optab[] = {
  {"bweightmax", required_argument, NULL, BWMAX_OPT},
  {"bweightmin", required_argument, NULL, BWMIN_OPT},
  {"vrbmax", required_argument, NULL, VRBMAX_OPT},
  {"vrbmin", required_argument, NULL, VRBMIN_OPT},
  {NULL, 0, NULL, 0}
};

int
main(int argc, char **argv){
  char *filename;
  char *outputbase;
  char output[256];
  int width, height;
  int optopt, c;
  SDF *sdfp;
  int vrbmax_arg=0, vrbmin_arg=0;
  int bweightmax_arg=0, bweightmin_arg=0;

  for(optopt=0; 
      (c=getopt_long_only(argc, argv, "", optab, NULL)) !=EOF;
      optopt=0){
    switch(c){
    case 0:
      /* a long option with a non-NULL flag */
      break;
    case BWMAX_OPT:
      bweightmax = atof(optarg);
      bweightmax_arg = 1;
      break;
    case BWMIN_OPT:
      bweightmin = atof(optarg);
      bweightmin_arg = 1;
      break;
    case VRBMAX_OPT:
      vrbmax = atof(optarg);
      vrbmax_arg = 1;
      break;
    case VRBMIN_OPT:
      vrbmin = atof(optarg);
      vrbmin_arg = 1;
      break;
    }
  }
  if( optind+1 > argc ){
    fprintf(stderr, "Usage:  %s [options] inputfile outputbase\n", argv[0]);
    exit(1);
  }
  filename = argv[optind];
  outputbase = argv[optind+1];

  sdfp = SDFopen(NULL, filename);
  if(sdfp == NULL ){
    Error("Could not SDFopen(NULL, %s), errstring=%s\n", filename, SDFerrstring);
  }
  SDFgetfloatOrDie(sdfp, "mweightmax", &mweightmax);
  SDFgetfloatOrDie(sdfp, "mweightmin", &mweightmin);
  SDFgetfloatOrDie(sdfp, "vrmmax", &vrmmax);
  SDFgetfloatOrDie(sdfp, "vrmmin", &vrmmin);

  if(!bweightmax_arg)
    SDFgetfloatOrDie(sdfp, "bweightmax", &bweightmax);
  if(!bweightmin_arg)
    SDFgetfloatOrDie(sdfp, "bweightmin", &bweightmin);
  if(!vrbmax_arg)
    SDFgetfloatOrDie(sdfp, "vrbmax", &vrbmax);
  if(!vrbmin_arg)
    SDFgetfloatOrDie(sdfp, "vrbmin", &vrbmin);
  SDFgetintOrDie(sdfp, "x_pixels", &width);
  SDFgetintOrDie(sdfp, "y_pixels", &height);
  SDFclose(sdfp);



  logbratio = log(bweightmax/bweightmin);
  Map(0, "mweight", mweight);
  Map(1, "bweight", bweight);
  Map(2, "vrb", vrb);
  Map(3, "vrm", vrm);

  
  sprintf(output, "%s.v.pgm", outputbase);
  outfp = fopen(output, "w");
  fprintf(outfp, "P5 # PGM-raw\n");
  fprintf(outfp, "%d %d\n", width, height);
  fprintf(outfp, "# Equal area map of redshift velocity from %s.\n",
	  filename);
  fprintf(outfp, "# Pix-range [2,254) corresponds to\n");
  fprintf(outfp, "#brightness-weighted average redshift velocities in:\n");
  fprintf(outfp, "# (%g, %g)\n", vrbmin, vrbmax);
  fprintf(outfp, "255\n");
  ReadBodies(filename, 4, sizeof(struct accumulator), na_arr, do_Zpixel);
  fclose(outfp);
  sprintf(output, "%s.v.txt", outputbase);
  outfp = fopen(output, "w");
  fprintf(outfp, "Recession velocity in range (%#5.2g, %#5.2g)", vrbmin, vrbmax);
  fclose(outfp);
  
  sprintf(output, "%s.m.pgm", outputbase);
  outfp = fopen(output, "w");
  fprintf(outfp, "P5 # PGM-raw\n");
  fprintf(outfp, "%d %d\n", width, height);
  fprintf(outfp, "# Equal-area map of 'apparent magnitudes' from %s.\n",
	  filename);
  fprintf(outfp, "# define m = -10.28 - 2.5*log10( M/M_{10} / (r/1kpc)^2\n" );
  fprintf(outfp, "# This would be right if light followed mass and solar values were correct for M/L\n");
  fprintf(outfp, "# Pix-range [2, 254) corresponds to\n");
  fprintf(outfp, "#magnitudes in the range:\n");
  fprintf(outfp, "# (%g, %g)\n",
	 magnitude(bweightmin), magnitude(bweightmax));
  fprintf(outfp, "255\n");
  ReadBodies(filename, 4, sizeof(struct accumulator), na_arr, do_Lpixel);
  fclose(outfp);
  sprintf(output, "%s.m.txt", outputbase);
  outfp = fopen(output, "w");
  fprintf(outfp, "App. mag in range (%#5.2g, %#5.2g)", 
	  magnitude(bweightmin), magnitude(bweightmax));
  fclose(outfp);
  
  exit(0);
}

