/* Read an 'equal-area' file, and write out a ppm file.
   This requires lots of fiddling to get a nice-looking image.
   It really ought to be interactive, or at least cmd-line driven.
   For now, we just edit the code. */

#include <math.h>
#include <stdio.h>
#include <stddef.h>
#include "protos.h"
#include "SDF.h"
#include "error.h"
#include "readbodies.h"
#include "eaimage.h"

static void hsvtorgb(float h, float s, float v, float *r, float *g, float *b);

float mweightmax, mweightmin;
float bweightmax, bweightmin;
float vrmmax, vrmmin;
float vrbmax, vrbmin;
float logbratio;

static void 
do_Zpixel(const struct accumulator *pix){
  float h;
  float r, g, b;

  if( pix->bweight == 0. ){
    r = g = b = 0;
  }else{
    /* Map the radial-velocity to hue in (5/6,0), m->b->c->g->y->r */
    h = 0.83333*(vrbmax - pix->vrb) / (vrbmax - vrbmin);
    hsvtorgb(h, 1.0F, 1.0F, &r, &g, &b);
  }
  putchar( (char)(r*255) );
  putchar( (char)(g*255) );
  putchar( (char)(b*255) );
}

static void 
do_Lpixel(const struct accumulator *pix){
  float v;
  float r, g, b;

  if( pix->bweight == 0. ){
    r = g = b = 0;
  }else{
    /* Map the 'brightness' weight logarithmically to 'value',
       i.e., gray-scale, and then gamma-correct.  It is also
       worth trying to map it to hue, possibly multi-wrapped.*/
    v = log( pix->bweight / bweightmin ) / logbratio;
    /* r = g = b = pow(v, 1/1.6666); */
    hsvtorgb(2.*v, 1.0F, 1.0F, &r, &g, &b);
  }
  putchar( (char)(r*255) );
  putchar( (char)(g*255) );
  putchar( (char)(b*255) );
}

struct name_address_s na_arr[8];

/* Here's a neat little macro that might be helpful in constructing
   a name_address array called na_arr. */
#define Map(i, Name, member) do{\
  na_arr[i].name = Name;\
  na_arr[i].offset = offsetof(struct accumulator, member);\
}while(0)

int
main(int argc, char **argv){
  char *filename = argv[1];
  int width, height;
  SDF *sdfp;

  sdfp = SDFopen(NULL, filename);
  SDFgetfloatOrDie(sdfp, "mweightmax", &mweightmax);
  SDFgetfloatOrDie(sdfp, "mweightmin", &mweightmin);
  SDFgetfloatOrDie(sdfp, "vrmmax", &vrmmax);
  SDFgetfloatOrDie(sdfp, "vrmmin", &vrmmin);

  SDFgetfloatOrDie(sdfp, "bweightmax", &bweightmax);
  SDFgetfloatOrDie(sdfp, "bweightmin", &bweightmin);
  SDFgetfloatOrDie(sdfp, "vrbmax", &vrbmax);
  SDFgetfloatOrDie(sdfp, "vrbmin", &vrbmin);
  SDFgetintOrDie(sdfp, "x_pixels", &width);
  SDFgetintOrDie(sdfp, "y_pixels", &height);
  SDFclose(sdfp);

  logbratio = log(bweightmax/bweightmin);
  Map(0, "mweight", mweight);
  Map(1, "bweight", bweight);
  Map(2, "vrb", vrb);
  Map(3, "vrm", vrm);

  printf("P6\n");
  printf("%d %d\n", width, 2*height);
  printf("255\n");
  ReadBodies(filename, 4, sizeof(struct accumulator), na_arr, do_Zpixel);
  ReadBodies(filename, 4, sizeof(struct accumulator), na_arr, do_Lpixel);
  exit(0);
}

/* Lifted and edited from ppmbrighten. */
static void
hsvtorgb( float h, float s, float v, float *r, float *g, float *b )
{
    if ( s == 0 )
    {
        *r = v;
        *g = v;
        *b = v;
    } else {
        float f, m, n, k;
        long i;

        h = h * 6.0;
        i = (int)floor(h);
        f = h - i;
        m = (v * (1.0 - s));
        n = (v * (1.0 - (s * f)));
        k = (v * (1.0 - (s * (1.0 - f))));

	/* i%6 lets the hue 'wrap' around for values outside of [0,1] */
        switch (i%6)
        {
        case 0:
            *r = v;
            *g = k;
            *b = m;
            break;
        case 1:
            *r = n;
            *g = v;
            *b = m;
            break;
        case 2:
            *r = m;
            *g = v;
            *b = k;
            break;
        case 3:
            *r = m;
            *g = n;
            *b = v;
            break;
        case 4:
            *r = k;
            *g = m;
            *b = v;
            break;
        case 5:
            *r = v;
            *g = m;
            *b = n;
            break;
        }
    }
}
